"""
Add UUIDS in data_dict Dict nodes contained in all groups in the database
This is needed to correctly identify structures if the groups are
exported as archives and then imported into other databases.
"""
from aiida import load_profile
from aiida.orm.utils import load_node, load_group
import project_utils as pu

load_profile()


def add_uuids_to_data_latest_data_dict(group_label_or_instance):
    """
    Take add corresponding UUIDs for all PKs listed in the most recent
    data_dict node in the group (if any)

    Args:
        group_label_or_instance: Group, int or label

    Returns:
        Dict or none
            return none if the onsidered group does not contain
            data_dicts or if the data_dcts already contain UUIDs.
    """
    # get data_dict Dict nodes in group. They are ordered by descending ctime,
    # meaning that the first is the most relevant.
    g = pu.get_group_from_str_or_instance(group_label_or_instance)
    data_dicts = pu.get_data_dicts_in_group(g)
    if not len(data_dicts):
        print(f"Group <{g.pk}> {g.label} contains no data_dict")
        return

    D = data_dicts[0]
    new_D = data_dicts[0].clone()
    new_keys = []
    for k, v in D.get_dict().items():
        if "_pk" in k:
            _l = k.split("_")
            if "pk" in _l[-1]:
                new_k = "_".join(_l[:-1] + ["uuids"])
                if new_k not in new_D.keys():
                    new_D[new_k] = [load_node(pk).uuid for pk in v]
                    new_keys.append(new_k)
                else:
                    print(f"Key \'{new_k}\' already exists in data_dict.")
            else:
                print(f"Nothing will be done with key {k}")
                continue
    if not len(new_keys):
        print("No UUIDs added to data_dict <{}> in group <{}> {}.".format(
            D.pk, g.pk, g.label))
        return

    new_D.store()
    g.add_nodes(new_D)
    print(f"data_dict node {new_D} has been added to group <{g.label}> {g.pk}")
    print(pu.print_dict_of_lists_content(new_D.get_dict()))

    return new_D


for group_id in range(26):
    try:
        g = load_group(group_id)
    except Exception as e:
        print(e)
        continue

    print(f"Group <{g.pk}> {g.label}")
    node_type_counts = {}
    for n in g.nodes:
        if type(n) in node_type_counts.keys():
            node_type_counts[type(n)] += 1
        else:
            node_type_counts[type(n)] = 1

    [print(f"\t{k}: {v} nodes.") for k, v in node_type_counts.items()]
    add_uuids_to_data_latest_data_dict(g)

print('Done.')
