# aiida_utils module
"""
ADD description


TODO:
    create modules:
            build_and_submit.py
                * vasp_relax
                * vasp_dos
                * vasp_bands
                * vasp_multistep_relax

"""

from aiida.orm import StructureData, KpointsData, Int, Float, Str, Bool, \
    Code, Dict, Group, List, QueryBuilder, RemoteData, FolderData, ArrayData, \
    SinglefileData, WorkChainNode
from aiida.orm.nodes import Node, ProcessNode
from aiida.common import AttributeDict, NotExistent, NotExistentAttributeError
from aiida.plugins import DataFactory, WorkflowFactory
from aiida.engine import submit, run_get_node, calcfunction
from aiida.orm.utils import load_node, load_group

from pymatgen.core.structure import Structure, IStructure, Molecule, IMolecule
from pymatgen.analysis.diffraction.neutron import NDCalculator
from pymatgen.analysis.diffraction. xrd import XRDCalculator
from pymatgen.io.vasp.outputs import Vasprun, Outcar
from pymatgen.io.vasp.inputs import Potcar
from pymatgen.analysis.nmr import ChemicalShielding, ElectricFieldGradient
from pymatgen.io.vasp.sets import MPNMRSet
from pymatgen.core.periodic_table import Element

from pyama.diffractionPkg.nanopdf import nanopdfData
from pyama.structureComparisonsPkg.distanceTools import distanceMatrixData
from pyama.uspexAnalysesPkg.uspexDataManipulation import uspexStructuresData

import os
import sys
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import tempfile

__version__ = '2023.04.19.01'

def get_versions():
    """
    Get version of modules used here as a dict
    """
    import aiida
    import aiida_vasp
    versions_dict = {
        'aiida_utils': __version__,
        'aiida-core': aiida.get_version(),
        'aiida-vasp': aiida_vasp.__version__,
    }
    return versions_dict


def get_mp_api_key_from_env_var():
    """
    Get Materials Project MP key from an environment variable
    
    Note: this is probably not a safe way to store a key 
    """
    mp_api_key = os.environ.get('MP_API_KEY')
    if mp_api_key:
        raise ValueError('No environement variable MP_API_KEY. Define it in your environment variables.')
    return mp_api_key


def get_structure_from_file(filename, primitive=False, sort=False,
                            merge_tol=0.0, description=None, verbosity=0,
                            store=False):
    """
    get a StructureData instance from a file using pymatgen
    """
    pmg_struct = Structure.from_file(filename, primitive=primitive, sort=sort,
                                     merge_tol=merge_tol)
    structure = StructureData()
    structure.set_pymatgen(pmg_struct)
    print(structure)
    # Set structure description
    if isinstance(description, str):
        structure.description = description
    else:
        if os.path.basename(filename).lower() in ['poscar', 'contcar',
                                                  'vasprun.xml']:
            # basename is not sufficiently informative in this case,
            # so we also include the parent directory
            parent = os.path.basename(os.path.dirname(filename))
            origin = os.path.join(parent, os.path.basename(filename))
        else:
            origin = os.path.basename(filename)

        structure.description = '{} structure obtained from {}'.format(
            structure.get_formula(), origin)
        if store:
            structure.store()
        if verbosity >= 1:
            print(('{} with composition {} and description \'{}\' has been '
                   'loaded from file : {}').format(
                  structure, structure.get_formula(), structure.description,
                  os.path.abspath(filename)))
    return structure


def get_default_potential_mapping(elements=None, structure=None,
                                  return_dict=False, verbosity=0):
    """
    Construct a defaut potential mapping from a StructureData, pymatgen
    Structure or list of elements.

    The default mapping simply attaches the potential with the element name
    (e.g. Li/POTCAR in the considered potential family)

    Args:
        elements: list, tuple or None (default: None)
            list of elements used to define the default potential mapping if
            potential_mapping is None. The default potentials will correspond
            to the element name. One should be careful with elements requiring
            the use of specific potentials.
        structure: StructureData, pymatgen (I)Structure, (I)Molecule, or None
            (default is None). A reference structure whose composition will be
            used to establish the default computer mapping (see above).
        return_dict: bool (default is False)
            if True a python dict will be returned instead of an Aiida Dict.
        verbosity: int (default is 1)
            Verbosity level.

    Returns:
        A potential_mapping Dict (or python dict if return_dict is True) associating elment names with a potential name (e.g. {'Si': 'Si'} to pick the POTENTIAL_FAMILY/Si/POTCAR VASP potential).
    """
    if structure is None and elements is None:
        raise Exception('get_default_potential_mapping functions requires '
                        'a structure or elements input.')
    if isinstance(structure, StructureData):
        mapping = {kind_name: kind_name for kind_name in
                   structure.get_kind_names()}
    elif isinstance(structure, (Structure, IStructure, Molecule, IMolecule)):
        mapping = {specie.name: specie.name for specie in structure.species}
    elif isinstance(elements, (list, tuple)):
        mapping = {element: element for element in elements}
    if verbosity >= 1:
        print('Default potential mapping:')
        for k, v in mapping.items():
            print(f'{k}: {v}')
    if return_dict:
        return mapping
    else:
        return Dict(dict=mapping)


def get_default_potential_family(code=None, return_str=False, default_code='vasp',
                                 default_vasp_potential_family='PBE.54',
                                 verbosity=0):
    """
    Get a default potential family for a given code.

    Args:
        code: aiida.orm.Code, str or None (default is None)
            Code label or Code instance. If None 'vasp' and the default
            'PBE.54' familily will be used.
    """
    potential_family = None
    # TODO: define default family for other codes and update the following
    # accordingly using the 'vasp' model
    if code is None:
        if default_code.lower() == 'vasp':
            potential_family = default_vasp_potential_family
    elif isinstance(code, str):
        if 'vasp' in code.lower():
            potential_family = default_vasp_potential_family
    elif isinstance(code, Code):
        if 'vasp' in code.label.lower():
            potential_family = default_vasp_potential_family

    if verbosity >= 0:
        print('Default potential family: \'{}\''.format(potential_family))

    if not return_str:
        potential_family = Str(potential_family)

    return potential_family


def get_default_potential_parameters(elements=None, structure=None, code=None,
                                     verbosity=0):
    potential_parameters = {
        'potential_family': get_default_potential_family(code=code,
                                                         verbosity=0),
        'potential_mapping': get_default_potential_mapping(elements=elements,
            structure=structure, verbosity=0)
    }
    if verbosity >= 1:
        print('Fetching default potential parameters:')
        for k, v in potential_parameters.items():
            print(f'{k}: {v}')
    return potential_parameters


def get_default_vasp_code_and_computer(verbosity=0):
    """
    Get code from a default code-computer combination for aiida_vasp
    """
    # Update the following according to defined user code and computer
    code_label = 'vasp-5.4.4'
    computer_label = 'jean-zay'
    code = Code.get_from_string('{}@{}'.format(code_label, computer_label))
    if verbosity >= 1:
        print('Default code: {}@{}'.format(code.label,
                                           code.get_computer_label()))
    return code


def get_computer_options(computer_label=None, wall_time_s=None, nb_of_nodes=1,
                         account=None, test_mode=True, return_attributedict=False,
                         verbosity=1):
    """
    An ad hoc utility function that should be adapted to the specific
    configurations of the computers used by the user and their schedulers

    prepare a dict default calculation parameters
    build_and_submit_multistep_relax(structure: StructureData,
                                     multistep_input_param: Dict,
                                     code: Code,
                                     relax_potential_family,
                                     relax_potential_mapping,
                                     relax_options):
    Args:
        computer_label: str (default is 'jean-zay')
            computer suffix in code@computer listed by 'verdi code list'
        wall_time_s: int (default is None)
            The default will be adapted based on the scheduler configurations
            of known computers.
        nb_of_nodes: int (default is 1)
            Number of requested nodes. The number of CPUs should be calculated
            accordingly based on the computer configuration (number of CPUs per
            node).
        test_mode: bool (default=True)
            Default parameters will be chosen such as to allow rapid execution
            of a job according to the computer's scheduler configuration.
        return_attributedict: bool (default is False)
            If True an aiida.common.extendeddicts.AttributeDict will be
            returned instead of an aiida.orm.Dict
        verbosity: int (default is 1)
            Verbosity level.

    Returns:
        An AttributeDict of computer options
    """
    if computer_label is None:
        computer_label = get_default_vasp_code_and_computer().get_computer_label()

    options = AttributeDict()

    # The section above must be edited according to the configuration of the
    # computers typically used by the user to minimize the amount of
    # information that should be provided
    if computer_label.lower() == 'jean-zay':
        web_site = 'http://www.idris.fr/eng/jean-zay/cpu/jean-zay-cpu-exec_partition_slurm-eng.html'

        options.resources = {'num_machines': nb_of_nodes,
                             'num_mpiprocs_per_machine': 40}
        if account is None:
            options.account = 'zqm@cpu'
        else:
            options.account = account
        if wall_time_s is None:
            options.max_wallclock_seconds = 7200  # 02:00:00
        else:
            options.max_wallclock_seconds = wall_time_s

        if test_mode and options.max_wallclock_seconds <= 7200:
            options.qos = 'qos_cpu-dev'
        elif options.max_wallclock_seconds > 72000 and nb_of_nodes <= 4:
            options.qos = 'qos_cpu-t4'
        elif options.max_wallclock_seconds <= 72000:
            options.qos = 'qos_cpu-t3'
        else:
            options.max_wallclock_seconds = 72000
            options.qos = 'qos_cpu-t3'
            print(('WARNING: inappropriate resource request for computer {}. '
                   'qos set to {} with a wall time of {} s. See Jean-Zay '
                   'webpage {} for QoS and associated limitations.').format(
                  computer_label, options.qos, options.max_wallclock_seconds,
                  web_site))
        options.queue_name = ''

    elif computer_label.lower() == 'curta':
        web_site = 'https://redmine.mcia.fr/projects/cluster-curta/wiki/Configuration_Scheduler'
        default_wall_time_s = 24*3600
        options.account = ''
        options.qos = ''
        options.resources = {'num_machines': nb_of_nodes,
                             'num_mpiprocs_per_machine': 32}
        if test_mode and nb_of_nodes == 1 and (wall_time_s is None or
                                               wall_time_s <= 12*3600):
            options.queue_name = 'visu'
            if wall_time_s is None:
                options.max_wallclock_seconds = 12*3600
            else:
                options.max_wallclock_seconds = wall_time_s
        elif wall_time_s is None:
            options.queue_name = 'ircer'
            options.max_wallclock_seconds = default_wall_time_s
        elif wall_time_s > 120*3600:
            options.queue_name = 'longq'
            if wall_time_s > 720*3600:
                options.max_wallclock_seconds = 720*3600
                print(('WARNING: inappropriate resource request for computer '
                       '{}. Resquested wall time ({} s) excesses limit of '
                       'queue \'longq\'. wall time set to {} s ({:.2f} h).  '
                       'See {} web site for information.').format(
                           computer_label, wall_time_s,
                           options.max_wallclock_seconds,
                           options.max_wallclock_seconds/3600, web_site))
            else:
                options.max_wallclock_seconds = wall_time_s
        else:  # wall_time_s <= 120*3600
            options.queue_name = 'ircer'
            options.max_wallclock_seconds = wall_time_s

    if verbosity >= 2:
        for k, v in options.items():
            print('{}: {}'.format(k, v))

    if not return_attributedict:
        options =  Dict(dict=options)

    return options

def get_default_incar_parameters(structure=None, elements=None,
                                 return_dict=False, return_attributedict=False,
                                 verbosity=0):
    """
    Read a default set of incar parameters

    MIND: These parameters do not include parameters related to the ionic
    relaxation settings, which are defined in the high-level relax parameters

    Args:
        structure: StructureData or None
            Used to determine whether elements requiring higher cutoff energies
            (C, N, O, F, He, Ne) are present, in which case the default should
            be high.
        elements: list, tuple or None
            Sets the encut value based on the highest enmax of the listed
            elements (see above).
        return_dict: bool (default is False)
            If True the function will return a python dict instead of an
            aiida.orm.Dict (unless return_attributedict is True)
        return_attributedict: bool (default is False)
            If True the function will return an aiida.common.extendeddicts.AttributeDict
            instead of an aiida.orm.Dict.


    Returns:
        A Dict to be direcly used as inputs.parameters for the vasp.relax
        workchain
    """
    cutoff_energy = 1.3 * 220  # > 1.3*ENMAX for all elments bu C, N, O, F, He,
                               # Ne using standard potentials
    if isinstance(structure, StructureData):
        elements = structure.get_kind_names()
    if isinstance(elements, (list, tuple)):
        elmt_set = set(elements)
        if (elmt_set & set(['C', 'N', 'O', 'F'])):
            cutoff_energy = 1.3 * 300
        elif 'He' in elmt_set:
            cutoff_energy = 1.3 * 360
        elif 'Ne' in elmt_set:
            cutoff_energy = 1.3 * 260
        elif 'B' in elmt_set:
            cutoff_energy = 1.3 * 240

    incar_parameters = {
        'incar': {
            'encut': cutoff_energy,  # > 1.3*ENMAX for all elments using std pot
            'ediff': 1e-6,
            'ncore': 8,
            'ismear': 0,
            'sigma': 0.05,
            'prec': 'High',
            'lreal': 'Auto',
        }
    }
    if verbosity >= 1:
        print('Reading default incar_parameters:')
        for k, v in incar_parameters.items():
            print('{}: {}'.format(k, v))
    if not return_dict and not return_attributedict:
        incar_parameters = DataFactory('dict')(dict=incar_parameters)
    elif return_attributedict:
        incar_parameters = AttributeDict(incar_parameters)

    return incar_parameters


def get_default_relax_parameters(fixed_cell=False, energy_cutoff=None,
                                 force_cutoff=None, verbosity=0):
    """
    Define a default set of relax parameters

    Change specific parameters as follows:
    relax_parameters = get_default_relax_parameters()
    relax_parameters.energy_cutoff.value = 0.001  # It is an aiida Float rather
                                                  # than a python float.

    Args:
        fix_cell: bool (default is False)
            Whether cell (volume and shape should be fixed).
        force_cutoff: float, bool or None (default is None)
            Whether to use the default (True) or a specified force-cutoff as
            the ionic convergence criterion
        energy_cutoff: float, bool or None (default is None)
            Whether to use the default (True) or a specified energy-cutoff as
            the ionic convergence criterion
        Verbosity: int (default is 0)
            Verbosity level

    Returns:
        An AttributeDict to be used as inputs.relax fir the vasp.relax
        workchain.
    """
    # Adjust as you wish
    default_force_cutoff = 0.01  # in eV/Angstroms
    default_energy_cutoff = 0.001  # in eV - TODO: calculate in eV by atom ?

    relax_parameters = AttributeDict()
    relax_parameters.perform = DataFactory('bool')(True)
    relax_parameters.algo = DataFactory('str')('cg')
    # Use user-defined
    if isinstance(force_cutoff, float):
        relax_parameters.force_cutoff = DataFactory('float')(force_cutoff)
    elif force_cutoff == True:  # force_cutoff is a boolean
        relax_parameters.force_cutoff = DataFactory('float')(
            default_force_cutoff)
    elif isinstance(energy_cutoff, float):
        relax_parameters.energy_cutoff = DataFactory('float')(energy_cutoff)
    elif energy_cutoff == True:
        relax_parameters.energy_cutoff = DataFactory('float')(
            default_energy_cutoff)

    else:
        relax_parameters.force_cutoff = DataFactory('float')(0.01)
    relax_parameters.positions = DataFactory('bool')(True)
    relax_parameters.shape = DataFactory('bool')(not fixed_cell)
    relax_parameters.volume = DataFactory('bool')(not fixed_cell)
    relax_parameters.steps = DataFactory('int')(200)

    if verbosity >= 1:
        print('Reading default relax_parameters:')
        for k, v in relax_parameters.items():
            print('{}: {}'.format(k, v))

    return relax_parameters


def print_process_inputs(inputs):
    """
    Print process inputs in a readable manner
    """
    for k, v in inputs.items():
        if isinstance(v, Dict):
            print('{}: {}'.format(k, v.get_dict()))
        elif isinstance(v, (Str, Int, Float, Dict, Bool)):
            print('{}: {}'.format(k, v.value))
        elif isinstance(v, KpointsData):
            print('{}: {}'.format(k, v.get_kpoints_mesh()))
        else:
            print('{}: {}'.format(k, v))


def build_and_submit_relax(structure: StructureData,
                           relax_code=None,
                           relax_potential_family=None,
                           relax_potential_mapping=None,
                           relax_options=None,
                           wall_time_s=None,
                           nb_of_nodes=1,
                           incar_parameters=None,
                           relax_parameters=None,
                           unsupported_parameters=None,
                           kpoints_spacing=0.4,
                           parser_settings=None,
                           clean_workdir=True,
                           test_mode=False,
                           dry_run=False,
                           verbosity=1):
    """
    Build vasp.relax workchain as automatically as possible and submit

    WARNING: make sure that function get_computer_options() is completed
    according to your computers and schedulers configurations.

    Args:
        structure: StructureData
            Input structure to relax
        relax_code: str, Code or None (default is None)
            Default will be loaded with get_default_vasp_code_and_computer()
        relax_potential_family: str or None (default is None)
            potential_family as listed by 'verdi data vasp-potcar listfamilies'.
            Default will be set according to relax_code and
            function get_default_potential_family().
        relax_potential_mapping: dict or None (default is None)
            dict associating elment names in the considered system with a
            potential name. If None the default mapping will be obtained from
            the structure.
        relax_options: dict, AttributeDict or None
            If None options will be defined automatically using the
            relax_code.get_computer_label() and function get_computer_options()
        wall_time_s: int, float or None (default is None)
            Calculation walltime in seconds. if not specified it will be set
            automatically according to the default queue, qos, etc., defined in
            get_computer_options().
        nb_of_nodes: int (default is 1)
            Number of calculation nodes. The number of cpus will be calculated
            automatically with get_computer_options().
        kpoints_spacing: float (default is 0.4)
            maximum spacing between k-points in the reciprocal space in Ang^-1
            used to automatically calculate a grid of k-points.
        unsupported_parameters: dict or None (default is None)
            dict listing VASP parameters not yet supported by the
            aiida_vasp ParametersMassager (see documentation here:
            https://aiida-vasp.readthedocs.io/en/latest/concepts/parameters.html)
            Here is the content of this documentation:
            In case you for instance perform developments in the VASP code,
            sometimes it makes sense to add a new VASP tag. This can be supplied
            in settings.inputs.unsupported_parameters as dict with the following
            specifications:
            unsupported_parameters = {
                'my_unsupported_parameter': {
                'default': 1.0,
                'description': 'Some description',
                'type': float,
                'values': [1.0, 2.0]
            }
        test_mode: bool (default is False)
            Will be used to pick relax_options ensuring a fast execution of
            calculations (with the risk of hitting the walltime) for test
            purposes.
        dry_run: bool (default is False)
            Build process but do not submit. Function will return None in this
            case.
        verbosity: int (default is 1)
            Verbosity level

    Returns:
        WorkChainNode if dry_run is False, None otherwise.
    """
    if relax_potential_family is None:
        relax_potential_family = get_default_potential_family(relax_code)
    if relax_potential_mapping is None:
        relax_potential_mapping = get_default_potential_mapping(
            structure=structure, return_dict=True,
            verbosity=verbosity)
    if isinstance(relax_code, str):
        relax_code = Code.get_from_string(relax_code)
    if relax_code is None:
        # TODO: adapt for code(s) other than VASP ?
        relax_code = get_default_vasp_code_and_computer(verbosity=verbosity)
    if isinstance(relax_options, dict):  # convert dict to AttributeDict
        relax_options = AttributeDict(relax_options)
    if relax_options is None:
        relax_options = get_computer_options(
            computer_label=relax_code.get_computer_label(),
            wall_time_s=wall_time_s, nb_of_nodes=nb_of_nodes,
            test_mode=test_mode, return_attributedict=True,
            verbosity=verbosity)

    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    kpoints_data = KpointsData()

    # TODO: define k_point_mesh from density
    kpoints_data.set_cell_from_structure(structure)
    kpoints_data.set_kpoints_mesh_from_density(kpoints_spacing)

    # Set the workchain to call. No () : we use the class rather than instance
    workchain = WorkflowFactory('vasp.relax')

    # Declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()
    # organize settings
    if isinstance(parser_settings, (dict, AttributeDict)):
        settings.parser_settings = parser_settings
    else:
        settings.parser_settings = {'output_params': ['total_energies',
                                                  'maximum_force'],
                                    'add_maximum_stress': False}
    # Add unsupported parameters to settings if any (setting a default value is sufficient):
    if isinstance(unsupported_parameters, dict):
        settings['unsupported_parameters'] = unsupported_parameters

    # set inputs for the following WorkChain execution
    inputs.code = relax_code
    # set structure
    inputs.structure = structure

    inputs.kpoints = kpoints_data

    # set parameters
    if isinstance(incar_parameters, (dict, AttributeDict)):
        inputs.parameters = DataFactory('dict')(dict=incar_parameters)
    else:
        inputs.parameters = get_default_incar_parameters(structure=structure,
                                                         verbosity=verbosity)

    if relax_parameters is None:
        inputs.relax = get_default_relax_parameters(verbosity=verbosity)
    elif isinstance(relax_parameters, dict):
        inputs.relax = AttributeDict(relax_parameters)
    elif isinstance(relax_parameters, AttributeDict):
        inputs.relax = relax_parameters

    # set potentials and their mapping
    inputs.potential_family = Str(relax_potential_family)
    inputs.potential_mapping = dict_data(dict=relax_potential_mapping)
    # set options
    inputs.options = dict_data(dict=relax_options)
    # set settings
    inputs.settings = dict_data(dict=settings)

    # set workchain related inputs
    inputs.verbose = Bool(True)  # For testing purposes

    if isinstance (clean_workdir, bool):
        inputs.clean_workdir = Bool(clean_workdir)
    elif isinstance (clean_workdir, Bool):
        inputs.clean_workdir = clean_workdir

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.

    if verbosity >= 2 or dry_run:
        print(('\n' + 100*'*' +
              '\nReady to submit {} workchain with inputs:\n'
              + 100*'*').format(workchain.__name__))
        print_process_inputs(inputs)
        print(100*'*')

    # DEBUGGING
    print('raw inputs:')
    for k, v in inputs.items():
        if isinstance(v, AttributeDict):
            print(f'{k}:')
            for k_k, k_v in v.items():
                print(f'    {k_k}: {k_v} (of type {type(k_v)})')
        else:
            print(f'{k}: {v} (of type {type(v)})')

    if dry_run:
        return None

    if test_mode:
        relax_node = run_get_node(workchain, **inputs)
    else:
        relax_node = submit(workchain, **inputs)
    return relax_node


def get_structure_from_pk_or_node(struct_or_pk):
    """
    """
    if isinstance(struct_or_pk, StructureData):
        return struct_or_pk
    elif isinstance(struct_or_pk, int):
        struct = load_node(struct_or_pk)
    elif isinstance(struct_or_pk, str):
        struct = load_node(uuid=struct_or_pk)
    else:
        sys.exit('Argument should be a StructureData, int or str.')
    if not isinstance(struct,StructureData):
        sys.exit('No stucture with specified PK or UUID in database.')
    return struct


def get_list_of_structures(structures_or_pks):
    """
    get a list of StructureData from a single or list of structure(s) or PK(s)

    Args:
        structures_or_pks: Int, str, StructureData or a list thereof
        May be a sinle PK, UIID, StructureData or a list thereof

    Returns:
        structures: list of StructureData
    """
    if isinstance(structures_or_pks, (int, str, StructureData)):
        structures_or_pks = [structures_or_pks]
    structures = [get_structure_from_pk_or_node(struct)
                  for struct in structures_or_pks]
    return structures


def get_node_from_id_or_node(node_or_id, of_types=None):
    """
    Get a node from a PK, UUID or node

    Args:
        node_or_id: int, str or node
            Node or node id (PK or UUID) to be converted to a node
        of_types: list, tuple or type name (dafault is None).
            Only allow the listeed type(s) of aiida nodes
            The function will return a TypeError if one or more
            nodes are not in of_types

    Returns:
        An aiida node
    """
    if issubclass(node_or_id.__class__, Node):
        node = node_or_id
    elif isinstance(node_or_id, (int, str)):
        node = load_node(node_or_id)
    else:
        sys.exit('Argument should be an aiida.orm.Node, int or str.')
    if of_types is not None:
        if not isinstance(node, of_types):
            raise TypeError('node {} is not of type {}'.format(node,
                                                               of_types))
    return node


def get_nodes_from_ids_or_nodes(nodes_or_ids, convert_to_single=False,
                                of_types=None):
    """
    Get a list of nodes from a single or a list of nodes, uuids or pks

    Will return an error if a node cannot be found

    Args:
        nodes_or_ids: int, str, aiida.orm.nodes.Node, list, tuple
            Node(s) or node id(s) to be converted to a list of nodes
        convert_to_single: bool (default is False)
            if true a list of one node will be converted to
            a node.
        of_types: list, tuple or type name (dafault is None).
            Only allow the listeed type(s) of aiida nodes
            The function will return a TypeError if one or more
            nodes are not in of_types

    Returns:
        A list of nodes or a node (if convert_to_single is True)
    """
    if issubclass(nodes_or_ids.__class__, (Node, int, str)):
        nodes_or_ids = [nodes_or_ids]
    if of_types is None:
        nodes = [get_node_from_id_or_node(node_or_id) for node_or_id in nodes_or_ids]
    else:
        nodes = []
        for node_or_id in nodes_or_ids:
            node = get_node_from_id_or_node(node_or_id)
            if not isinstance(node, of_types):
                raise TypeError('node_or_id {} is not of type {}'.format(node,
                                                                         of_types))
            nodes.append(node)
    if convert_to_single and len(nodes) == 1:
        [nodes] = nodes
    return nodes


def check_data_names(data_names, structures):
    """
    Check data_names length and type and return a copy to avoide in-place modif
    """
    # Check data_names
    if isinstance(data_names, (list, tuple)):
        if len(data_names) != len(structures):
            print('WARNING: data_names and structures have different length.')
            print('Structure PKs = ',[struct.pk for struct in structures])
            print('data_names = ', data_names)
            _data_names = None
        else:
            _data_names = data_names.copy()
    elif isinstance(data_names, str):
        _data_names = [data_names]
    elif data_names is not None:
        print('WARNING: data_names should be a list or tuple. Using the default.')
        _data_names = None
    if data_names is None:
        _data_names = [f'structure {struct.pk}' for struct in structures]

    return _data_names


def get_total_x_ray_scattering(structure_or_pk, plot_pdf=True,
                               plot_title=None, structure_description=None,
                               experimental_file=None,
                               experimental_title=None,
                               expt_rel_y_shift=0.0,
                               show_experiment=False,
                               show_difference=False,
                               diff_rel_y_shift=-0.25,
                               **kwargs):
    """
    Calculate and possibly plot total X-ray scattering data from a structure

    Experimental data may be added to the data in the form
    of a 2-column space/tab serated file without header.
    To use more elaborate files one must use the nanopdf
    module directly.

    Args:
        plot_pdf: bool (fdefault is True)
            Whether plot shoult be shown and fig and ax returned in addition
            to r and the exact pdf numpy.darrays.
        plot_title: str or None (default is None)
            Title of the plot
        experimental_file: str or None (default is None)
            File containing the experimental data in a simple
            two-comumn format with r in Angstroms (first column)
            and the pdf intensities (second column). No header.
        experimental_title: str or None (default is None)
            Title associated with experimental data (used in legend)
        show_experiment: bool or None
            Will be be automatically set to True if experimental_file
            is a valid file.
        expt_rel_y_shift: float (default is 0.0)
            Shift of experimental data relative to simulation
            in fraction of the simulated data amplitude.
        show_difference: bool or None
            Whether a difference between experiment and simulation
            should be calculated and drawn.

        kwargs:
            may include all arguments used by nanopdfData, among
            which are R_max, sigma, lambda1, verbosity
    """
    structure = get_structure_from_pk_or_node(structure_or_pk)
    pmg_struct = structure.get_pymatgen()
    nd = nanopdfData(**kwargs)
    nd.calculate_final_exact_pdf_from_structure(pmg_struct)

    # TODO : allow setting title
    if plot_pdf:
        if not isinstance(structure_description, str):
            if len(structure.description) > 0:
                structure_description = '{} <{}>: simulated PDF'.format(structure.description,
                                                         structure.pk)
            else:
                structure_description = '{} <{}>: simulated PDF'.format(structure.get_formula(),
                    structure.pk)

        fig, ax = nd.plot_exact_pdf(title=plot_title, data_name=structure_description,
                                    experimental_file=experimental_file,
                                    experimental_title=experimental_title,
                                    show_experiment=show_experiment,
                                    expt_rel_y_shift=expt_rel_y_shift,
                                    show_difference=show_difference,
                                    diff_rel_y_shift=diff_rel_y_shift)
        return nd.R, nd.exactPDF, fig, ax
    else:
        return nd.R, nd.exactPDF


def plot_multiple_total_x_ray_scattering(structures_or_pks, data_names=None,
        plot_title='Exact total X-ray scattering', experimental_file=None,
        experimental_title=None, y_shift_factor=1.1, **kwargs):
    """
    Plot total X-ray scattering for a series of structures (or PKs)
    """
    structures = get_list_of_structures(structures_or_pks)

    legend = check_data_names(data_names, structures)

    fig, ax = plt.subplots()
    y_shift = 0
    r_max = 0
    for index, structure in enumerate(structures):
        r, pdf = get_total_x_ray_scattering(structure, plot_pdf=False,
            plot_title=None, experimental_file=experimental_file,
            experimental_title=experimental_title, **kwargs)
        ax.plot(r, pdf + y_shift)
        if y_shift == 0:
            y_range = max(pdf)-min(pdf)
            y_shift = y_shift_factor * y_range
        else:
            y_shift += y_shift_factor * y_range
        if max(r) > r_max:
            r_max = max(r)

    if experimental_file is not None:
        # Load experimental ND pattern
        expt_data = np.loadtxt(experimental_file, delimiter=None, skiprows=0)
        # Adding experimental ND data to plot
        ax.plot(expt_data[:,0], expt_data[:,1] + y_shift,
                color='black')
        if experimental_title is None:
            experimental_title = 'Experiment'
        legend += [experimental_title]
    ax.set(xlim=[0, max(r)])
    if isinstance(plot_title, str):
        ax.set_title(plot_title)
    ax.set_xlabel('r (\u212B)')
    ax.set_ylabel('Intensity (AU)')
    ax.legend(legend,loc='upper left', bbox_to_anchor= (1.0, 1.0))
    fig.tight_layout()
    # plt.show()

    return fig, ax


def plot_distance_matrix(struct_or_ids, r_max=10.0, sigma=0.02,
                         system_name=None, structure_names=[]):
    """
    Plot a matrix of distances between structures with
    pyama.structureComparisonsPkg.distancTools.distanceMatrixData
    Distances are numbers between 0 (identical structures) and 1
    and are based on partial radial distribution functions (RDFs)

    Args:
        struct_or_ids: list or tuple
            List of PKs, UUIDs or StructureData instances.
        r_max: float (default is 10.0)
            Maximum radius for the calculation of partial RDFs
        sigma: float (default is 0.02)
            Gaussian smearing factor in Angstrom for RDFs calculations
            Avoids numerical problems.
        system_name: str (default is None)
            Will be used in the figure title
        structure_names: list ir tuple (default is [])
            Names of structures
    """
    structures = get_list_of_structures(struct_or_ids)
    dmd = distanceMatrixData(Rmax=r_max, sigma=sigma)
    dmd.calculate_distance_matrix([struct.get_pymatgen() for struct in structures],
                                  structureIDs=[struct.pk for struct in structures],
                                  systemName=system_name)
    if len(structure_names) not in [0, len(structures)]:
        structure_names = []
    fig, ax = dmd.plot_distance_matrix(tickLabels=structure_names)
    return fig, ax

def add_smoothing_functions(pattern, fwhm=0.2, eta=0.5,
                            two_theta_range=[0, 120],
                            step=None):
    """
    Add pseudo-Voigt broadening to a diffraction pattern (pymatgen)

    Args:
        pattern:
            pymatgen diffraction pattern as obtained with XRDCalculator
            or NDCalculator
        fwhm: float (default=0.2)
            Full width at half maximum in units of 2*theta (°)
        eta: float (default=0.5)
            Lorentzian to Gaussian ratio (1 for pure Lorentzian, 0 for
            pure Gaussian profile)
        two_theta_range:
        step: float or None
            two_theta step for the x axis. Default is 0.1*fwhm

    Returns:
        (x, y): tuple
            x:  two_theta axis between two_theta_range[0] and
                two_theta_range[1] with step.
            y:  simulated diffracted intensities at x.
    """
    H = fwhm  # fwhm in 2(theta) units
    if eta < 0 or eta > 1:
        print('WARNING : eta should be between 0 and 1. Using the default (0.5).')
        eta = 0.5
    if step==None:
        step = 0.1*H

    x = np.arange(two_theta_range[0], two_theta_range[1], step)
    x_hkl = pattern.x
    I_hkl = pattern.y
    # Initialize y
    y = np.zeros(len(x))
    for i in range(len(x_hkl)):
        # Gaussian broadening
        G = (2/H)*np.sqrt(np.log(2)/np.pi) * np.exp(-(4*np.log(2)/(H*H)) * np.square(x - x_hkl[i]))
        # Gaussian broadening
        L = 2/(np.pi * H) / (1 + 4/(H*H) * np.square(x - x_hkl[i]))
        y += I_hkl[i] * (eta * L + (1 - eta) * G)
    # Normalize:
    y = y/max(y)

    return x, y

def plot_diffraction_pattern(structure_or_pk, data_name=None,
                             diffraction_type='X-ray',
                             wavelength=None,
                             two_theta_range=(0, 80),
                             fwhm=0.2, eta=0.5,
                             experimental_file=None,
                             experimental_title=None,
                             push_legend=True):
    """
    Plot neutron or X-ray diffraction pattern for an input StructureData instance
    """
    structure=get_structure_from_pk_or_node(structure_or_pk)

    # Plot simulated neutron or x-raydiffraction pattern with Pymatgen
    pmg_struct = structure.get_pymatgen()
    if diffraction_type.lower() in ['xrd', 'x-ray', 'x-rays']:
        diffraction_type_name = 'X-ray diffraction'
        if wavelength is None:
            wavelength = 'CuKa'
        calculator=XRDCalculator(wavelength=wavelength, symprec=0,
                                 debye_waller_factors=None)
    elif diffraction_type.lower() in ['nd', 'neutron', 'neutrons']:
        diffraction_type_name = 'neutron diffraction'
        if wavelength is None:
            wavelength = 2.522
        calculator=NDCalculator(wavelength=wavelength, symprec=0,
                                debye_waller_factors=None)
    else:
        sys.exit('diffraction_type should be neutron/ND or X-ray/XRD.')

    pattern=calculator.get_pattern(pmg_struct, scaled=True,
                                   two_theta_range=two_theta_range)

    x, y = add_smoothing_functions(pattern, fwhm=fwhm,
        eta=eta, two_theta_range=two_theta_range)
    fig, ax = plt.subplots()
    # ax.plot(pattern.x, pattern.y, color='blue')

    ax.plot(x, y, color='blue')
    if isinstance(data_name, str):
        legend = [data_name]
    else:
        legend = ['Structure {}'.format(structure.pk)]

    if experimental_file is not None:
        # Load experimental ND pattern
        expt_data = np.loadtxt(experimental_file,delimiter=None,skiprows=0)
        # Adding experimental ND data to plot
        ax.plot(expt_data[:,0], expt_data[:,1]/max(expt_data[:,1])+
                 1.1*max(y),color='black')
        if experimental_title is None:
            legend.append('Experiment')
        else:
            legend.append(experimental_title)

    ax.set_title(('Simulated {} pattern for structure {}').format(
        diffraction_type_name, structure.pk))
    ax.set(xlabel='2\u03B8 (°)',ylabel='Normalized Intensity (AU)',
           xlim=two_theta_range)
    if push_legend:
        ax.legend(legend,loc='upper left', bbox_to_anchor= (1.0, 1.0))
        plt.tight_layout()
    else:
        ax.legend(legend)
    plt.show()

    return fig, ax

def copy_vasp_files_to_dir(process_or_ids, root_dir=None, verbosity=1):
    """
    Copy retrieved or remote input/output VASP files associated with process(es) to a directory

    Directory name will be <PK>_exit-code-<EXIT_STATUS>
    """
    processes = get_nodes_from_ids_or_nodes(process_or_ids)

    # TODO: offer the possibility to chose the files to copy
    files_to_copy = ['INCAR', 'POSCAR', 'KPOINTS', 'vasp_output', 'OUTCAR', 'OSZICAR', 'vasprun.xml']

    if root_dir is None:
        root_dir = os.getcwd()

    for calc in processes:
        # TODO: test whether nodes are processes
        if calc.process_label != 'VaspCalculation':
            print('Node {} is not a VaspCalculation process.'.format(calc.pk))
            continue
        dir_name = os.path.join(root_dir, '{}_exit-code-{}'.format(calc.pk, calc.exit_status))
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        for file_name in files_to_copy:
            file_origin = None
            abs_file_name = os.path.abspath(os.path.join(dir_name, file_name))
            try:
                file_content = calc.outputs.retrieved.get_object_content(file_name)
                with open(abs_file_name, 'w') as f:
                    f.write(file_content)
                file_origin = calc.outputs.retrieved
                if verbosity >= 1:
                    print('{}<{}>: File {} copied from retrieved output {}.'.format(
                        calc.process_label, calc.pk, abs_file_name, file_origin.pk))
            except Exception as e:
                if verbosity >= 2:
                    print('{}<{}>: File {} not found in retrieved output {}: {}'.format(
                        calc.process_label, calc.pk, file_name, calc.outputs.retrieved.pk, e))
                try:
                    calc.outputs.remote_folder.getfile(file_name, abs_file_name)
                    file_origin = calc.outputs.remote_folder
                    if verbosity >= 1:
                        print('{}<{}>: File {} copied from remote_folder output {}.'.format(
                            calc.process_label, calc.pk, abs_file_name, file_origin.pk))
                except Exception as e:
                    if verbosity >= 1:
                        print(('{}<{}>: file {} could not be copied from retrieved or remote_folder '
                           'outputs: {}.').format(calc.process_label, calc.pk, file_name, e))

def print_subprocesses_status(proc_or_id, separate_by=[]):
    """
    print exit_status (exit_message) or process_state for all subprocesses of a given process

    Args:
        proc_or_id: ProcessNode, Int or Str
            Process node, PK or UUID
        separate_by: list or tuple (default is [])
            print a line of '-' as soon as descendants with
            a certain process_label are detected
    """
    parent_process = get_node_from_id_or_node(proc_or_id)
    # convert separate_by to a list if necessary
    if not isinstance(separate_by, (list, tuple)):
        separate_by = [separate_by]
    for proc in parent_process.called_descendants:
        if proc.process_label in separate_by:
            print(100*'-')
        if proc.process_state.value == 'finished':
            print('{}<{}>: {} ({})'.format(proc.process_label, proc.pk, proc.exit_status, proc.exit_message))
        else:
            print('{}<{}>: {} ({})'.format(proc.process_label, proc.pk, proc.process_state.name))

def import_structures_and_store_in_group(structure_files, group_label,
                                         structure_names=None,
                                         print_file_abs_path=True, verbosity=1):
    """
    TO BE COMPLETED
    """
    try:
        group = load_group(group_label)
    except NotExistent as e:
        if verbosity >= 2:
            print(e)
        group = Group(label=group_label)
        group.store()
        if verbosity >= 1:
            print('Group {}<{}> has been created.'.format(group.label,
                                                          group.pk))
    # Convert single file to list
    if isinstance(structure_files, str):
        structure_files = [structure_files]

    orig_group_length = len(group.nodes)
    imported_structures = []
    for struct_index, struct_file in enumerate(structure_files):
        # TODO: Test whether an equivalent structure already exists in group

        struct = StructureData()
        struct.set_pymatgen_structure(Structure.from_file(struct_file))
        if structure_names is None:
            struct_name = f'{struct.get_formula()}'
        else:
            struct_name = structure_names[struct_index]

        if print_file_abs_path:
            file_name = os.path.abspath(struct_file)
        elif struct_file.lower() in ['poscar', 'contcar']:
            file_name = os.path.basename(os.path.split(os.path.abspath(struct_file)))
        else:
            file_name = os.path.basename(struct_file)
        struct.label = '{} structure imported from file {}'.format(struct_name,
                                                                   file_name)
        imported_structures.append(struct)
        struct.store()
        if verbosity >= 2:
            print(('Structure {} of composition {} and imported '
                   'from file {} with label \'{}\'.').format(struct.pk,
                   struct.get_formula(), os.path.abspath(struct_file),
                   struct.label))
        group.add_nodes(struct)

    if verbosity >= 1:
        print('{} structures added to group {}<{}>.'.format(
            len(group.nodes) - orig_group_length, group.label, group.pk))

    return imported_structures


def load_reference_systems(labels=None, pattern=None,
                           ref_struct_group_label='reference_structures',
                           convert_to_single=True, verbosity=1):
    if isinstance(labels, str):
        label_filter = labels
    elif isinstance(labels, (list, tuple)):
        label_filter = {'in': labels}
    if isinstance(pattern, str):
        label_filter = {'like': '%{}%'.format(pattern)}

    qb = QueryBuilder()
    qb.append(Group, filters={'label': ref_struct_group_label}, tag='ref_struct_group')
    qb.append(StructureData, with_group='ref_struct_group',
              filters={'label': label_filter}, tag='structures')
    ref_systems = [row[0] for row in qb.all()]
    if verbosity >= 1:
        for ref_system in ref_systems:
            print('Reference system {} with label {} loaded from group {}'.format(
                ref_system.pk, ref_system.label, ref_struct_group_label))
    if convert_to_single and len(ref_systems) == 1:
        [ref_systems] = ref_systems
    return ref_systems


def print_converge_parameters(converge_process_node):
    for k in converge_process_node.inputs.converge.keys():
        try:
            print('{}: {}'.format(k,
                converge_process_node.inputs.converge[k].value))
        except AttributeError:
            try:
                print('{}: {}'.format(k,
                    converge_process_node.inputs.converge[k].get_array('array')))
            except Exception as e:
                print('{}: {} - {}'.format(k, type(e), e))
        except Exception as e:
            print('{}: {} - {}'.format(k, type(e), e))


def set_potentials_family_and_mapping_input(inputs,
        potential_family=None, potential_mapping=None,
        verbosity=0):
    # set potentials and their mapping
    if potential_family is None:
        inputs.potential_family = get_default_potential_family(
            inputs.code, verbosity=verbosity)
    elif isinstance(potential_family, str):
        inputs.potential_family = Str(potential_family)
    elif isinstance(potential_family, Str):
        inputs.potential_family = potential_family
    else:
        raise TypeError('potential_family should be an str, Str or None.')

    if potential_mapping is None:
        inputs.potential_mapping = get_default_potential_mapping(
            structure=inputs.structure, verbosity=verbosity)
    elif isinstance(potential_mapping, (dict, AttributeDict)):
        inputs.potential_mapping = Dict(dict=potential_mapping)
    elif isinstance(potential_mapping, Dict):
        inputs.potential_mapping = potential_mapping
    else:
        raise TypeError('potential_mapping should be a dict, Dict or None.')
    return inputs


def build_and_submit_dos(structure: StructureData,
                         code=None,
                         potential_family=None,
                         potential_mapping=None,
                         options=None,
                         wall_time_s=None,
                         nb_of_nodes=1,
                         incar_parameters=None,
                         unsupported_parameters=None,
                         kpoints_spacing=0.08,
                         nb_of_energy_points=1201,
                         parser_settings=None,
                         clean_workdir=True,
                         test_mode=True,
                         dry_run=False,
                         process_description=None,
                         verbosity=1):
    """
    Build a dos calculation with the vasp.verify workchain as automatically
    as possible and submit

    WARNING: make sure that function get_computer_options() is completed
    according to your computers and schedulers configurations.

    Default potential mapping and incar parameters may be modified in
    functions:
        - get_default_incar_parameters,
        - get_default_potential_mapping,
        - get_default_potential_family

    Args:
        structure: StructureData
            Input structure to relax
        code: str, Code or None (default is None)
            Default will be loaded with get_default_vasp_code_and_computer()
        potential_family: str or None (default is None)
            potential_family as listed by 'verdi data vasp-potcar listfamilies'.
            Default will be set according to code and
            function get_default_potential_family().
        potential_mapping: dict or None (default is None)
            dict associating elment names in the considered system with a
            potential name. If None the default mapping will be obtained from
            the structure.
        options: dict, AttributeDict or None
            If None options will be defined automatically using the
            code.get_computer_label() and function get_computer_options()
        wall_time_s: int, float or None (default is None)
            Calculation walltime in seconds. if not specified it will be set
            automatically according to the default queue, qos, etc., defined in
            get_computer_options().
        nb_of_nodes: int (default is 1)
            Number of calculation nodes. The number of cpus will be calculated
            automatically with get_computer_options().
        kpoints_spacing: float (default is 0.05)
            maximum spacing between k-points in the reciprocal space in Ang^-1
            used to automatically calculate a grid of k-points.
            WARNING: this parameter is 2*pi smaller than the KSPACING
            used in VASP INCAR files. VASP KSPACING default of 0.5 corresponds
            to kpoints_spacing ~ 0.08 Ang^1.
        unsupported_parameters: dict or None (default is None)
            dict listing VASP parameters not yet supported by the
            aiida_vasp ParametersMassager (see documentation here:
            https://aiida-vasp.readthedocs.io/en/latest/concepts/parameters.html)
            Here is the content of this documentation:
            In case you for instance perform developments in the VASP code,
            sometimes it makes sense to add a new VASP tag. This can be supplied
            in settings.inputs.unsupported_parameters as dict with the following
            specifications:
            unsupported_parameters = {
                'my_unsupported_parameter': {
                'default': 1.0,
                'description': 'Some description',
                'type': float,
                'values': [1.0, 2.0]
            }
        clean_workdir: bool or Bool (default is True)
            Whether (potentially remote) workdir should be cleaned.
            Set to False to use for a restart, e.g. to calculate bands
            from a pre-computed CHGCAR file. Prefer this option to adding
            CHGCAR to the list of retrieved files, as this takes too much
            space on the repository.
        test_mode: bool (default is True)
            Will be used to pick options ensuring a fast execution of
            calculations (with the risk of hitting the walltime) for test
            purposes. Default is set to True because dos calculations are
            fast for most systems.
        dry_run: bool (default is False)
            Build process but do not submit. Function will return None in this
            case.
        process_description: str or None (default is None)
            Description that will be added as the description attribute
            of the process node.
        verbosity: int (default is 1)
            Verbosity level

    Returns:
        WorkChainNode if dry_run is False, None otherwise.
    """

    structure = get_structure_from_pk_or_node(structure)

    if isinstance(code, str):
        code = Code.get_from_string(code)
    if code is None:
        # TODO: adapt for code(s) other than VASP ?
        code = get_default_vasp_code_and_computer(verbosity=verbosity)
    if isinstance(options, dict):  # convert dict to AttributeDict
        options = AttributeDict(options)
    if options is None:
        options = get_computer_options(
            computer_label=code.get_computer_label(),
            wall_time_s=wall_time_s, nb_of_nodes=nb_of_nodes,
            test_mode=test_mode, return_attributedict=True,
            verbosity=verbosity)

    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    kpoints_data = KpointsData()

    # TODO: define k_point_mesh from density
    kpoints_data.set_cell_from_structure(structure)
    kpoints_data.set_kpoints_mesh_from_density(kpoints_spacing)

    # Set the workchain to call. No () : we use the class rather than instance
    workchain = WorkflowFactory('vasp.verify')

    # Declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()
    # organize settings
    if isinstance(parser_settings, (dict, AttributeDict)):
        settings.parser_settings = parser_settings
    else:
        settings.parser_settings = {'add_dos': True,
                                    'add_misc': ['total_energies',
                                                 'notifications',
                                                 'run_status',
                                                 'run_stats',
                                                 'version',
                                                 'fermi_level',
                                                ]}
    # Add unsupported parameters to settings if any (setting a default value is sufficient):
    if isinstance(unsupported_parameters, dict):
        settings['unsupported_parameters'] = unsupported_parameters

    # set inputs for the following WorkChain execution
    inputs.code = code
    # set structure
    inputs.structure = structure

    inputs.kpoints = kpoints_data

    # set parameters
    if isinstance(incar_parameters, (dict, AttributeDict)):

        if 'lorbit' not in incar_parameters['incar'].keys():
            incar_parameters['incar']['lorbit'] = 11
        if 'nedos' not in incar_parameters['incar'].keys():
            incar_parameters['incar']['nedos'] = nb_of_energy_points
        inputs.parameters = DataFactory('dict')(dict=incar_parameters)
    else:
        # Get default incar parameters as an AttributeDict that can be modified
        incar_parameters = get_default_incar_parameters(structure=structure,
            return_attributedict=True, verbosity=verbosity)
        # TODO: add DOS-specific parameters
        incar_parameters.incar['ismear'] = -5
        if 'sigma' in incar_parameters.incar.keys():
            incar_parameters.incar.pop('sigma')
        incar_parameters.incar['lorbit'] = 11
        incar_parameters.incar['nedos'] = nb_of_energy_points
        inputs.parameters = DataFactory('dict')(dict=incar_parameters)

    inputs = set_potentials_family_and_mapping_input(inputs,
        potential_family, potential_mapping)

    # set options
    if isinstance(options, (dict, AttributeDict)):
        inputs.options = dict_data(dict=options)
    # set settings
    if isinstance(settings, (dict, AttributeDict)):
        inputs.settings = dict_data(dict=settings)

    # set workchain related inputs
    inputs.verbose = Bool(True)  # For testing purposes

    if isinstance (clean_workdir, bool):
        inputs.clean_workdir = Bool(clean_workdir)
    elif isinstance (clean_workdir, Bool):
        inputs.clean_workdir = clean_workdir

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.

    if verbosity >= 2 or dry_run:
        print(('\n' + 100*'*' +
              '\nReady to submit {} workchain with inputs:\n'
              + 100*'*').format(workchain.__name__))
        print_process_inputs(inputs)
        print(100*'*')

    if verbosity >= 3:  # DEBUGGING
        print('raw inputs:')
        for k, v in inputs.items():
            if isinstance(v, AttributeDict):
                print(f'{k}:')
                for k_k, k_v in v.items():
                    print(f'    {k_k}: {k_v} (of type {type(k_v)})')
            else:
                print(f'{k}: {v} (of type {type(v)})')

    if dry_run:
        return None

    process_node = submit(workchain, **inputs)
    if verbosity >= 1:
        print('{}<{}> has been submitted.'.format(process_node.process_label,
                                                  process_node.pk))
    # Add description attribute to process node
    if isinstance(process_description, str):
        process_node.description = process_description
        if verbosity >= 1:
            print('Description: \'{}\''.format(process_node.description))

    return process_node


def build_and_submit_bands(prerun_node,
                           structure=None,
                           code=None,
                           potential_family=None,
                           potential_mapping=None,
                           options=None,
                           wall_time_s=None,
                           nb_of_nodes=1,
                           incar_parameters=None,
                           unsupported_parameters=None,
                           kpoints_distance=0.05,
                           gaussian_smearing=True,
                           smearing_sigma=0.05,
                           bands_decompose_wave=False,
                           bands_decompose_bands=False,
                           bands_lm=False,
                           bands_phase=False,
                           bands_wigner_seitz_radius=None,
                           parser_settings=None,
                           test_mode=True,
                           dry_run=False,
                           process_description=None,
                           verbosity=1):
    """
    Build a bands calculation with the vasp.bands workchain as automatically
    as possible from a previous run and submit

    WARNING: make sure that function get_computer_options() is completed
    according to your computers and schedulers configurations.

    Default potential mapping and incar parameters may be modified in
    functions:
        - get_default_incar_parameters,
        - get_default_potential_mapping,
        - get_default_potential_family

    TODO: update arguments

    Args:
        prerun_node: ProcessNode
            Node containing the remote_folder in which restart files
            will be found.
        structure: StructureData, str, int or None (default is None)
            If None the structure will be obtained from the pre-run node
            outputs (if any) or inputs.
        code: str, Code or None (default is None)
            Default will be loaded with get_default_vasp_code_and_computer()
        potential_family: str, Str or None (default is None)
            potential_family as listed by 'verdi data vasp-potcar listfamilies'.
            Default will be set according to code and
            function get_default_potential_family().
        potential_mapping: dict, Dict or None (default is None)
            dict associating elment names in the considered system with a
            potential name. If None the default mapping will be obtained from
            the structure.
        options: dict, AttributeDict or None
            If None options will be defined automatically using the
            code.get_computer_label() and function get_computer_options()
        wall_time_s: int, float or None (default is None)
            Calculation walltime in seconds. if not specified it will be set
            automatically according to the default queue, qos, etc., defined in
            get_computer_options().
        nb_of_nodes: int (default is 1)
            Number of calculation nodes. The number of cpus will be calculated
            automatically with get_computer_options().
        kpoints_distance: float (default is 0.05)
            maximum spacing between k-points in the reciprocal space in Ang^-1
            used to automatically calculate a grid of k-points.
            WARNING: this parameter is 2*pi smaller than the KSPACING
            used in VASP INCAR files. VASP KSPACING default of 0.5 corresponds
            to kpoints_spacing ~ 0.08 Ang^1.
        gaussian_smearing: bool (default is True)
            Whether a Gaussian smearing should be used.
        smearing_method_index: int (default is 0, Gaussian smearing)
            ISMEAR parameter in VASP. 0 for Gaussian smearing.
            DO NOT USE THE TETRAHEDRON METHOD (-4 or -5) FOR BANDS
        smearing_sigma: (default is 0.05)
        bands_decompose_wave: boool (default is False)
            Whether to decompose the wave function.
        bands_decompose_bands: bool (default is False)
            Whether to decompose the band structure on each atom.
        bands_lm: bool (default: False)
            Whether to further decompose the decomposition into l- and
            m-states.
        bands_phase: bool (default is False)
            Further decompose the l- and m-state decomposition into phases.
        bands_wigner_seitz_radius: list, List or None (default is None)
            The Wigner-Seitz radius for each atom type in AA as a list. If set,
            the internal projectors are not utilized.
        unsupported_parameters: dict or None (default is None)
            dict listing VASP parameters not yet supported by the
            aiida_vasp ParametersMassager (see documentation here:
            https://aiida-vasp.readthedocs.io/en/latest/concepts/parameters.html)
            Here is the content of this documentation:
            In case you for instance perform developments in the VASP code,
            sometimes it makes sense to add a new VASP tag. This can be supplied
            in settings.inputs.unsupported_parameters as dict with the following
            specifications:
            unsupported_parameters = {
                'my_unsupported_parameter': {
                'default': 1.0,
                'description': 'Some description',
                'type': float,
                'values': [1.0, 2.0]
            }
        test_mode: bool (default is True)
            Will be used to pick options ensuring a fast execution of
            calculations (with the risk of hitting the walltime) for test
            purposes. Default is set to True because dos calculations are
            fast for most systems.
        dry_run: bool (default is False)
            Build process but do not submit. Function will return None in this
            case.
        process_description: str or None (default is None)
            Description that will be added as the description attribute
            of the process node.
        verbosity: int (default is 1)
            Verbosity level

    Returns:
        WorkChainNode if dry_run is False, None otherwise.
    """
    # Check prerun_node type: pk, UUID, process node or remote folder
    if isinstance(prerun_node, (str, int)):
        prerun_node = load_node(prerun_node)

    if isinstance(prerun_node, ProcessNode):
        restart_folder = prerun_node.outputs.remote_folder
    elif isinstance(prerun_node, RemoteData):
        restart_folder = prerun_node
    if verbosity >= 1:
        print(('Bands calculations will be conducted with charges pre-'
               'calculated in restart folder {} with process {}<{}>').format(
                   restart_folder.pk, restart_folder.creator.process_label ,
                   restart_folder.creator.pk))

    # Get structure from restart folder if None
    if structure is None:
        try:
            structure = restart_folder.creator.relax.structure
            if verbosity >= 1:
                print(('Using input structure<{}> from pre-run {}<{}> '
                       'outputs.relax').format(structure.pk,
                        restart_folder.creator.process_label,
                        restart_folder.creator.pk))
        except Exception as e:
            if verbosity >= 2:
                print(e)
            try:
                structure = restart_folder.creator.outputs.structure
                if verbosity >= 1:
                    print(('Using input structure<{}> from pre-run {}<{}> '
                           'outputs').format(structure.pk,
                            restart_folder.creator.process_label,
                            restart_folder.creator.pk))
            except Exception as e:
                if verbosity >= 1:
                    print(('Structure could not be obtained from pre-run '
                           '{}<{}> output: {}').format(
                            restart_folder.creator.process_label,
                            restart_folder.creator.pk, e))
                try:
                    structure = restart_folder.creator.inputs.structure
                    if verbosity >= 1:
                        print(('Using input structure<{}> from pre-run {}<{}> '
                               'inputs').format(structure.pk,
                                restart_folder.creator.process_label,
                                restart_folder.creator.pk))
                except Exception as e:
                    print(('Structure could not be obtained from '
                           'pre-run {}<{}> output or input: {}').format(
                            restart_folder.creator.process_label,
                            restart_folder.creator.pk, e))


    if isinstance(code, str):
        code = Code.get_from_string(code)
    if code is None:
        # TODO: adapt for code(s) other than VASP ?
        code = get_default_vasp_code_and_computer(verbosity=verbosity)
    if isinstance(options, dict):  # convert dict to AttributeDict
        options = AttributeDict(options)
    if options is None:
        options = get_computer_options(
            computer_label=code.get_computer_label(),
            wall_time_s=wall_time_s, nb_of_nodes=nb_of_nodes,
            test_mode=test_mode, return_attributedict=True,
            verbosity=verbosity)

    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    # Set the workchain to call. No () : we use the class rather than instance
    workchain = WorkflowFactory('vasp.bands')

    # Declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()
    # organize settings
    if isinstance(parser_settings, (dict, AttributeDict)):
        settings.parser_settings = parser_settings
    else:
        settings.parser_settings = {'output_params': ['total_energies',
                                                  'maximum_force'],
                                    'add_bands': True}
    # Add unsupported parameters to settings if any (setting a default value is sufficient):
    if isinstance(unsupported_parameters, dict):
        settings['unsupported_parameters'] = unsupported_parameters

    # set inputs for the following WorkChain execution
    inputs.code = code
    # set structure
    inputs.structure = structure

    # Setting restart folder from which pre-computed charges will be obtained
    inputs.restart_folder = restart_folder

    # set parameters and smearing (values in INCAR will take precedence)
    smearing = AttributeDict()
    if isinstance(incar_parameters, (dict, AttributeDict)):
        inputs.parameters = DataFactory('dict')(dict=incar_parameters)
    else:
        # Get default incar parameters as an AttributeDict that can be modified
        incar_parameters = get_default_incar_parameters(structure=structure,
            return_attributedict=True, verbosity=verbosity)
        # TODO: add DOS-specific parameters
        incar_parameters.incar['icharg'] = 11
        incar_parameters.incar['lorbit'] = 11
        inputs.parameters = DataFactory('dict')(dict=incar_parameters)

    if 'ismear' in inputs.parameters['incar'].keys():
        if inputs.parameters['incar']['ismear'] == 0:
            smearing.gaussian = Bool(True)
        elif inputs.parameters['incar']['ismear'] in [-4, -5]:
            raise ValueError(['Bands calculation is not compatible with the tetrahedron '
                              'smearing method (smearing_method_index = -4, -5). '
                              'Use (e.g.) Gaussian method (smearing_method_index = 0) '
                              'instead.'])
        else:
            smearing.gaussian = Bool(False)
    else:
        smearing.gaussian = Bool(gaussian_smearing)
    if 'sigma' in incar_parameters.incar:
        smearing.sigma = Float(incar_parameters.incar['sigma'])
    else:
        smearing.sigma = Float(smearing_sigma)
    inputs.smearing = smearing

    # set bands input
    bands = AttributeDict()
    if isinstance(kpoints_distance, Float):
        bands.kpoints_distance = kpoints_distance
    else:
        bands.kpoints_distance = Float(kpoints_distance)
    # TODO: set bands.decompose_wave, bands.decompose_bands, bands.lm,
    #       bands.phase, bands.wigner_seitz_radius
    # Currently these are all set to False (default in vasp.bands workchain)
    bands.decompose_wave = Bool(bands_decompose_wave)
    bands.decompose_bands = Bool(bands_decompose_bands)
    bands.lm = Bool(bands_lm)
    bands.phase = Bool(bands_phase)
    if isinstance(bands_wigner_seitz_radius, list):
        bands.wigner_seitz_radius = List(bands_wigner_seitz_radius)
    elif isinstance(bands_wigner_seitz_radius, List):
        bands.wigner_seitz_radius = bands_wigner_seitz_radius
    inputs.bands = bands

    # set potentials and their mapping
    inputs = set_potentials_family_and_mapping_input(inputs,
        potential_family, potential_mapping)

    # set options
    if isinstance(options, (dict, AttributeDict)):
        inputs.options = DataFactory('dict')(dict=options)
    # set settings
    if isinstance(settings, (dict, AttributeDict)):
        inputs.settings = DataFactory('dict')(dict=settings)

    # set workchain related inputs
    inputs.verbose = Bool(True)  # For testing purposes

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.

    if verbosity >= 2 or dry_run:
        print(('\n' + 100*'*' +
              '\nReady to submit {} workchain with inputs:\n'
              + 100*'*').format(workchain.__name__))
        print_process_inputs(inputs)
        print(100*'*')

    if verbosity >= 3:  # DEBUGGING
        print('raw inputs:')
        for k, v in inputs.items():
            if isinstance(v, AttributeDict):
                print(f'{k}:')
                for k_k, k_v in v.items():
                    print(f'    {k_k}: {k_v} (of type {type(k_v)})')
            else:
                print(f'{k}: {v} (of type {type(v)})')

    if dry_run:
        return None

    process_node = submit(workchain, **inputs)
    if verbosity >= 1:
        print('{}<{}> has been submitted.'.format(process_node,
                                                  process_node.pk))
    # Add description to process
    if isinstance(process_description, str):
        process_node.description = process_description
        if verbosity >= 1:
            print('Description: \'{}\''.format(process_node.description))

    return process_node


def get_temporary_file_from_retrieved_folder(process_or_node, file_name):
    """
    Args:
    file_name: str
        file name (basename) in retrieved folder.
    Returns:
        file object
    """
    # Get retrieved data
    if isinstance(process_or_node, FolderData):
        retrieved = process_or_node
    else:
        process_node = get_node_from_id_or_node(process_or_node)
        retrieved = process_node.outputs.retrieved

    if file_name not in retrieved.list_object_names():
        raise FileNotFoundError('No {} file in retrieved folder <{}>'.format(
            file_name, retrieved.pk))

    tmp_path = tempfile.mkdtemp()
    with open(os.path.join(tmp_path, 'vasprun.xml'), 'w') as file:
        file.write(retrieved.get_object_content('vasprun.xml'))

    return file


def get_band_gap_from_retrieved_vasprun(process_or_node,
                                        return_pymatgen_vasprun=False,
                                        verbosity=1):
    """
    Get band gap from retried vasprun.xml file using pymatgen

    Args:
        process_or_node: FolderData, ProcessData, int, str
            Process node, process id (UUID or PK) or FolderData
            corresponding to a retrieved output
        return_pymatgen_vasprun: bool (default is False)
            whether the retrieved pymatgen Vasprun instance
            should be returned
        verbosity: int (default is 1)
            Verbosity level

    Returns:
        dict with kets direct (bool), energy (float) and transition (str)
        Vasprun instance if return_pymatgen_vasprun argument is True
    """
    # Get retrieved data
    if isinstance(process_or_node, FolderData):
        retrieved = process_or_node
    else:
        process_node = get_node_from_id_or_node(process_or_node)
        retrieved = process_node.outputs.retrieved

    if 'vasprun.xml' not in retrieved.list_object_names():
        raise FileNotFoundError('No vasprun.xml in retrieved folder <{}>'.format(
            retrieved.pk))

    tmp_path = tempfile.mkdtemp()
    with open(os.path.join(tmp_path, 'vasprun.xml'), 'w') as f:
        f.write(retrieved.get_object_content('vasprun.xml'))

    vr = Vasprun(os.path.join(tmp_path, 'vasprun.xml'))
    gap_dict = vr.get_band_structure().get_band_gap()
    if verbosity >= 1:
        if gap_dict['direct']:
            print(('Direct band gap of {:3f} eV detected (transition {}) '
                   'from {}<{}>').format(gap_dict['energy'],
                gap_dict['transition'], process_node.process_label,
                process_node.pk))
        else:
            print('No direct band gapdetected from {}<{}>'.format(
                process_node.process_label, process_node.pk))
    if return_pymatgen_vasprun:
        return gap_dict, vr
    else:
        return gap_dict


def get_nmr_parameters_from_vasp_process(proc_or_id):
    """
    Get NMR paremeters from an aiida-vasp process

    Procedure uses pymatggen to get the data from
    the retrieved VASP OUTCAR file
    """
    process_node = get_node_from_id_or_node(proc_or_id)
    tmp_path = tempfile.mkdtemp()
    retrieved = process_node.outputs.retrieved
    with open(os.path.join(tmp_path, 'OUTCAR'), 'w') as file:
        file.write(retrieved.get_object_content('OUTCAR'))
    outcar_dict = Outcar(filename=os.path.join(tmp_path, 'OUTCAR')).as_dict()
    nb_of_sites = len(process_node.inputs.structure.sites)

    nmr_list = [{
        'chemical_shielding': ChemicalShielding.from_maryland_notation(
            *outcar_dict['nmr_cs']['valence and core'][i]),
        'electric_field_gradient': ElectricFieldGradient(outcar_dict['nmr_efg']['raw'][i]),
        'efg_parameters': outcar_dict['nmr_efg']['parameters'][i]}
        for i in range(nb_of_sites)]

    # Define specific parameters from ChemicalShielding instance:
    # e.g. cs.haeberlen_values.sigma_iso
    for i in range(nb_of_sites):
        nmr_list[i]['cs_sigma_iso'] = nmr_list[i]['chemical_shielding'
            ].haeberlen_values.sigma_iso
        nmr_list[i]['cs_delta_sigma_iso'] = nmr_list[i]['chemical_shielding'
            ].haeberlen_values.delta_sigma_iso
        nmr_list[i]['cs_eta'] = nmr_list[i]['chemical_shielding'
            ].haeberlen_values.eta
        nmr_list[i]['cs_omega'] = nmr_list[i]['chemical_shielding'
            ].maryland_values.omega
        nmr_list[i]['cs_span'] = nmr_list[i]['cs_omega']
        nmr_list[i]['cs_kappa'] = nmr_list[i]['chemical_shielding'
            ].maryland_values.kappa
        nmr_list[i]['cs_skew'] = nmr_list[i]['chemical_shielding'
            ].maryland_values.kappa
        nmr_list[i]['efg_cq'] = nmr_list[i]['efg_parameters']['cq']
        nmr_list[i]['efg_eta'] = nmr_list[i]['efg_parameters']['eta']
        nmr_list[i]['efg_nuclear_quadrupole_moment'] = nmr_list[i]['efg_parameters']['nuclear_quadrupole_moment']
        nmr_list[i]['efg_v_xx'] = nmr_list[i]['electric_field_gradient'].V_xx
        nmr_list[i]['efg_v_yy'] = nmr_list[i]['electric_field_gradient'].V_yy
        nmr_list[i]['efg_v_zz'] = nmr_list[i]['electric_field_gradient'].V_zz
    return nmr_list

@calcfunction
def store_nmr_parameters_from_process(retrieved_node: FolderData):
    """
    Store an ArrayData containing calculated NMR paremeters in database

    parameters are parsed from the retrieved VASP OUTCAR file
    using pymatgen's Outcar class and ChemicalShielding and
    ElectricFieldGradient classes.


    """
    process_node = retrieved_node.creator
    """
    print('NMR paramaters Array will be created from retrieved output {} of process {}<{}>'.format(
        retrieved_node.pk, process_node.process_label, process_node.pk))
    """
    nmr_list = get_nmr_parameters_from_vasp_process(process_node)
    array = ArrayData()
    for prop in ['cs_sigma_iso', 'cs_delta_sigma_iso', 'cs_eta',
                 'cs_omega', 'cs_span', 'cs_kappa', 'cs_skew',
                 'efg_cq', 'efg_eta', 'efg_nuclear_quadrupole_moment',
                 'efg_v_xx', 'efg_v_yy', 'efg_v_zz']:
        array.set_array('nmr_{}'.format(prop), np.array([_[prop] for _ in nmr_list]))
    array.description = 'NMR paramaters parsed from retrieved output {} of process {}<{}>'.format(
        retrieved_node.pk, process_node.process_label, process_node.pk)
    """
    for name in array.get_arraynames():
        print('array[\'{}\'] = {}'.format(name, array.get_array(name)))
    """
    return array


def get_planewave_cutoff_from_potcar(potcar, cutoff_scaling=1.3,
                                     rounding_decimals=0):
    """
    Get planewave-cutoff energy based on recommended ENMAX values in pseudo

    Args:
        potcar:
            pymatgen.io.vasp.inputs.Potcar object.
        cutoff_scaling: float (default: 1.3)
            scaling factor applied to max(enmax(i)). Should be at least
            1.3 for variable-cell relaxations or accurate energy calculations
        rounding_decimals: int (default: 0)
            Number of decimals for the rounding of the returned cutoff value.

    returns:
        cutoff_scaling * max(ENMAX(i)) with {i} the elements in potcar
        object
    """
    enmaxs = [pot.keywords['ENMAX'] for pot in potcar]
    return np.round(cutoff_scaling * max(enmaxs), decimals=rounding_decimals)


def get_nb_of_valence_electrons(pmg_structure, potcar=None):
    """
    Calculate number of valance electrons based on pmg_structure and pseudos

    Args:
        pmg_structure: pymatgen Structure
        potcar: pymatgen Potcar or None (default is None)
            if None potcar is generated from the structure
            with the default pymatgen functional.

    returns:
        nb_of_electrons: Int
            The number of valence electrons
    """
    if potcar is None:
        potcar = Potcar(pmg_structure.symbol_set)
    nelect_by_type = {pot.element: pot.nelectrons for pot in potcar}
    nb_of_electrons = 0
    for atom_type in pmg_structure.symbol_set:
        nb_of_electrons += nelect_by_type[atom_type] * pmg_structure.composition[atom_type]
    return int(nb_of_electrons)


def get_nb_of_empty_bands(nb_of_cpus, nb_of_valence_electrons=None, structure=None,
                          potcar=None, return_nb_of_bands=False, verbosity=1):
    """
    Get the number of bands that vasp will use automatically
    based on the number of requested processors

    Works for paired electrons only
    """
    # Get a pymatgen (I)Structure

    # Get the number of valence electrons
    if nb_of_valence_electrons is None:
        if isinstance(structure, StructureData):
            pmg_structure = structure.get_pymatgen()
        elif isinstance(structure, (Structure, IStructure)):
            pmg_structure = structure
        else:
            raise TypeError(('A structure (aiida orpymatgen) should be defined '
                             'if nb_of_valence_electrons is None'))
        nb_of_valence_electrons = get_nb_of_valence_electrons(pmg_structure,
                                                              potcar=potcar)
    coeff = 1
    while 1:
        if coeff * nb_of_cpus > np.ceil(nb_of_valence_electrons / 2):
            break
        else:
            coeff += 1
    nb_of_bands = int(coeff * nb_of_cpus)
    nb_of_empty_bands = int(nb_of_bands - np.ceil(nb_of_valence_electrons / 2))
    if verbosity >= 2:
        print('{} bands and {} empty bands for {} valence electrons.'.format(
            nb_of_bands, nb_of_empty_bands, nb_of_valence_electrons))
    if return_nb_of_bands:
        return nb_of_empty_bands, nb_of_bands
    else:
        return nb_of_empty_bands


def get_best_nb_of_taks_per_node(nb_of_valence_electrons, max_nb_of_tasks_per_node,
                                 target_nb_of_bands=None, min_nb_of_bands=None,
                                 max_nb_of_bands=None, extra_bands_bands_rel_tol=0.25,
                                 nb_of_nodes=1, min_nb_of_tasks_per_node=None, 
                                 min_to_max_nb_of_tasks_ratio=0.667, 
                                 target_extra_bands_ratio=0.10, verbosity=1):
    """
    Get a optimimum nb of tasks per node to approach a target nb of bands
    taking the number of cores per node into account.

    In some VASP calculations NCORE (the number of cores acting on one band) 
    should be set to 1. In such cases the number of bands ultimately used
    will be the first multiple of the number of requested CPUs above the
    requested NBANDS. This functions searches a number of cores per node
    as close as possible to the maximum (nb of physical cores per node)
    that gives a number of bands close to a target value, which by default
    is a number of empty bands 10% the number of occupied bands.
    
    This functions works for paired-electrons systems.
    
    TODO:
    The default NBANDS calculated by VASP seems a clever choice of target
    number of bands: 
    NBANDS 	= max(NELECT/2+NIONS/2,NELECT*0.6) 	
    This requires that either the number of atoms or the structure is provided...

    Args:
        nb_of_valence_electrons: int
            number of electrons that are not included in 
            pseudopotential. May be obtained with function
            get_nb_of_valence_electrons().
        max_nb_of_tasks_per_node: int
            Should be set to the number of processors
            per node.
        target_nb_of_bands: int or None (default is None)
            by default it will be set to (1 + target_extra_bands_ratio)
            times the number of occupied bands (nb_of_valence_electrons/2)
        min_nb_of_bands: int or None (default is None)
            Will be set by default to:
            nb_of_occupied_bands + (1 - extra_bands_bands_rel_tol) * target_extra_bands
        max_nb_of_bands: int or None (default is None)
            Will be set by default to:
            nb_of_occupied_bands + (1 + extra_bands_bands_rel_tol) * target_extra_bands
        extra_bands_bands_rel_tol: float (default is 0.25)
            Defines the interval between:
            - the minimum and the target number of extra bands
            - the maximum and the target number of extra bands
        nb_of_nodes: int (default is 1)
            Number of requested nodes. The number of requested CPUs
            will then have to be set to:
            nb_of_nodes * nb_of_tasks_per_node 
        min_nb_of_tasks_per_node: int or None (default is None)
            The minium number of CPUs requested per node.
            By default it will be set to:
            min_to_max_nb_of_tasks_ratio * max_nb_of_tasks_per_node
        min_to_max_nb_of_tasks_ratio: float (default is 0.667)
            Default ratio between the minimum allowed number
            of requested CPUs per node and the total number of CPUs.
            This default value should be appropriate in most cases.
            per node (designated here as max_nb_of_tasks_per_node)
        target_extra_bands_ratio: float (default is 0.10)
            By default, a number of extra bands of 10% of the number
            occupied bands is used. Smaller value could potentially
            be used in systems with many electrons.
        verbosity=1
        
    Returns:
        nb_of_tasks_per_node: int
    """
    nb_of_occupied_bands = int(np.ceil(0.5 * nb_of_valence_electrons))
    if target_nb_of_bands is None:  # Set min_nb_of_bands to NELECT/2 + 10%
        target_nb_of_bands = int(np.ceil((1 + target_extra_bands_ratio)
                                         * nb_of_occupied_bands))
        if verbosity >= 2:
            print('target nb of bands set to {} ({} empty bands).'.format(
                target_nb_of_bands, target_nb_of_bands - nb_of_occupied_bands))
    target_extra_bands = np.ceil(target_nb_of_bands - nb_of_occupied_bands)
    if verbosity >= 2:
        print('target nb of extra bands is {}'.format(target_extra_bands))

    if min_nb_of_bands is None:
        min_nb_of_bands = int(np.ceil(nb_of_occupied_bands + 
            (1 - extra_bands_bands_rel_tol) * target_extra_bands))
    if max_nb_of_bands is None:  # Set max_nb_of_bands to NELECT/2 + 10%
        max_nb_of_bands = int(np.ceil(nb_of_occupied_bands + 
            (1 + extra_bands_bands_rel_tol) * target_extra_bands))
    if min_nb_of_tasks_per_node is None:
        min_nb_of_tasks_per_node = int(np.ceil(min_to_max_nb_of_tasks_ratio * max_nb_of_tasks_per_node))
    if verbosity >= 2:
        print(('Trying to find an optimum number of tasks per node between {} and {} '
               'to obtain number of bands between {} and {} ({} to {} empty bands).'
               ).format(min_nb_of_tasks_per_node, max_nb_of_tasks_per_node,
                        min_nb_of_bands, max_nb_of_bands,
                        min_nb_of_bands - nb_of_occupied_bands,
                        max_nb_of_bands - nb_of_occupied_bands))
    # Decrease number of tasks untill target_nb_of_bands is obtained or ???
    nb_of_tasks_per_node = max_nb_of_tasks_per_node
    best_overshoot = np.inf
    best_nb_of_tasks_per_node = nb_of_tasks_per_node
    while 1:
        nb_of_cpus = nb_of_tasks_per_node * nb_of_nodes
        if verbosity >= 2:
            print('Trying nb_of_tasks_per_node = {} ({} cpus):'.format(
                nb_of_tasks_per_node, nb_of_cpus))
        (nb_of_empty_bands, nb_of_bands) = get_nb_of_empty_bands(nb_of_cpus,
            nb_of_valence_electrons, return_nb_of_bands=True, verbosity=verbosity)
        if nb_of_tasks_per_node <= min_nb_of_tasks_per_node:
            print('No good value found. Retrieving best compromise: ')
            nb_of_tasks_per_node = best_nb_of_tasks_per_node
            break
        elif np.isclose(nb_of_bands, target_nb_of_bands) or (
                nb_of_bands >= min_nb_of_bands and  nb_of_bands <= max_nb_of_bands):
            if verbosity >= 2:
                print('Good match found:')
            break
        elif nb_of_bands > max_nb_of_bands:
            overshoot = nb_of_bands - max_nb_of_bands
            if overshoot < best_overshoot:
                best_overshoot= overshoot
                best_nb_of_tasks_per_node = nb_of_tasks_per_node
                if verbosity >= 2:
                    print('Best compromise so far.')
            
        else:
            if verbosity >= 2:
                print('failed attempt: nb_of_bands ({}) < min_nb_of_bands ({}).'.format(
                    nb_of_bands, min_nb_of_bands))
        nb_of_tasks_per_node -= 1
        if verbosity >= 2:
            print('Decreasing nb_of_tasks_per_node to {}.'.format(nb_of_tasks_per_node))

    if verbosity >= 1:
        print(('nb_of_tasks_per_node = {} ({} cpus) -> {} bands, including {} '
               'empty bands.').format(nb_of_tasks_per_node, nb_of_cpus,
                                      nb_of_bands, nb_of_empty_bands))
    
    return int(nb_of_tasks_per_node)



def get_quadrupolar_moment_mapping(potcar_or_elements, isotopes=None):
    """
    Get a mapping of quadrupolar moments vs element.

    WARNING: the list of isotopes defined in pymatgen (v. 2020.4.29) is still
    incomplete (e.g. no isotope defined for Ga, As)

    Args:
        potcar: pymatgen.io.vasp.input.Poscar instance
            used to obtain the list of symbols in the order compatible
            with the poscar file.

        isotopes: list of str (default is None)
            Each element in the list should be in the form 'Li-7'.
            If None the lowest-mass nmr-active isotope defined in pymatgen periodic_table.

    Returns:
        quadrupolar_moment_mapping: a nested dict of the form:
            {'Al': {'Q_mb': VALUE_IN_millibarn, 'isotope': 'Al-27'}}
    """
    quadrupolar_moment_mapping = {}

    if isinstance(potcar_or_elements, Potcar):
        elements = potcar_or_elements.symbols
    elif isinstance(potcar_or_elements, (list, tuple)):
        elements = potcar_or_elements
    else:
        raise TypeError('potcar_or_elements should be a Potcar or list of string.')

    picked_isotopes = []
    for atom_type in elements:
        q_dict = Element(atom_type).nmr_quadrupole_moment
        if len(q_dict) == 0:
            #TODO: check whether isotopes exist in file nuclearQuadrupoleMoments_StoneNJ_2013.csv
            quadrupolar_moment_mapping[atom_type] = {
                'Q_mb': 1.0, 'isotope': None}
        else:
            count = 0
            for k, v in q_dict.items():
                if count == 0:
                    quadrupolar_moment_mapping[atom_type] = {
                        'Q_mb': v, 'isotope': k}
                elif isotopes is None:
                    quadrupolar_moment_mapping[atom_type] = {
                        'Q_mb': 1.0, 'isotope': None}
                elif count >= 1 and k in isotopes:
                    quadrupolar_moment_mapping[atom_type]['Q_mb'] = v
                    quadrupolar_moment_mapping[atom_type]['isotope'] = k
                count += 1
    return quadrupolar_moment_mapping


def build_and_submit_nmr(structure: StructureData,
                         pmg_structure=None,
                         code=None,
                         potential_family=None,
                         potential_mapping=None,
                         options=None,
                         wall_time_s=None,
                         account=None,
                         nb_of_nodes=1,
                         incar_parameters=None,
                         unsupported_parameters=None,
                         kpoints_spacing=0.25,
                         parser_settings=None,
                         clean_workdir=True,
                         test_mode=False,
                         dry_run=False,
                         pymatgen_functional='PBE_54',
                         planewave_cutoff_energy=None,
                         planewave_cutoff_scaling=1.5,
                         verbosity=1):
    """
    Build vasp.relax workchain as automatically as possible and submit

    WARNING: make sure that function get_computer_options() is completed
    according to your computers and schedulers configurations.

    Args:
        structure: StructureData
            Input structure to relax
        code: str, Code or None (default is None)
            Default will be loaded with get_default_vasp_code_and_computer()
        potential_family: str or None (default is None)
            potential_family as listed by 'verdi data vasp-potcar listfamilies'.
            Default will be set according to code and
            function get_default_potential_family().
        potential_mapping: dict or None (default is None)
            dict associating elment names in the considered system with a
            potential name. If None the default mapping will be obtained from
            the structure.
        options: dict, AttributeDict or None
            If None options will be defined automatically using the
            code.get_computer_label() and function get_computer_options()
        wall_time_s: int, float or None (default is None)
            Calculation walltime in seconds. if not specified it will be set
            automatically according to the default queue, qos, etc., defined in
            get_computer_options().
        nb_of_nodes: int (default is 1)
            Number of calculation nodes. The number of cpus will be calculated
            automatically with get_computer_options().
        kpoints_spacing: float (default is 0.4)
            maximum spacing between k-points in the reciprocal space in Ang^-1
            used to automatically calculate a grid of k-points.
        unsupported_parameters: dict or None (default is None)
            dict listing VASP parameters not yet supported by the
            aiida_vasp ParametersMassager (see documentation here:
            https://aiida-vasp.readthedocs.io/en/latest/concepts/parameters.html)
            Here is the content of this documentation:
            In case you for instance perform developments in the VASP code,
            sometimes it makes sense to add a new VASP tag. This can be supplied
            in settings.inputs.unsupported_parameters as dict with the following
            specifications:
            unsupported_parameters = {
                'my_unsupported_parameter': {
                'default': 1.0,
                'description': 'Some description',
                'type': float,
                'values': [1.0, 2.0]
            }
        test_mode: bool (default is False)
            Will be used to pick options ensuring a fast execution of
            calculations (with the risk of hitting the walltime) for test
            purposes.
        dry_run: bool (default is False)
            Build process but do not submit. Function will return None in this
            case.
        verbosity: int (default is 1)
            Verbosity level

    Returns:
        WorkChainNode if dry_run is False, None otherwise.
    """
    # Use pymatgen's MPNMRSet to fetch recommended pseudopotentials
    if pmg_structure is None:
        pmg_structure = structure.get_pymatgen()
    if pymatgen_functional is not None:
        nmr_set = MPNMRSet(pmg_structure,
                           user_potcar_functional=pymatgen_functional)
    else:
        # Use PMG_DEFAULT_FUNCTIONAL defined in ~/.pmgrc.yaml
        nmr_set = MPNMRSet(pmg_structure)

    if potential_family is None:
        potential_family = get_default_potential_family(code,
            return_str=True)

    # Set potential family from the vasp MPNMRSet
    if potential_mapping is None:
        potential_mapping = {at_type: pot for at_type, pot in zip(
           nmr_set.poscar.site_symbols, nmr_set.potcar_symbols)}

    if isinstance(code, str):
        code = Code.get_from_string(code)
    if code is None:
        # TODO: adapt for code(s) other than VASP ?
        code = get_default_vasp_code_and_computer(verbosity=verbosity)
    if isinstance(options, dict):  # convert dict to AttributeDict
        options = AttributeDict(options)
    if options is None:
        options = get_computer_options(
            computer_label=code.get_computer_label(),
            wall_time_s=wall_time_s, nb_of_nodes=nb_of_nodes,
            test_mode=test_mode, account=account,
            return_attributedict=True, verbosity=verbosity)

    # Adapt computer options to optimize number of processors
    # in accordance with a target number of bands (10% of occupied bands by default)
    nb_of_valence_electrons = get_nb_of_valence_electrons(pmg_structure,
                                                          nmr_set.potcar)
    max_nb_of_tasks_per_node = options["resources"]["num_mpiprocs_per_machine"]
    min_nb_of_tasks_per_node = int(np.ceil(2/3*max_nb_of_tasks_per_node))
    options["resources"]["num_mpiprocs_per_machine"] =  get_best_nb_of_taks_per_node(
        nb_of_valence_electrons, max_nb_of_tasks_per_node, 
        extra_bands_bands_rel_tol=0.25,  nb_of_nodes=options["resources"]["num_machines"], 
        min_nb_of_tasks_per_node=min_nb_of_tasks_per_node, verbosity=verbosity)

    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    kpoints_data = KpointsData()

    # TODO: define k_point_mesh from density
    kpoints_data.set_cell_from_structure(structure)
    kpoints_data.set_kpoints_mesh_from_density(kpoints_spacing)

    # Set the workchain to call. No () : we use the class rather than instance
    # workchain = WorkflowFactory('vasp.verify')  # does not work for versions > 2.5.0
    workchain = WorkflowFactory('vasp.vasp')

    # Declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()
    # organize settings
    if isinstance(parser_settings, (dict, AttributeDict)):
        settings.parser_settings = parser_settings
    else:
        settings.parser_settings = {'output_params': ['total_energies'],
                                    'add_maximum_stress': False,
                                    'add_forces': True}
    # Add unsupported parameters to settings if any (setting a default value is sufficient):
    if isinstance(unsupported_parameters, dict):
        settings['unsupported_parameters'] = unsupported_parameters

    # set inputs for the following WorkChain execution
    inputs.code = code
    # set structure
    inputs.structure = structure

    inputs.kpoints = kpoints_data

    # set parameters
    if isinstance(incar_parameters, (dict, AttributeDict)):
        inputs.parameters = dict_data(dict=incar_parameters)
    else:
        if planewave_cutoff_energy is None:
            planewave_cutoff_energy = get_planewave_cutoff_from_potcar(
                nmr_set.potcar, cutoff_scaling=planewave_cutoff_scaling)
        quadrupolar_moment_mapping = get_quadrupolar_moment_mapping(
            nmr_set.poscar.site_symbols)

        inputs.parameters = dict_data(dict={'incar': {
            "SYSTEM"   :   '{} NMR calculation'.format(structure.get_formula()),
            "GGA"      :   'PE' ,       # Perdew-Burke-Ernzerhof (PBE) XC functional
            "ENCUT"    :   planewave_cutoff_energy,        # Plane-wave energy cutoff (eV); adjust as needed
            "NSW"      :   0 ,        # Maximum number of ionic relaxation steps
            "ISYM"     :   0 ,        # seemingly recommended.
            "EDIFF"    :   1e-7,        # Electronic self-consistency criterion (convergence threshold)
            "ALGO"     :  'Normal',     # Electronic relaxation algorithm (Normal is recommended)
            "ISMEAR"   :   0   ,        # no fancy smearing
            "SIGMA"    :   0.05,
            "LREAL"    :  "Auto",        # Make sure there is no problem with
            "PREC"     :  "Accurate",
                # NMR-related parameters
            "LEFG"     :  True,
            "QUAD_EFG" :  [quadrupolar_moment_mapping[atom_type]['Q_mb'] for atom_type
                           in nmr_set.poscar.site_symbols],
            "NELM"     :  100,       # leave time for convergence in shielding calc.
            "LCHIMAG"  :  True,      # to switch on linear response for chemical shifts
            "DQ"       :  0.001,     # often the default is sufficient
            "ICHIBARE" : 1,          # often the default is sufficient
            "LNMR_SYM_RED": False,   # ?
            "NLSPLINE" : False,   # only needed if LREAL is NOT set.
            "LREAL"    : False,         # helps for speed for large systems, not needed
            "NBANDS"   : int(np.ceil(0.5*nb_of_valence_electrons)),   # NELECT/2 will let VASP take the closest multiple of NCPUs above.
            "NCORE"    : 1
    }})

    # set potentials and their mapping
    inputs.potential_family = Str(potential_family)
    inputs.potential_mapping = dict_data(dict=potential_mapping)
    
    # set options
    inputs.options = dict_data(dict=options)
    
    # set settings
    inputs.settings = dict_data(dict=settings)

    # set workchain related inputs
    inputs.verbose = Bool(True)  # For testing purposes

    if isinstance (clean_workdir, bool):
        inputs.clean_workdir = Bool(clean_workdir)
    elif isinstance (clean_workdir, Bool):
        inputs.clean_workdir = clean_workdir

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.

    if verbosity >= 2 or dry_run:
        print(('\n' + 100*'*' +
              '\nReady to submit {} workchain with inputs:\n'
              + 100*'*').format(workchain.__name__))
        print_process_inputs(inputs)
        print(100*'*')

    if verbosity >= 3:  # DEBUGGING
        print('raw inputs:')
        for k, v in inputs.items():
            if isinstance(v, AttributeDict):
                print(f'{k}:')
                for k_k, k_v in v.items():
                    print(f'    {k_k}: {k_v} (of type {type(k_v)})')
            else:
                print(f'{k}: {v} (of type {type(v)})')

    if dry_run:
        return None

    if test_mode:
        process_node = run_get_node(workchain, **inputs)
    else:
        process_node = submit(workchain, **inputs)

    if verbosity >= 1:
        try:
            print(('Process {}<{}> submitted with function build_and_submit_nmr() '
                   'on {} structure {}.').format(process_node.process_label, process_node.pk,
                                                 structure.get_formula(), structure.pk))
        except Exception as e:
            print('ERROR printing summary at the end of function build_and_submit_nmr():')
            print(e)
            print('process_node = {} (of type {})'.format(process_node, type(process_node)))
    return process_node


def get_group_from_str_or_instance(group_name_or_instance, verbosity=1):
    if isinstance(group_name_or_instance, str):
        group_label = group_name_or_instance
        try:
            group = load_group(
                label=group_label)
        except NotExistent:
            sys.exit('No group with label {} found in database.'.format(
                group_label))
    elif isinstance(group_name_or_instance, int):
        try:
            group = load_group(
                pk=group_name_or_instance)
            group_label
        except NotExistent:
            sys.exit('No group with pk {} found in database.'.format(
                group_name_or_instance))
    elif isinstance(group_name_or_instance, Group):
        group = group_name_or_instance
    else:
        sys.exit('group_name_or_instance should be valid aiida.orm.Group ' +
                 'instance or Group label.')
    if verbosity >= 2:
        print(f'Group name: {group.label}')
    return group


def create_group_from_uspex_data(uspex_path, structure_selection, group_label=None,
                                 group_description=None, uspex_run_tag=None, uspex_run_id=None,
                                 verbosity=1):
    """
    Create a group of structures and metadata associated with a uspex calculation

    Group should include:
        - input files: Parameters.txt (possibly files in Specific folder) as SingleFileData
        - main output files (SingleFileData):
            * OUTPUT.txt
            * gatheredPOSCARS
            * gatheredPOSCARS_unrelaxed
            * Individuals
            * goodStructures
        - a selection of structures:
            This implies (i) selecting the structure with usd.get_structure_IDs().
            The method will have to be extended to select files with other methods,
            including randomly.

    In complement, one should defined a function to load uspexStructuresData from the stored
    Individuals file.

    IMPLEMENTATION IN PROGRESS

    Args:
        uspex_path: str
            path to a UPSEX results directory.
        structure_selection: list or tuple of int
            If a list or tuple is provided
    """
    uspex_abs_path = os.path.abspath(uspex_path)
    usd = uspexStructuresData(fileOrDirectoryName=uspex_path)

    if isinstance(structure_selection, (list, tuple)):
        uspex_ids = [int(uspex_struct_id) for uspex_struct_id in structure_selection]
        """
        TODO: add more complex selection methods in a dict
        {'lowest_energies': NB_OF_STRUCTURES,
         'random': NB_OF_STRUCTURES,
         'lowest_energies': NB_OF_STRUCTURES,
         'random_seed': SEED
        }
        """
    else:
        raise TypeError('Invalid structure selection. Use list of USPEX structure IDs.')

    poscars = usd.get_poscars_from_IDs(uspex_ids)
    # structures_metadata = {}

    if uspex_run_tag is None:
        # Get uspex run name from directory name
        uspex_run_tag = os.path.basename(os.path.split(uspex_abs_path)[0])

    origin = 'uspex_{}'.format(uspex_run_tag)
    if uspex_run_id is not None:
        origin += '_{}'.format(uspex_run_id)

    if group_label is None:
        group_label = origin
    try:
        group = load_group(group_label)

    except NotExistent:
        group = Group(label=group_label)
        group.store()
        if group_description is None:
            group_description = 'Structures obtained from USPEX run {}'.format(uspex_run_tag)

        # Store original USPEX files into repository and add to group
        files_to_store = ['Parameters.txt', 'OUTPUT.txt', 'goodStructures', 'goodStructures_POSCARS',
                          'gatheredPOSCARS', 'gatheredPOSCARS_unrelaxed']

        # TODO: check whether files exist
        for _file in files_to_store:
            file_name = os.path.join(uspex_abs_path, _file)
            aiida_file = SinglefileData(file_name)
            aiida_file.description = 'USPEX {} file from run \'{}\''.format(
                _file, uspex_run_tag)
            aiida_file.store()
            aiida_file.extras = {'original_location': uspex_abs_path}
            group.add_nodes(aiida_file)

    if group_description is not None:
        group.description = group_description

    # TODO: Create a link to the original files.

    ids_sorted_by_fitness = [usd.IDs[index] for index in np.argsort(usd.fitnesses)]
    ids_sorted_by_energy = [usd.IDs[index] for index in np.argsort(usd.enthalpies)]

    aiida_stuct_pks = []
    for uspex_id, poscar in zip(uspex_ids, poscars):
        structure = StructureData(pymatgen=poscar.structure)
        structure.description = '{} uspex structure with ID {} extracted from run {}'.format(
            structure.get_formula(), uspex_id, origin)
        uspex_struct_index = list(usd.IDs).index(uspex_id)
        structure.store()
        structure.extras = {
            'origin': origin,
            'original_id': str(uspex_id),
            'uspex_data': {
                'ID': uspex_id,
                'generation_nb': usd.generationNumbers[uspex_struct_index],
                'energy': usd.enthalpies[uspex_struct_index],
                'fitness': usd.fitnesses[uspex_struct_index],
                'kpoints': usd.kpoints[uspex_struct_index],
                'space_group_nb': usd.symmGroupNb[uspex_struct_index],
                'Q_entr': usd.Q_entr[uspex_struct_index],
                'A_order': usd.A_order[uspex_struct_index],
                'S_order': usd.S_order[uspex_struct_index],
                'rank_by_fitness': ids_sorted_by_fitness.index(uspex_id),
                'rank_by_energy': ids_sorted_by_energy.index(uspex_id)
            }
        }
        # Store in database
        if verbosity >= 2:
            print('Structure {} stored in database with pk {}.'.format(
                structure.description, structure.pk))
        if verbosity >= 3:
            print('with extras: \n{}'.format(structure.extras))

        aiida_stuct_pks.append(structure.pk)
        group.add_nodes(structure)

    return group


def write_retrieved_file_from_process(proc_or_id, file_name, temp_file_name='tmp.txt'):
    """
    Get NMR paremeters from an aiida-vasp process

    Procedure uses pymatggen to get the data from
    the retrieved VASP OUTCAR file
    """
    process_node = get_node_from_id_or_node(proc_or_id)
    retrieved = process_node.outputs.retrieved
    retrieved_content = retrieved.get_object_content('OUTCAR')
    temp_file = os.path.abspath(temp_file_name)
    with open(temp_file, 'w') as file:
        file.write(retrieved_content)
    print(('Content of file {} from retrieved output node {} of process {} '
           'written in file {}.').format(file_name, retrieved.pk, process_node.pk,
                                         temp_file))

    return temp_file


def get_dataframe_from_nmr_calc_group(group_or_label, verbosity=1):
    """
    Get a pandas dataframe from NMR calculations in a group

    Args:
        group_or_label: Group, str or int
            group instance, label of pk

    Returns:
        dataframe: pandas dataframe
    """
    group = get_group_from_str_or_instance(group_or_label)

    qb = QueryBuilder()
    qb.append(Group, filters={'label': group.label}, tag='group')
    qb.append(WorkChainNode, with_group='group', tag='processes')

    mydict = {
        'process_pk': [],
        'proc_exit_status': [],
        'calc_pk': [],
        'calc_exit_status': [],
        'struct_pk': [],
        'formula': [],
        'nb_of_atoms': [],
        'remote_path': [],
    }

    for row in qb.all():
        process_node = row[0]
        calc_node = [proc for proc in process_node.called_descendants if 'Calculation' in proc.process_label][-1]
        mydict['process_pk'].append(process_node.pk)
        mydict['proc_exit_status'].append(process_node.exit_status)
        mydict['calc_pk'].append(calc_node.pk)
        mydict['calc_exit_status'].append(calc_node.exit_status)
        mydict['struct_pk'].append(process_node.inputs.structure.pk)
        mydict['formula'].append(process_node.inputs.structure.get_formula())
        mydict['nb_of_atoms'].append(len(process_node.inputs.structure.sites))
        try:
            mydict['remote_path'].append(calc_node.outputs.remote_folder.get_remote_path())
        except Exception as e:
            mydict['remote_path'].append('N.A. ({})'.format(e))

    dataframe = pd.DataFrame(mydict)
    if verbosity >= 1:
        print(dataframe)

    return dataframe


