"""
Multi-step structure-relaxation workflow to be re-used for any structure

Authors:
Sylvian Cadars (1), Olivier Masson (1), Assil Bouzid (1)

Affiliations:
(1) Institute for Research in Ceramics (IRCER), University of Limoges, CNRS

Please cite:
S. Cadars, O. Masson, J.-P. laval, F.E. Shuaib, G. Delaizir, A. Bouzid, A. Piarristeguy
Insights into the structural complexity and local disorder of crystalline
AsTe3 from Semi-Automated First-Principles Modelling
Phys. Chem. Chem. Phys. 2024
DOI: https://doi.org/10.1039/D4CP02941K

This is directly adapted from :
https://aiida-vasp.readthedocs.io/en/latest/workchains/writing_workchains.html

In fact the workchain written therein, vased on vasp.vasp worchain rather than vasp.relax
may in fact be more general.
It is not yet clear to me what these are
see :
    help(WorkflowFactory('vasp.master'))
    help(WorkflowFactory('vasp.vasp'))
    help(WorkflowFactory('vasp.relax')) # should be used instead of vasp.vasp to
                                        # relax a structucre within a workflow
It seems that the vasp.relax workchain includes relaxation-specific inputs.

Because _relax_workchain is based on vasp.relax inputs contains
inputs.parameters.relax and all keywords therein :
inputs.parameters.relax.perform, algo, energy_cutoff, force_cutoff, steps,
positions, shape, volume, convergence_on, convergence_absolute (use asolute
rather than relative units for volume and position convergence),
convergence_max_iterations, convergence_volume, convergence_positions,
convergence_shape_lengths, convergence_shape_angles.

All these paramrters are intended to be code-independent.

More info on the vasp.relax workchain class definition :
https://github.com/aiida-vasp/aiida-vasp/blob/develop/aiida_vasp/workchains/relax.py

For an example of workflow using the relaxation specific parameters :
see https://github.com/aiida-vasp/aiida-vasp/blob/develop/examples/run_relax.py

"""
from aiida.engine import WorkChain, append_, while_, calcfunction
from aiida.plugins import DataFactory, WorkflowFactory
from aiida.orm import KpointsData, Dict, Str, Bool, Float, Int, load_node
from aiida.orm import WorkChainNode, QueryBuilder, CalcJobNode
from aiida.common.extendeddicts import AttributeDict
from aiida.common.exceptions import NotExistent, NotExistentAttributeError, NotExistentKeyError
from aiida_vasp.utils.workchains import compose_exit_code
from aiida.orm.utils import load_node

import re
import numpy as np
import pandas as pd

def _convert_value(value):
    """
    modify string value to appropriate python type and/or value

    Help function to get_multistep_input_param_from_csv
    """
    if str(value).lower() == 'false' :
        value = False
    elif str(value).lower() == 'true' :
        value = True
    elif str(value).lower() == 'none' :
        value = None
    elif  str(value).isnumeric() :
        try :
            value = int(value)
        except ValueError :
            value = float(value)

    return value

def get_multistep_input_param_from_csv(csv_file_name) :
    """
    Create multistep_input_param dict input to vaspMultistepelaxWorkchain from a csv file

    Parameters :
        csv_file_name : name of csv file containing the input parameters of the different
                        relaxation steps

    Returns :
        multistep_input_param : dict to be used as input for the VaspMultistepRelaxWorkchain

    The csv file must be constructed as follows, for an example with two steps :
    one relaxation step and one energy calculation.
    There may be as many steps (and therefore columns with label 'step_x') as needed.
    Parameters (incar or relax) appearing in one step shouls appear in all, with
    'none' (case insensitive) as value if they should not be specified in the
    inputs of the corresponding relaxation step.

    namespace_0,namespace_1,step_0,step_1
    description,None,'Geometry optimization - free volume','Single-point energy - accurate'
    incar,istart,0,0
    incar,icharg,2,2
    incar,encut,200,240
    incar,ediff,1.00E-04,5.00E-05
    incar,ismear,0,5
    incar,sigma,0.1,None
    relax,perform,TRUE,FALSE
    relax,algo,cg,None
    relax,energy_cutoff,1.00E-03,None
    relax,steps,100,None
    relax,positions,TRUE,None
    relax,shape,TRUE,None
    relax,volume,TRUE,None

    The resulting multistep_input_param dict will be organized as follows :

    multistep_input_param = {
        'step_0' : {
            'description' : 'description of relaxation step 0',
            'incar' : {
                'encut': 200,
                ...
            },
            'relax' : {
                'perform' : True,
                ...
            },
            'kpoints_spacing' : VALUE_IN_INVERSE_ANGSTROM,
            'max_iterations' : # may be set to 1 for all steps but the last 2 (final relax step and SPE)
        },
        'step_1' : {
            'description' : 'description of relaxation step 1',
            'incar' : {
                'encut': 300,
                ...
            },
            'relax' : {
                'perform' : False,
                ...
            },
            'kpoints_spacing' : VALUE_IN_INVERSE_ANGSTROM
            'max_iterations' : 5 # may be set to 1 for all steps but the last 2 (final relax step and SPE)
        },
        ...
    }


    """

    df = pd.read_table(csv_file_name,delimiter=',',header=0,index_col=None)

    [nbOfRows,nbOfCols] = df.shape
    # select incar rows
    description_row = df.loc[df.namespace_0 == 'description']
    # Create a dict of descriptions stripped from redundant "" and/or '' quotes
    descriptions = {col_index-2 :  description_row.iloc[0,col_index].strip("'").strip('"') for col_index in range(2,nbOfCols)}
    incar_rows = df.loc[df.namespace_0 == 'incar']
    [nb_of_incar_rows,_] = incar_rows.shape
    relax_rows = df.loc[df.namespace_0 == 'relax']
    [nb_of_relax_rows,_] = relax_rows.shape
    kpoints_spacing_row = df.loc[df.namespace_0 == 'kpoints_spacing']
    kpoints_spacings = {col_index-2 :  float(kpoints_spacing_row.iloc[0,col_index]) for col_index in range(2,nbOfCols)}
    max_iterations_row = df.loc[df.namespace_0 == 'max_iterations']
    max_iterations = {col_index-2 :  int(max_iterations_row.iloc[0,col_index]) for col_index in range(2,nbOfCols)}
    # TODO (???) add a try and except IndexError ???

    multistep_input_param = {}
    for step_index in range(0,nbOfCols-2) :
        multistep_input_param['step_{}'.format(step_index)] = {
            'description' : descriptions[step_index],
            'incar' : {},
            'relax' : {},
            'kpoints_spacing' : kpoints_spacings[step_index],
            'max_iterations' : max_iterations[step_index]
        }

        for row_index in range(0,nb_of_incar_rows) :
            value = _convert_value(incar_rows.iloc[row_index,step_index+2])
            if value != None :
                multistep_input_param['step_{}'.format(step_index)]['incar'][incar_rows.iloc[row_index,1]] = value

        for row_index in range(0,nb_of_relax_rows) :
            value = _convert_value(relax_rows.iloc[row_index,step_index+2])
            if value != None :
                multistep_input_param['step_{}'.format(step_index)]['relax'][relax_rows.iloc[row_index,1]] = value

    return multistep_input_param

def sort_node_list_by_pk(nodes) :
    sorted_nodes = [nodes[i] for i in np.argsort(np.array([node.pk for node in nodes]))]
    return sorted_nodes

@calcfunction
def create_multistep_summary(**kwargs) :
    """
    Create summary dictionary from data nodes of multiple 'VaspRelaxWorkchain's

    input arguments : kwargs is a dict with structure :
        {link_label : data_node}
        wherein 'link_label' has the following format :
        link_label = 'calc_{calc_index}_misc' for misc outputs of the 'VaspRelaxWorchain's
        link_label = 'calc_{calc_index}_kpoints' for kpoints inputs of the 'VaspRelaxWorchain's
        link_label = 'relax_{relax_index}_calc_{calc_index}_misc' for misc
        outputs of the child 'VaspCalculation' processes.

    Below is an example of how kwargs may be created, sarting from the workchain node
    from which the 'VaspRelaxWorchain's of interest have been called :

    data_nodes = {}
    relax_index = 0

    for child in sort_node_list_by_pk(master_chain_node.called) :
        if child.process_type == 'aiida.workflows:vasp.relax':
            data_nodes[f'relax_{relax_index}_misc'] = child.outputs.misc
            data_nodes[f'relax_{relax_index}_kpoints'] = child.inputs.kpoints
            calc_index = 0
            for grandchild in sort_node_list_by_pk(child.called_descendants) :
                if grandchild.process_type == 'aiida.calculations:vasp.vasp' :
                    data_nodes[f'relax_{relax_index}_calc_{calc_index}_misc'] = grandchild.outputs.misc
                    calc_index += 1
            relax_index += 1
    summary = create_multistep_summary(**data_nodes)

    """
    multistep_summary = {}

    # Creating dicts which will have a structure of the form {relax_index : data}
    # that will be all be sorted in the end
    data_namespaces = [
        'cpu_time_s',
        'electronic_converged',
        'step_final_energy',
        'maximum_force',
        'kpoints_mesh',
        'ionic_steps',
                      ]
    data = {namespace : {} for namespace in data_namespaces}

    current_relax_index = 0
    for link_label, data_node in sorted(kwargs.items()) :

        # Locate misc dict output of 'VaspRelaxWorkchain's
        # and
        if re.match(r'(relax_\d+)\_misc', link_label) !=  None :
            temp = re.findall(r'\d+', link_label)
            [relax_index] = list(map(int, temp))
            # extract parameters avaiable therein
            try :
                data['electronic_converged'][relax_index] = data_node['run_status']['electronic_converged']
                data['step_final_energy'][relax_index] = data_node['total_energies']['energy_extrapolated']
                data['maximum_force'][relax_index] = data_node['maximum_force']
            except KeyError :
                data['electronic_converged'][relax_index] = False
                # Cannot use np.nan because it cannot be json-serializable in database
                data['step_final_energy'][relax_index] = None
                data['maximum_force'][relax_index] = None

        elif re.match(r'(relax_\d+)\_kpoints', link_label) !=  None :
            temp = re.findall(r'\d+', link_label)
            [relax_index] = list(map(int, temp))
            data['kpoints_mesh'][relax_index] = data_node.get_kpoints_mesh()[0]

        elif re.match(r'(relax_\d+)(_calc_\d+)\_misc', link_label) !=  None :
            # (aiida-vasp > 1.2.1) Read directly number of steps from :

            temp = re.findall(r'\d+', link_label)
            [relax_index,calc_index] = list(map(int, temp))
            # Initialize cpu_time_s and ionic_steps for every relaxation step
            if relax_index not in data['cpu_time_s'].keys() :
                data['cpu_time_s'][relax_index] = 0
            if relax_index not in data['ionic_steps'].keys() :
                data['ionic_steps'][relax_index] = 0
            # Increment cpu_time_s and ionic_steps for every calculation in one relax step
            # Read number of steps from outputs.misc["run_status"]["last_iteration_index"][0] (aiida-vasp > 1.2.1)
            try :
                # Will only work for processes run with aiida-vasp > 1.2.1
                data['ionic_steps'][relax_index] += data_node['run_status']['last_iteration_index'][0]
            except KeyError : # energies passed as a dict rather than an array
                pass
            try :
                data['cpu_time_s'][relax_index] += data_node['run_stats']['total_cpu_time_used']
            except KeyError :
                pass
            try :
                # Check that ionic convergence is achived in at least one of the child
                # VaspCalculation's of the considered VaspRelaxWorkchain step
                if data_node['run_status']['ionic_converged'] :
                    data['ionic_converged'][relax_index] = True
            except KeyError :
                pass

    for namespace in data_namespaces :
        multistep_summary[namespace] = [data[namespace][key] for key in sorted(data[namespace].keys())]
    # print('multistep_summary = ',multistep_summary)
    return Dict(dict=multistep_summary)


def get_vasp_calc_descendants(relax_node_id, order_by_dict={'ctime': 'asc'}):
    """
    Get vasp descendants of a VaspMultiStepRelaxWorkChain node using a DB query

    Results may be ordered with argument order_by_dict, which uses the same syntax
    as the QueryBuilder().order_by() method.

    TODO : In principle this function may be simplified by a with_descendant=''
    keyword argument (thus sequential queries within neted workflows and jobs),
    but this does not seem to work...

    Args:
        relax_node_id : int
            VaspMultiStepRelaxWorkChain

        order_by_dict : dict
            dict describing the query result ordering as in the QueryBuilder.order_by()
            method.

    returns:
        a list of child VaspCalculation CalcJobNodes
    """
    qb = QueryBuilder()
    qb.append(cls=WorkChainNode,
               filters={
                   'id': relax_node_id
               },
               tag='multi_step_relax_nodes'
            )
    qb.append(cls=WorkChainNode, with_incoming=qb.get_used_tags()[0],
               filters={
                   'attributes.process_label': 'RelaxWorkChain'
               },
               tag='relax_nodes'
            )
    qb.append(cls=WorkChainNode, with_incoming=qb.get_used_tags()[1],
               filters={
                   'attributes.process_label': 'VerifyWorkChain'
               },
               tag='verify_nodes'
            )
    qb.append(cls=WorkChainNode, with_incoming=qb.get_used_tags()[3],
               filters={
                   'attributes.process_label': 'VaspWorkChain'
               },
               tag='vasp_workchain_nodes'
            )
    qb.append(cls=CalcJobNode, with_incoming=qb.get_used_tags()[5],
               filters={
                   'attributes.process_label': 'VaspCalculation',
               },
               tag='vasp_calculations')
    qb.order_by({CalcJobNode: order_by_dict})
    return qb.all(flat=True)


def get_final_energy(relax_node_id, verbose=False):
    """
    Get the final exnergy associated with a vaspMultiStepRelaxWorkchain node

    Return and print associated exception if the final energy cannot be found.
    """
    final_energy = None
    multistep_relax_node = load_node(relax_node_id)
    try:
        final_energy = multistep_relax_node.outputs.misc['total_energies'][
            'energy_extrapolated']
    except Exception as e:
        print('{}<{}>: {}'.format(multistep_relax_node.process_label,
                                  multistep_relax_node.pk, e))
    if verbose:
        print('{}<{}>: Final energy : {} eV'.format(
              multistep_relax_node. process_label,multistep_relax_node.pk,
              final_energy))

    return final_energy


def get_total_nb_of_ionic_steps(relax_node_id, verbose=False):
    """
    Get the total number of ionic steps performed in a
    vaspMultiStepRelaxWorkchain process

    Return and print associated exception if the total number of ionic steps
    cannot be found in the multistep_summary output.
    """
    total_nb_of_ionic_steps = None
    multistep_relax_node = load_node(relax_node_id)
    try:
        total_nb_of_ionic_steps = sum(
            multistep_relax_node.outputs.multistep_summary['ionic_steps'])
    except Exception as e:
        print('{}<{}>: {}'.format(multistep_relax_node.process_label,
                                  multistep_relax_node.pk, e))
    if verbose:
        print('{}<{}>: Total number of ionic steps performed : {}'.format(
              multistep_relax_node. process_label,multistep_relax_node.pk,
              total_nb_of_ionic_steps))

    return total_nb_of_ionic_steps

def get_penultimate_step_max_force(relax_node_id, step_index=-2,
                                   verbose=False):
    """
    Get maximum force from penultimate (or other) step of a mutli-step relax process

    The maximum force of the required step (by default the penultimate) is
    first searched in the multistep_summary Dict output. If the value is None
    (e.g. because relaxation was not complete) the value will be obtained
    a database request to successive child processes called by the
    VaspMultistepRelaxWorkChain.

    Args:
        relax_node_id: int.
            PK number of the relaxation process.
            TODO : allow the use of uuid.

        step_index : int. default is -2
            The default -2 values cooresponds to the penultimate step and
            assumes that the final step is a single-point energy that
            should be ignored from the standpoint of forces.

    Returns:
        The maximum force value as a float (or None)
    """

    max_force = None
    multistep_relax_node = load_node(relax_node_id)
    max_force = multistep_relax_node.outputs.multistep_summary[
        'maximum_force'][step_index]
    if max_force is not None:
        if verbose:
            print(('Max force  = {:.3f} eV/Angstrom (obtained from' +
                   'multistep_summary Dict output {}).').format(
                        max_force, multistep_relax_node.outputs.multistep_summary.pk))
    else:
        vasp_calc_descendants = get_vasp_calc_descendants(relax_node_id,
            order_by_dict={'ctime': 'asc'})
        calc_node = vasp_calc_descendants[step_index]
        try:
            max_force = calc_node.outputs.misc['maximum_force']
        except Exception as ex:
            print(('{}<{}>: Exception of type {} occured with message :\n' +
                  '{}.').format(calc_node.process_label, calc_node.pk,
                                type(ex).__name__, ex))
        if verbose:
            print(('Max force  = {:.3f} eV/Angstrom (obtained from ' +
                   'penultimate {}<{}>).').format(
                   max_force, calc_node.process_label, calc_node.pk))

    return max_force


def print_descendant_vasp_calc_statuses(relax_node_id, print_relax_workchains=True,
                                        print_force=True):
    """
    Add description here
    """
    for descendant in get_vasp_calc_descendants(relax_node_id):
        exit_message = ''
        if not descendant.exit_status:
            exit_message = ' - ' + descendant.exit_message
            print('{}<{}> exited with status {}{}'.format(
                descendant.process_label, descendant.pk,
                descendant.exit_status, exit_message))


class VaspMultiStepRelaxWorkchain(WorkChain):

    # Here we use the vasp.relax workchain in all cases and simply
    # set spec.inputs.relax.perform = False for the final SPE calculation.
    # See spec.outline definition in :
    # https://github.com/aiida-vasp/aiida-vasp/blob/develop/aiida_vasp/workchains/relax.py
    _next_workchain = WorkflowFactory('vasp.relax') # '_' prefix means internal use
    # Initialize verbosity
    _verbose = False

    @classmethod
    def define(cls,spec):
        super(VaspMultiStepRelaxWorkchain,cls).define(spec)
        # Exclude all input parameters that will be defined externally :
        # this inludes all parameters defined in the incar file equivalent that
        # shall be set in an input_dict containing the details of each step.
        # TODO : decide whether one should exclude vasp.relax worchain relax input
        # Exclude here all input parameters that are not used because they are defined
        # for each relax step based on inputs.multistep_input_param
        # One needs to understand the diffference between inputs.verify.max_iterations
        # and inputs.max_iterations

        # TODO: exclude kpoints as it is set from the multistep_param
        spec.expose_inputs(cls._next_workchain,exclude=[
            'structure','parameters'
        ])
        # There may be a conceptual (and data provenance) problem here because I do not use the initial
        # relax and kpoints parameters of the VaspRelax workchain but define them later in
        # init_next_workchain based on multistep_input_param
        # However if I exclude them from the expose_inputs I do not know how to expose them later
        # because the workchain instance does not have a self.expose_inputs()
        """
        spec.expose_inputs(cls._next_workchain,exclude=[
            'structure','parameters','max_iterations','kpoints','relax'])
        """
        spec.input('structure',valid_type=(DataFactory('structure'), DataFactory('cif')))

        # WARNING : multistep_input_param is currently a list of dicts (each unstored).
        # It could alternatively be a list of (potentially-storable) AiiDA Dict. The
        # second option seems to add unnecessary complexity (parameters will be saved
        # as Dict nodes at every step of the workflow) and could be slow.
        spec.input('multistep_input_param',valid_type=(DataFactory('dict')))

        spec.outline(
            cls.initialize,
            while_(cls.run_next_workchains) (
                cls.init_next_workchain,
                cls.run_next_workchain,
                cls.verify_next_workchain,
                cls.update_structure,
                cls.increment_step_index
            ),
            # The final (SPE) step specificity is well accounted for by vasp.relax
            cls.results,
            cls.store_relaxed,
            cls.get_multistep_summary,
            cls.finalize
        )

        # The final step may include a DOS calculation, in which case the vasp.vasp
        # outputs are probably more appropriate. On the other had the DOS may also be
        # a subsequent step within in a higher-level workflow.

        spec.expose_outputs(cls._next_workchain) # already includes relax.structure
        # This does not work : spec.output('multistep_summary',valid_type=(DataFactory('Dict')))
        spec.output('multistep_summary',valid_type=Dict)

        spec.exit_code(0, 'NO_ERROR', message='the sun is shining')
        spec.exit_code(500, 'ERROR_UNKNOWN', message='unknown error detected in the vaspMultiStepRelax workchain')
        spec.exit_code(420, 'ERROR_NO_CALLED_WORKCHAIN',
                       message='no called workchain detected')
        spec.exit_code(501,'ERROR_UNKNOWN_RELAX_INPUT_PARAMETER',
                      message='Unkown relax input paremeter in multistep_input_param')
        spec.exit_code(506, 'ERROR_SAME_INPUT_AND_FINAL_STRUCTURE',
                       message='final relaxed and initial structures are identical.')
        # TODO : define exit_status codes (inspired (and yet different ?) from vasp.relax
        # step_number will be defined and pass from one _nect_workchain to the next in spec.ctx

    def initialize(self):
        """
        Here we create the inputs required to run the 'vasp.relax' workchain
        and store them in the context. No need to redo what is already done vasp.relax
        """

        # Initialize inputs and context
        self.ctx.inputs = AttributeDict()
        self.ctx.inputs.update(self.exposed_inputs(self._next_workchain))
        try:
            self._verbose = self.inputs.verbose.value
            self.ctx.inputs.verbose = self.inputs.verbose
        except AttributeError:
            pass

        if self._verbose :
            self.report('Initializing {} version 2021-08-30.01.'.format(self.__class__.__name__))

        # Initialize settings : this where one can decide what information should be parsed.
        # See : https://aiida-vasp.readthedocs.io/en/latest/concepts/parsing.html
        # No need to parse energies to count the number of ionic steps in individual 'VaspCalculation's
        # since aiida-vasp v. 1.2.1
        # No need to specify the parsing of the structure. This is taken care of by the
        # VaspRelaxWorkchain
        """
        dict_entry = {'add_energies' : True}
        if 'settings' in self.inputs:
            self.report('settings found in inputs : {}'.format(self.inputs.settings.get_dict()))
            settings = AttributeDict(self.inputs.settings.get_dict())
            try:
                settings.parser_settings.update(dict_entry)
            except AttributeError:
                settings.parser_settings = dict_entry
        else:
            settings = AttributeDict({'parser_settings': dict_entry})
        # Note : VaspRelaxWorkchain excepts if self.ctx.inputs.settings is defined as an
        # AttributeDict rather than an AiiDA data dict. This constrasts with the
        # settings initialization given within the vasp.relax or vasp.converge workchains
        self.ctx.inputs.settings = Dict(dict = settings)
        self.report('Settings updated in context  : {}'.format(self.ctx.inputs.settings.get_dict()))
        """

        # Initialize structure : should be re re-done because it is excluded in definition
        self.ctx.current_structure = self.inputs.structure

        # Initialize step_index in context
        self.ctx.step_index = 0

        # Initialize ctx.multistep_summary : a dict where all values will be lists
        # wherein each element will be associated with a relaxation step index.
        # An AiiDA Dict would be stored in the database at the end of the initialize
        # method and therefore become immutable. It seems that a standard dict should
        # used instead (see aiida-vasp "Equilibrium volume of FCC Si, take 2")
        self.ctx.multistep_summary = {
            'ionic_steps': [],
            'inonic_converged': [],
            'kpoints_mesh': [],
            'electronic_converged': [],
            'maximum_force': [],
            'step_final_energy': [],
            'cpu_time_s': [],
        }
        if self._verbose :
            self.report('multistep_summary initialized in context : {})'.format(
                self.ctx.multistep_summary))

    def run_next_workchains(self):
        """
        While loop condition based on step_index

        Last (SPE) step included in the loop. Its specificity (no relaxation
        performed in principle, although this is not mendatory).
        """
        nb_of_steps = len(list(self.inputs.multistep_input_param.keys()))
        while_loop_condition = bool(self.ctx.step_index < nb_of_steps)
        if while_loop_condition and self._verbose :
            self.report("Relaxation step index {}/{}.".format(
                self.ctx.step_index,nb_of_steps-1))

        return while_loop_condition

    def init_next_workchain(self):
        """
        Initialize next workchain

        The relaxation step index should be present in the context.
        It is important to note that the input.parameters and input.relax have not
        be defined at this point because they should be extracted from
        inputs.multistep_input_param and self.ctx.relax.step_index

        There may be no need to define them elsewhere than in context because only context
        inputs are used to submit the calculation, and these are then stored to the database.
        One must simply ensure that these meet the required format of the vasp.relax workchain.
        """

        # Trying to expose inputs from _next_workchain (RelaxWorkChain)
        # to retrieve inputs that are defined within each step
        #
        # self.ctx.inputs = self.__class__.expose_inputs(self._next_workchain) # DOES NOT WORK

        # Set up input parameters from list multistep_input_param
        # Note that the structure is inputs.parameters['incar'], which takes the content of
        # multistep_input_param[STEP_INDEX]['INCAR'] as value
        incar_dict = {'incar' : self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['incar']}
        parameters = Dict(dict=incar_dict) # already an AiiDA Dict ?
        description = self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['description']
        if self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['relax']['perform'] :
            calc_type = 'relaxation'
        else :
            calc_type = 'single-point energy'
        parameters['incar']['system'] = '{} - step {} - {} - {}'.format(
            self.ctx.current_structure.get_formula(),self.ctx.step_index,calc_type,description)
        self.ctx.inputs.parameters = parameters

        # Initialize inputs.max_iterations
        self.ctx.inputs.max_iterations = Int(
            self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['max_iterations'])
        # TODO : set default values in case these parameters are not defined in multistep_input_param :
        # max_iterations = 5 if this is the last relaxation step
        # max_iteration = 1 otherwise
        #   find steps for wherein relax.perform = True
        #   set max_iterations = 5 in all others
        #   ...
        #

        # Initialize relax parameters
        # Here we need to go through all possible relax input parameters of the vasp.relax
        # workchain and convert the corresponding multistep_input_param in the appropriate
        # aiida data format.

        # Loop through all relax inputs in multistep_input_param
        multistep_relax_items = list(self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['relax'].items())
        #IMPROVE : read type of every available vasp.relax relax input item and
        # set to this type and value. Use the following.
        """
        available_relax_items = self.inputs.relax.items() # values are nodes
        print('Current step relax input items:')
        for key,value in multistep_relax_items :
            print(key,' : ',value)
        print('Available vasp.relax relax items:')
        for key,node in available_relax_items :
            print(key,' : ',node) # this prints the node. use node.value to print value.
        """
        # TODO : test all possible relax keywords of the vasp.relax workchain
        bool_relax_namespaces = ['convergence_absolute',
                                 'convergence_on',
                                 'perform',
                                 'positions',
                                 'shape',
                                 'volume']
        str_relax_namespaces = ['algo']
        float_relax_namespaces = ['energy_cutoff',
                                  'force_cutoff',
                                  'convergence_volume',
                                  'convergence_positions',
                                  'convergence_shape_lengths',
                                  'convergence_shape_angles'
                                 ]
        int_relax_namespaces = ['convergence_max_iterations',
                                'steps']
        for key,value in multistep_relax_items:
            if key.lower() in bool_relax_namespaces:
                self.ctx.inputs.relax[key.lower()] = Bool(value)
            elif key.lower() in str_relax_namespaces:
                self.ctx.inputs.relax[key.lower()] = Str(value)
            elif key.lower() in float_relax_namespaces :
                self.ctx.inputs.relax[key.lower()] = Float(value)
            elif key.lower() in int_relax_namespaces :
                self.ctx.inputs.relax[key.lower()] = Int(value)
            else :
                error_message = 'Unknown relax input parameter : '+key
                # TODO : Raise exception (must derive from BaseException)
        """
        # One needs to verify whether parameters defined in inputs.relax
        # potentially overwrite those defined in input.parameters
        # in which case a specific keyword in multi_step_input_param should be set.
        # Another alternative is to create an array for each with increased risk
        # of errors.
        """

        # TODO : Load the structure from external inputs (step 0) or from
        # the previous step.
        self.ctx.inputs.structure = self.ctx.current_structure
        self.set_kpoints_mesh_from_spacing(
            self.inputs.multistep_input_param['step_{}'.format(self.ctx.step_index)]['kpoints_spacing'])

    def set_kpoints_mesh_from_spacing(self,kpts_spacing):
        """
        Set k-point mesh based on self.ctx.current_structure and kpts_spacing in A^⁻1
        """
        kpts_data = KpointsData()
        kpts_data.set_cell_from_structure(self.ctx.current_structure)
        kpts_data.set_kpoints_mesh_from_density(kpts_spacing)
        self.ctx.inputs.kpoints = kpts_data

    def run_next_workchain(self):
        running = self.submit(self._next_workchain, **self.ctx.inputs)

        # No need to check convergence. Achieved in vasp.relax workchain.

        # return copied from vasp.relax workchain
        return self.to_context(workchains=append_(running))

    def verify_next_workchain(self):
        """
        Verify and inherit exit status from child workchains.

        Copied from vasp.relax workchain.
        """

        try:
            workchain = self.ctx.workchains[-1]
        except IndexError:
            self.report('There is no {} in the called workchain list.'.format(self._next_workchain.__name__))
            return self.exit_codes.ERROR_NO_CALLED_WORKCHAIN  # pylint: disable=no-member

        # Inherit exit status from last workchain (supposed to be
        # successfull)
        next_workchain_exit_status = workchain.exit_status
        next_workchain_exit_message = workchain.exit_message
        if not next_workchain_exit_status:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
        else:
            self.ctx.exit_code = compose_exit_code(next_workchain_exit_status, next_workchain_exit_message)
            self.report('The called {}<{}> returned a non-zero exit status. '
                        'The exit status {} is inherited'.format(workchain.__class__.__name__, workchain.pk, self.ctx.exit_code))
            # Make sure at the very minimum we attach the misc node (if it exists) that contains notifications and other
            # quantities that can be salvaged
            try:
                self.out('misc', workchain.outputs['misc'])
            except NotExistent:
                pass

    def update_structure(self) :
        """
        update ctx.current_structure based on latest-run output
        """
        workchain = self.ctx.workchains[-1]

        # For debugging purposes : print relaxWorkchain outputs at this stage of the process
        """
        self.report("Listing {}<{}> outputs :".format(workchain.__class__.__name__,workchain.pk))
        for output in workchain.outputs :
                self.report("\t- {}.".format(output))
        """
        if self.ctx.inputs.relax.perform :
            relaxed_structure = None
            structure_origin = ''
            try :
                relaxed_structure = self.ctx.outputs['relax']['structure']
                structure_provenance = 'relax.structure output of {}<{}>'.format(
                    workchain.process_label,workchain.pk)

            except AttributeError :
                self.report("No relax output structure found for {}<{}>. Possibly "
                            "because the final single-point is being performed.".format(
                                workchain.process_label,workchain.pk))
                try : # For AiiDA versions < 1.6.2
                    relaxed_structure = self.ctx.outputs['relax__structure']
                    structure_provenance = 'relax__structure output of {}<{}>'.format(
                        workchain.process_label,workchain.pk)
                except AttributeError :
                    self.report('The {}<{}> for the relaxation run did not have an '
                            'output relax.structure (or relax__structure for AiiDA '
                            'versions older than 1.6.2). Relaxation may have failed '
                            'with exit code zero.'.format(
                                workchain.process_label, workchain.pk))
                    if self.ctx.exit_code.status in [401, 11, 505] or (
                            self.ctx.exit_code.status == 0 and workchain.inputs.relax.perform):
                        # Maximum number of iterations exceeded
                        # No output is generated in this case.
                        # Latest structure retrieved from the latest VaspCalculation process
                        # among the called_descendants
                        if self.ctx.exit_code.status == 0 and workchain.inputs.relax.perform:
                            self.report('No relaxed structure found whereas exit_status is 0 ' +
                                        'and a relaxation was requested. ')
                        self.report('Looking for output structure among VaspCalculation descendants.')
                        # Order descendant by increasing pk
                        (relaxed_structure, structure_provenance) = self._find_output_structure_among_descendants()

            if relaxed_structure != None :
                self.report('Updating current structure in context from {}.'.format(
                    structure_origin))
                self.ctx.previous_structure = self.ctx.current_structure
                self.ctx.current_structure = relaxed_structure
            else :
                self.report('No relaxed structure found in {}<{}> or descendants.'.format(
                    workchain.process_label,workchain.pk))

            # TODO : consider testing that either the maximum number of ionic steps
            # (relax.steps) has been reached or inonic convergence was achieved.
            # Avoid redundance with tests already performed in vasp.relax

    def _find_output_structure_among_descendants(self):
        relaxed_structure = None
        structure_provenance = ''
        workchain = self.ctx.workchains[-1]

        sorted_indexes = np.argsort(np.asarray(
            [descendant.pk for descendant in workchain.called_descendants]))
        sorted_descendants = [workchain.called_descendants[index] for index in sorted_indexes]
        for descendant in sorted_descendants:
            if descendant.process_label == 'VaspCalculation' :
                if 'structure' in list(descendant.outputs._get_keys()) :
                    # TODO: make sure ouput structure is different from input
                    # structure ?
                    self.report("Relaxed structure {} found for "
                                "{} descendant {}.".format(
                                    descendant.outputs.structure.pk,
                                    descendant.process_label,
                                    descendant.pk))
                    relaxed_structure = descendant.outputs.structure
                    structure_provenance = "structure output of {}<{}> descendant of {}<{}>.".format(
                            descendant.process_label,descendant.pk,
                            workchain.process_label,workchain.pk)
        return (relaxed_structure, structure_provenance)


    def increment_step_index(self):
        """ increment multistep_param step_index in context """
        self.ctx.step_index += 1

    def results(self):
        # TODO : see what to do with this...
        """Attach all outputs from next_workchain to this workchain.

        This should include relax.structure in principle but we nevertheless
        add it in store_relaxed to ensure that the relaxed structure is stored
        in outputs even if the latest VaspRelaxWorkchain failed.

        How it is done in vasp.converge :
        workchain = self.ctx.workchains[-1]
        self.out_many({key: workchain.outputs[key] for key in workchain.outputs})

        How it is done in vasp.relax :
        workchain = self.ctx.workchains[-1]
        self.out_many(self.exposed_outputs(workchain, self._next_workchain))

        """
        workchain = self.ctx.workchains[-1]
        # The following excepts with AttributeError: 'VaspMultiStepRelaxWorkchain'
        # object has no attribute 'exposed_ouputs'
        try :
            self.out_many(self.exposed_ouputs(workchain, self._next_workchain))
        except AttributeError :
            if self._verbose :
                self.report("The following outputs from {}<{}> will be stored.".format(
                    workchain.__class__.__name__,workchain.pk))
                for key in workchain.outputs :
                    self.report("\t {} : {}".format(key,workchain.outputs[key]))
            self.out_many({key: workchain.outputs[key] for key in workchain.outputs})

    def store_relaxed(self):
        """
        Store relaxed structure in worchain outputs
        """
        if self.ctx.current_structure.pk == self.inputs.structure:
            self.report(('Final and initial structures are the same: {}.' +
                         'A known origin of this problem is when aiida-vasp ' +
                         ' RelaxWorkChain exits with 0 code and no final ' +
                         ' output structure.').format(self.inputs.structure.pk))
            return self.exit_codes.ERROR_SAME_INPUT_AND_FINAL_STRUCTURE

        self.report("Storing relaxed structure from context current_structure.")
        self.out('relax.structure', self.ctx.current_structure)

    def get_multistep_summary(self):
        """
        Store multistep_summary dict in worchain outputs

        TODO : deal with missing misc
        """

        """
        from aiida.common.exceptions import NotExistentAttributeError

        master_chain_node = load_node(7483) # OK with 4700.
        data_nodes = {}
        relax_index = 0
        for child in master_chain_node.called_descendants :
            if child.process_type == 'aiida.workflows:vasp.relax':
                try :
                    data_nodes[f'relax_{relax_index}_misc'] = child.outputs.misc
                except NotExistentAttributeError :
                    print('NotExistentAttributeError exception.')
                    data_nodes[f'relax_{relax_index}_misc'] = {}
        """

        # Create a dict of {link_label : data_node} to use as keyword arguments to calcfunction multistep_summary
        # All data_node inputs will be passed along with a link label taht will show in the provenance graph.
        data_nodes = {}
        relax_index = 0
        for child in self.ctx.workchains :
            if child.process_type == 'aiida.workflows:vasp.relax':
                try :
                    data_nodes[f'relax_{relax_index}_misc'] = child.outputs.misc
                except NotExistentAttributeError :
                    if self._verbose :
                        self.report('Missing misc output to {}<{}>'.format(
                            child.__class__.__name__,child.pk))
                        self.report('Corresponding data in multistep_summary will be set to None.')
                    data_nodes[f'relax_{relax_index}_misc'] = {}

                data_nodes[f'relax_{relax_index}_kpoints'] = child.inputs.kpoints
                calc_index = 0
                for grandchild in sort_node_list_by_pk(child.called_descendants) :
                    if grandchild.process_type == 'aiida.calculations:vasp.vasp' :
                        try :
                            data_nodes[f'relax_{relax_index}_calc_{calc_index}_misc'] = grandchild.outputs.misc
                        except NotExistentAttributeError :
                            if self._verbose :
                                self.report('Missing misc output to {}<{}>.'.format(
                                    grandchild.__class__.__name__,grandchild.pk))
                                self.report('Corresponding data in multistep_summary will be set to None.')
                            data_nodes[f'relax_{relax_index}_calc_{calc_index}_misc'] = Dict(dict={})
                        calc_index += 1
                relax_index += 1
        if self._verbose :
            self.report('Using the following data_nodes dict will be used as keyword arguments for calcfunction create_multistep_summary :')
            self.report('   data_nodes = {}'.format(data_nodes))
        multistep_summary = create_multistep_summary(**data_nodes)

        # Provenance is stored because multistep_summary is an AiiDA dict created within a calcfunction
        self.out('multistep_summary',multistep_summary)

        # write a report
        self.report("{} stored in ouputs : {}<{}>.".format(
            'multistep_summary',
            self.outputs['multistep_summary'].__class__.__name__,
            self.outputs['multistep_summary'].pk))
        self.report('Its content is the following :')
        for key in self.outputs['multistep_summary'].attributes:
            self.report('{} : {}'.format(key,self.outputs['multistep_summary'][key]))

    def finalize(self):
        """ Finalize the workchain """
        self.report("Finalizing VaspMultiStepRelaxWorkchain.")
