"""
Utility functions specific to the AsTe3 System

Authors:
Sylvian Cadars (1), Olivier Masson (1), Assil Bouzid (1)

Affiliations:
(1) Institute for Research in Ceramics (IRCER), University of Limoges, CNRS

Please cite:
S. Cadars, O. Masson, J.-P. laval, F.E. Shuaib, G. Delaizir, A. Bouzid, A. Piarristeguy
Insights into the structural complexity and local disorder of crystalline
AsTe3 from Semi-Automated First-Principles Modelling
Phys. Chem. Chem. Phys. 2024
DOI: https://doi.org/10.1039/D4CP02941K

Compatibility:
    aiida-core 1.6.5
    aiida-vasp

Usage :
    In this example this project_utils.py file would be placed in directory:
    $HOME/aiida_projects/MY_PROJECT/

    Add PATH_TO_THIS_FILE to PYTHONPATH
    e.g. for a conda install:
    Edit file:
    $CONDA_PREFIX/etc/conda/activate.d/env_vars.sh
    and add the follong line:
    export PYTHON_PATH=$YTHON_PATH:$HOME/aiida/projects

    Source the file above or deactivate and reactivate your cnda environment.

    Then in your program or workflow use:
    import MY_PROJECT/project_utils as pu

    pu.print_versions()  # to display the versions of the modules used here.

"""

from aiida import load_profile
from aiida.orm import CalcJobNode, Dict, QueryBuilder, Group, WorkChainNode
from aiida.orm import StructureData, WorkFunctionNode, CalcFunctionNode
from aiida.orm import SinglefileData, Code, KpointsData, Str, Int, Bool, Float, List
from aiida.plugins import DataFactory, CalculationFactory

from aiida.engine import submit
from aiida.engine import workfunction, calcfunction, ExitCode, run_get_node
from aiida.orm.utils import load_node, load_group
from aiida.common import NotExistent, AttributeDict, NotExistentAttributeError, \
                         IntegrityError
from aiida.tools.data.array.kpoints import get_explicit_kpoints_path

from vaspMultiStepRelaxWorkchainModule import get_penultimate_step_max_force , \
    VaspMultiStepRelaxWorkchain, get_multistep_input_param_from_csv

from aiida_vasp.workchains.moleculardynamics import MolecularDynamicsWorkChain
from dosAndBands import DosAndBandsWorkChain

from pyama.structureComparisonsPkg import distanceTools as dt

import numpy as np
import sys
import time
import os
import random
import pandas as pd
import fileinput
import json
from matplotlib import pyplot as plt
from matplotlib.colorbar import Colorbar
from matplotlib.collections import PathCollection
from ase.visualize import view
from pymatgen.core import Structure, Species
from pymatgen.analysis.diffraction.neutron import NDCalculator
from pymatgen.analysis.diffraction. xrd import XRDCalculator
from pyama.diffractionPkg.nanopdf import nanopdfData

load_profile()

__version__ = '2022.12.07.1'

def get_version():
    """
    Use a version number to allow reproducibility
    """
    # UPDATE THIS LINE WHENEVER CHANGES ARE MADE. USE FORMAT 'YYYY.MM.DD_X'
    return __version__

def get_versions():
    """
    Get version of modules used here as a dict
    """
    import aiida
    import aiida_vasp
    import aiida_supercell
    import vaspMultiStepRelaxWorkchainModule
    versions_dict = {
        'aiida-core': aiida.get_version(),
        'aiida-vasp': aiida_vasp.__version__,
        'aiida-supercell': aiida_supercell.__version__,
        '{}'.format(os.path.basename(os.path.splitext(__file__)[0])): get_version(),
        '{}_location'.format(os.path.basename(os.path.splitext(__file__)[0])): __file__,
        'vaspMultiStepRelaxWorkchainModule_location': vaspMultiStepRelaxWorkchainModule.__file__
    }
    return versions_dict

def print_versions():
    versions_dict = get_versions()
    for k,v in versions_dict.items():
        print('{}: {}'.format(k, v))

@workfunction
def enum_As_Te_in_181_supercell(supercell_code: Code,
                                structure: StructureData,
                                test_mode: Bool):
    """
    Prepare builder and run supercell calculation on initial structure.

    Args:
        supercell_code : Code
            Supercell code object.
        structure :
            initial structure with disorder

    Returns:
        calc_outputs : outputs of the SupercellCalculation process

    TODO :
        Associate link_labels to calcfunction inputs
        This may be done at function call by setting **kwargs as parameters
        and calling function with :
        function_name(link_label_1=)

        Consider specifying user parameters outside this function
        with additional arguments such as SUPERCELL_OPTIONS, etc.
    """
    SupercellCalculation = CalculationFactory('supercell')

    builder = SupercellCalculation.get_builder()
    builder.code = supercell_code
    builder.structure = structure
    builder.charges = Dict(dict={'As*': 0, 'Te*': 0})
    builder.calculate_coulomb_energies = Bool(False)
    builder.charge_balance_method = Str('no')
    builder.merge_symmetric = Bool(True)
    builder.supercell_size = List(list=[1, 8, 1])
    builder.metadata.options.resources = {  # pylint: disable = no-member
        'num_machines': 1,
        'num_mpiprocs_per_machine': 1,
    }
    builder.save_as_archive = Bool(False)
    builder.metadata.options.max_wallclock_seconds = 1 * 30 * 60

    # Uncomment to sample specific structures rather than all
    if test_mode.value:
        builder.sample_structures = Dict(dict={'random': 5})
    else:
        builder.sample_structures = Dict(dict={})  # generate all structures

    """
    WARNING : submit returns the calc_node wheres run returns a Dict of
    output nodes.
    calc_node = submit(builder)
    """
    # Use run rather than submit to ensure that job finishes before return
    calc_outputs, calc_node = run_get_node(builder)
    if calc_node.exit_status:
        print('{}<{}>: process exited with code {}: {}', calc_node.process_label,
              calc_node.pk, calc_node.exit_status, calc_node.exit_message)

    return calc_outputs


def generate_Te_As_configurations(supercell_code_name='supercell@localhost',
                                  test_mode=False,
                                  supercell_struct_group_name=None):
    # Setting supercell_struct_group_name to default value depending on
    # wether test_mode is True or False
    if supercell_struct_group_name is None:
        if test_mode:
            supercell_struct_group_name='test_group'
        else:
            supercell_struct_group_name='AsTe3_Te-based_sc181_As-Te-configurations_no-opt'

    """
    Prepare supercell calculation and group resulting structures

    TODO: Set default supercell_struct_group_name to an appropriate value
    """
    step1_init_struct_file = SinglefileData(file=os.path.join(os.getcwd(),
                                                              'AsTe3_Te-metal-cell.cif'))
    supercell_code = Code.get_from_string(supercell_code_name)

    print(('Running function generate_Te_As_configurations() with group '
           'label \'{}\' and code {}.').format(supercell_struct_group_name,
           supercell_code))

    step1_calc_outputs = enum_As_Te_in_181_supercell(supercell_code,
                                                     step1_init_struct_file,
                                                     test_mode=Bool(test_mode))
    step1_calc_node = step1_calc_outputs['output_parameters'
        ].get_incoming().first()[0]
    exit_message = ''
    if step1_calc_node.exit_status:
        exit_message = ' : '+step1_calc_node.exit_message
    print('{}<{}> exit_status : {}{}'.format(step1_calc_node.process_label,
                                             step1_calc_node.pk,
                                             step1_calc_node.exit_status,
                                             exit_message))
    # Create a group containing all supercell-step-1 output structures
    try:
        group = Group(label=supercell_struct_group_name)
        group.store()
    except IntegrityError:
        print('Group with label \'{}\' already exists. Clearing nodes.'.format(
            supercell_struct_group_name))
        group = load_group(label=supercell_struct_group_name)
        group.clear()

    # TODO : assign all step1_output_struct to a group.
    group.add_nodes(list(step1_calc_node.outputs.output_structures.values()))
    # Loop through 1st-supercell-step output structures to introduce
    # Te(0)/Te(2-) disorder on the (now fixed) Te sites and find the lowest-
    # Coulomb-energy As(3+)(fixed)/Te(0)/Te(2-) configuration.

    print('{} structures generated with {}<{}> and stored in group {}<{}>'.format(
        len(list(step1_calc_node.outputs.output_structures.values())),
        step1_calc_node.process_label, step1_calc_node.pk,
        group.label, group.pk,))

    return group

@workfunction
def enum_Te0_Te2m_in_supercell(supercell_code: Code, structure: StructureData) :
    """
    Distribute Te(0)/Te(2-) in a structure containing Te atoms and calculated electrostatic energies with aiida-supercell plugin

    Args:
        supercell_code: Code
            supercell code object
        structure: StructureData
            AiiDA structure object
    """
    SupercellCalculation = CalculationFactory('supercell')
    builder = SupercellCalculation.get_builder()
    builder.code = supercell_code
    builder.structure = structure
    # charges are set previously site by site within the CIF file
    # builder.charges = Dict(dict={'As*': 3, 'Te0+': 0, 'Te2-':-2})
    builder.calculate_coulomb_energies = Bool(True)
    builder.charge_balance_method = Str('try')
    builder.merge_symmetric = Bool(True)
    builder.supercell_size = List(list=[1, 1, 1])
    builder.metadata.options.resources = {  #pylint: disable = no-member
        'num_machines': 1,
        'num_mpiprocs_per_machine': 1,
    }
    builder.save_as_archive = Bool(False)
    builder.metadata.options.max_wallclock_seconds = 1 * 30 * 60

    # Extract only the lowest_energy strutcure
    builder.sample_structures = Dict(dict={
        'low_energy': 1,
    })

    # Use run rather than submit to ensure that job finishes before return
    calc_outputs, calc_node = run_get_node(builder)
    if calc_node.exit_status:
        print('{}<{}> with input structure {} exited with code {}: {}.'.format(
              calc_node.process_label, calc_node.pk, calc_node.inputs.structure.pk,
              calc_node.exit_status, calc_node.exit_message))
    else:
        print('{}<{}> with input structure {} run successfully.'.format(
              calc_node.process_label, calc_node.pk, calc_node.inputs.structure.pk))

    return calc_outputs
# end of function enum_Te0_Te2m_in_supercell

@calcfunction
def introduce_Te_ox_state_disorder(initial_structure : StructureData) :
    """
    Replace all Te sites in initial_structure by Te(0)/Te(2-)

    Use pymatgen to replace all Te sites by two distinct (albeit overlapped)
    sites Te(0)/Te(2-) with 50%/50% occup.

    Returns :
        SinglefileData created from the CIF file of the disordered structure.
    """
    pmg_struct = initial_structure.get_pymatgen_structure()
    new_pmg_struct = pmg_struct.copy()
    for site_index,site in enumerate(pmg_struct.sites) :
        if site.species.elements[0].name == 'Te' :
            new_pmg_struct[site_index] = {'Te2-':0.5},site.frac_coords
            new_pmg_struct.append({Species.from_string('Te0') : 0.5},site.frac_coords)
        elif site.species.elements[0].name == 'As' :
            new_pmg_struct[site_index] = {'As3+':1},site.frac_coords
    new_pmg_struct.sort()
    # Export disordered pymatgen structure as cif
    export_name = os.path.join(os.getcwd(),'tmp.cif')
    new_pmg_struct.to(fmt='cif',filename=export_name)

    # Open cif file to modify site labels
    # (which does not seem to be possible within pymatgen)
    lines=[]
    in_sites_loop = False
    for line in fileinput.input(export_name,inplace=True) :
        if '_atom_site_occupancy' in line :
            in_sites_loop = True
        elif in_sites_loop :
            mylist = line.split()
            if len(mylist) >= 2 :
                if mylist[0] == 'Te0+' :
                    mylist[1] = 'Te1'
                elif mylist[0] == 'Te2-' :
                    mylist[1] = 'Te2'
                elif mylist[0] == 'As3+' :
                    mylist[1] = 'As3'
                line = ' '.join(mylist)+os.linesep
        sys.stdout.write(line)

    # return SinglefileData associated with the cif file.
    return SinglefileData(export_name)

def calculate_lowest_E_C(supercell_code_label='supercell@localhost',
                         test_mode=False, supercell_struct_group_label=None):
    """
    Introduce Te(0)/Te(2-) disorder and calculate electrostatic energies in all
    As/Te configurations

    TODO? define as a workfunction to track process evolution ?
    in this case the function should take supercell_code and
    supercell_struct_group rather than labels as the inputs.

    Args:
        supercell_code_label: str (default is 'supercell@localhost'
            Label of the aiida_supercell code
        test_mode: bool (default is False)
            If true the default supercell_struct_group_label is 'test_group'
        supercell_struct_group_label: str or None (default is None)
            Label of the group containing all structures obtained with
            generate_Te_As_configurations(). Its default value depends on
            test_mode: 'test_group is test_mode is True, and
            'AsTe3_Te-based_sc181_As-Te-configurations_no-opt' otherwise.
    """
    # Set default supercell_struct_group depending on test_mode
    if supercell_struct_group_label is None:
        if test_mode:
            supercell_struct_group_label = 'test_group'
        else:
            supercell_struct_group_label = 'AsTe3_Te-based_sc181_As-Te-configurations_no-opt'
    supercell_code = Code.get_from_string(supercell_code_label)

    # Load structures from group
    qb = QueryBuilder()
    qb.append(Group,
              filters={
                  'label': supercell_struct_group_label,
                  'type_string': 'core'
              },
              tag='structures_group')
    qb.append(StructureData,with_group='structures_group')
    structures = qb.all(flat=True)

    # Loop through 1st-supercell-step output structures to introduce Te(0)/Te(2-)
    # disorder on the (now fixed) Te sites and find the lowest-Coulomb-energy
    # As(3+)(fixed)/Te(0)/Te(2-) configuration.
    step2_calc_nodes = []
    for struct_index, step1_output_struct in enumerate(structures) :

        # TODO : get supercell structure_id from structure
        supercell_id = get_supercell_id_from_structure(step1_output_struct)
        print('\tIntroducing Te(0)/Te(2-) disorder in struct. {}<pk = {}> ({} out of {})'.format(
            supercell_id, step1_output_struct.pk, struct_index, len(structures)-1))
        step2_init_struct_file = introduce_Te_ox_state_disorder(
            step1_output_struct)
        step2_calc_outputs = enum_Te0_Te2m_in_supercell(supercell_code,
                                                        step2_init_struct_file)
        step2_calc_node = \
            step2_calc_outputs['output_parameters'].get_incoming().first()[0]
        step2_calc_nodes.append(step2_calc_node)
        exit_message = ''
        if step2_calc_node.exit_status:
            exit_message = ' : ' + step2_calc_node.exit_message
        print('\t{}<{}> exit_status : {}{}'.format(
            step2_calc_node.process_label, step2_calc_node.pk,
            step2_calc_node.exit_status, exit_message))
        # print the number of combinations
        if not step2_calc_node.exit_status:
            # TODO : add exception if ""symmetrically_distinct" key is not
            # present

            print('Looking for lowest-energy config among {} '.format(
                    step2_calc_node.outputs.output_parameters['Number_of_structures']['symmetrically_distinct'])
                  + 'symmetrycally-distinct combinations.')

        for sc_id, structure in step2_calc_node.outputs.output_structures.items():
            print('\t\t{}<pk = {}> '.format(sc_id,structure.pk))
            (min_Ewald_energy,structure_id,structure_pk) = get_lowest_Te0_Te2m_Coulomb_energy(step2_calc_node)
            print('\tmin_Ewald_energy = {:.4f} eV (structure {}<pk = {}>'.format(
                min_Ewald_energy,structure_id,structure_pk))

    print(('End of function calculate_lowest_E_C. Lowest-Coulomb-energy Te(0)/Te(2-) '
           'configurations have been found for {} structures.').format(
          len(step2_calc_nodes)))


def get_supercell_id_from_edge_label(edge_label):
    supercell_id = ''
    if 'output_structures__' in edge_label:
        supercell_id = edge_label[len('output_structures__'):]
    return supercell_id


def get_structure_pk_from_supercell_id(supercell_calc_node: CalcJobNode,
                                       supercell_id: str):
    """
    Get pk of AiiDA's StructureData output of a SupercellCalcuation

    Parameters:
        supercell_calc_node : CalcJobNode
            The SupercellCalculation process node
        supercell_id : str
            a string of the form 'i<selection_mode>00345' with :
                selection_mode = 'r','l','h', etc. (see supercell manual)
    """
    structure = supercell_calc_node.outputs['output_structures'][supercell_id]
    return structure.pk


def get_supercell_id_from_structure(structure_or_pk):
    """
    Get supercell ID associated with a structure (or its PK)

    The structure
    """
    supercell_id = None
    if isinstance(structure_or_pk, int):
        structure = load_node(structure_or_pk)
    else:
        structure = structure_or_pk

    if not isinstance(structure, StructureData):
        print('WARNING: no structure with pk {} in database.'.format(
            structure.pk))
        return None
    for incoming_link in list(structure.get_incoming()):
        if incoming_link.node.process_label == 'SupercellCalculation':
            supercell_id = get_supercell_id_from_edge_label(
                incoming_link.link_label)
    if supercell_id is None:
        print(('WARNING: structure {} does not seem to originate ' +
               'from a SupercellCalculation process.').format(
                structure.pk))
    return supercell_id


def get_lowest_Te0_Te2m_Coulomb_energy(supercell_calc_node: CalcJobNode):
    """
    Get lowest-coulombic-energy structure from SupercellCalculation process

    Args :
        supercell_calc_node : SupercellCalculation process node

    Returns :
        A tuple of (min_Ewald_energy,supercell_structure_id,structure_pk)
        with :
            min_Ewald_energy : float
                the Ewald energy associated with the optimal repartitions
                of charges in the considered system
            supercell_structure_id : str
                supercel-structure identifier in the form
                'i<selection_mode><index>' with selection_mode = 'l', 'h',
                'r', etc.
            structure_pk : int
                the structure id in AiiDA database
    """

    # Loop through structures in outputs.output_parameters
    output_param = supercell_calc_node.outputs.output_parameters
    energies_dict = {}
    for name, info in output_param['Structures_info'].items():
        if name[1] == 'l':
            energies_dict[name] = info['coulombic_energy']
    energies = list(energies_dict.values())
    min_index = np.argmin(energies)
    supercell_structure_id = list(energies_dict.keys())[min_index]
    structure_pk = get_structure_pk_from_supercell_id(supercell_calc_node,
                                                      supercell_structure_id)
    min_Ewald_energy = energies[min_index]

    return min_Ewald_energy, supercell_structure_id, structure_pk


def print_dict_of_lists_content(data_Dict, max_nb_of_values_to_print=4,
                                prefix=''):
    """
    print data Dict content entirely or in a truncated fashion

    data Dict should be of the form :
    {
        'key_1': list_1,
        'key_2': list_2,
        ...
    }
    wherein all lists should have the same length.
    """
    if isinstance(data_Dict, Dict):
        items = list(data_Dict.attributes_items())
    elif isinstance(data_Dict, dict):
        items = list(data_Dict.items())

    for key, value in items:
        if isinstance(value, list):
            if len(value) <= max_nb_of_values_to_print:
                print(f'{prefix}{key}: {value}')
            else:
                print(f'{prefix}{key}: [{value[0]}, {value[1]}, ... {value[-1]}]')
        else:
            print(f'{prefix}{key}: {value}')
# end of module local_utils


def find_processes_with_exit_status(group_or_label, node_labels=[''],
        process_label='VaspMultiStepRelaxWorkchain', with_direction='incoming',
        searched_exit_status=0, verbosity=1):
    """
    Find processes with certain exit status(es) with structures in a group as inputs or outputs

    Args:
        group_or_label: Group or str
            Instance or label of the group containing the structures used as
            input (with_direction='incoming') or output (with_direction=outgoing)
            of the searched processes.
        with_direction: str (default='incoming')
            whether strucctures in a group should be considered in the search
            as either :
            - 'with_incoming' : in this case outgoing processes (using group
            structures as inputs) will be returned.
            or
            - 'with_outgoing' : in this case incoming processes (having the
            group 's structures as outputs) will be returned.
        searched_exit_status: int or dict
            An integer desgnating a specific exit_status
            or a dict of the form {'in': LIST_OF_EXIT_STATUSES} used as
            a query filter on attributes.exit_status
    """

    # DEBUGGING
    if verbosity >= 2:
        print('with_direction = ', with_direction)

    struct_group = get_group_from_str_or_instance(group_or_label)
    # Search structures in group
    qb = QueryBuilder()
    qb.append(cls=Group, filters={'label': struct_group.label},
              tag='struct_group')
    qb.append(cls=StructureData, with_group='struct_group',
              tag='group_structures')

    if isinstance(node_labels, str):
        node_labels = [node_labels]
    if verbosity >= 1:
        print(f'{qb.count()} structures in group {struct_group.label}.')
    # Search failed processes using group_structures as outputs
    if with_direction == 'outgoing':
        qb.append(cls=WorkChainNode, with_outgoing='group_structures',
                  filters={
                    'attributes.process_label': process_label,
                    'attributes.process_state': 'finished',
                    'attributes.exit_status': searched_exit_status,
                    'label': {'in': node_labels},
                  },
                  tag='incoming_processes',
                  project=['*']
                  )
    elif with_direction == 'incoming':
        qb.append(cls=WorkChainNode, with_incoming='group_structures',
                  filters={
                    'attributes.process_label': process_label,
                    'attributes.process_state': 'finished',
                    'attributes.exit_status': searched_exit_status,
                    'label': {'in': node_labels},
                  },
                  tag='outgoing_processes',
                  project=['*']
                  )
    else:
        sys.exit('with_direction argument should be either \'incoming\' or \'outgoing\'')
    if verbosity >= 1:
        print(('{} processes with exit status {} found with_{} structures ' +
               'in group {}').format(
            qb.count(), searched_exit_status, with_direction, struct_group.label))
    processes = qb.all(flat=True)
    return processes


def relabel_process_if_identical_in_and_out_struct_pk(process_node,
                                                      failed_label='failed',
                                                      verbosity=1,
                                                      test_mode=False):
    """
    Change the label of a process if input and output structures have identical PKs

    Args:
        process_node:
        failed_label: str (default: 'failed')
            Label to be given to the process if input and output structures
            have identical PKs.
        verbosity: int (default=1)
            Verbosity level
        test_mode: bool (default: False)
            if True the label will not be changed. The function will only print
            out that identical input and output structure PKs have been found
            (if verbosity >= 1)

    Returns:
        True if process was relabeled. False otherwise (including if test_mode
        == False)
    """

    is_process_relabeled = False
    # Check that process exit_status is 0 and process_state is 'finished'
    if process_node.exit_status or \
            process_node.process_state.value != 'finished':
        if verbosity >= 2:
            print(('{}<{}>: process unfinished or with non-zero exit ' +
                   'status. Label not changed.').format(
                       process_node.process_label, process_node.pk))
        return is_process_relabeled

    # Change label if output and input structures have identical pks
    if 'structure' in list(process_node.outputs):
        if process_node.outputs.structure.pk == \
                process_node.inputs.structure.pk:
            if not test_mode:
                process_node.label = failed_label
                is_process_relabeled = True
            if verbosity >= 1 or test_mode:
                print(('{}<{}>: input and ouput structure are identical. ' +
                       'Label changed to \'{}\'.').format(
                       process_node.process_label, process_node.pk,
                       failed_label))
    elif 'relax' in list(process_node.outputs):
        if process_node.outputs.relax.structure.pk == \
                process_node.inputs.structure.pk:
            if not test_mode:
                process_node.label = failed_label
                is_process_relabeled = True
            if verbosity >= 1 or test_mode:
                print(('{}<{}>: input structure and ouput relax.structure ' +
                       'are identical. Label changed to \'{}\'.').format(
                       process_node.process_label, process_node.pk,
                       failed_label))
    else:
        if verbosity >= 1:
            print(('{}<{}>: No structure or relax.structure output ' +
                   'found.').format(
                    process_node.process_label, process_node.pk))
    return is_process_relabeled


def relabel_processes_with_identical_in_and_out_struct(group_or_label,
        node_labels=[''], process_label='VaspMultiStepRelaxWorkchain',
        failed_label='failed', include_outgoing=True, include_incoming=True,
        verbosity=1, test_mode=False):
    """
    Search failed processes associated with structures in a group and change their label

    The main goal of this function is to eliminate processes with a 0
    exit_status but whose input and output structures are identical, which
    causes infinite loops in recursive searches within nested processes.

    Args:
        struct_group : Group
            Group containing structures being outputs and/or inputs to the
            searched processes.
        process_label : str (default = 'VaspMultiStepRelaxWorkchain')
            Designation of the processes searched (node.process_label
            attribute)
        node_labels : list
            Listy of node labels to be searched.
        failed_label: str (default='failed')
            Label that will be given to the identified failed process nodes
        include_outgoing: bool (default=True)
            wheteher processes using the group's structures as inputs
            should be included in the search
        include_incoming: bool (default=True)
            wheteher processes using the group's structures as outputs
            should be included in the search
        verbosity: int (default=1)
            Verbosity level

    Returns:
        pks of relabeled_processes
    """
    struct_group = get_group_from_str_or_instance(group_or_label)

    nb_of_relabeled_incoming_processes = 0
    nb_of_relabeled_outgoing_processes = 0
    relabeled_process_pks = []

    searched_exit_status = 0
    if include_incoming:
        # NB : 'with_outgoing' structures means incoming processes
        processes = find_processes_with_exit_status(struct_group,
            node_labels=node_labels, process_label=process_label,
            with_direction='outgoing',
            searched_exit_status=searched_exit_status, verbosity=verbosity)
        if verbosity >= 1:
            print(('{} {} processes with labels {} and exit status(es) {} ' +
                   'have structures in group {} as outputs.').format(
                   len(processes), process_label, node_labels,
                   searched_exit_status, struct_group.label))

        for process_node in processes:
            is_relabeled = relabel_process_if_identical_in_and_out_struct_pk(
                process_node, failed_label=failed_label,
                verbosity=verbosity, test_mode=test_mode)
            if is_relabeled:
                relabeled_process_pks.append(process_node.pk)

        nb_of_relabeled_incoming_processes = len(relabeled_process_pks)
        if verbosity >= 1:
            print('{} incoming processes have been relabeled.'.format(
                  nb_of_relabeled_incoming_processes))

    if include_outgoing:
        # NB : 'with_incoming' structures means outgoing processes
        processes = find_processes_with_exit_status(struct_group,
            node_labels=node_labels, process_label=process_label,
            with_direction='incoming', searched_exit_status=0)
        if verbosity >= 1:
            print(('{} {} processes with labels {} and exit status(es) {} ' +
                   'use structures in group {} as inputs.').format(
                   len(processes), process_label, node_labels,
                   searched_exit_status, struct_group.label))

        for process_node in processes:
            is_relabeled = relabel_process_if_identical_in_and_out_struct_pk(
                process_node, failed_label=failed_label,
                verbosity=verbosity, test_mode=test_mode)
            if is_relabeled:
                relabeled_process_pks.append(process_node.pk)

        nb_of_relabeled_outgoing_processes = len(relabeled_process_pks) - \
            nb_of_relabeled_incoming_processes
        if verbosity >= 1:
            print('{} outgoing processes have been relabeled.'.format(
                  nb_of_relabeled_outgoing_processes))

    return relabeled_process_pks


def retrieve_and_store_lowest_E_C(test_mode=False, supercell_struct_group=None,
                                  data_label='lowest_E_C_data',
                                  data_description=None, verbosity=1):
    """
    Retrieve and store lowest E_C and related data for structures in a group

    This should ideally have been done at the same time as the calculation
    of the Coulomb energies. Here this is done intead separately in a post
    -processing step.
    """
    # Set default supercell_struct_group depending on test_mode
    if supercell_struct_group is None:
        if test_mode:
            supercell_struct_group = 'test_group'
        else:
            supercell_struct_group = 'AsTe3_Te-based_sc181_As-Te-configurations_no-opt'

    supercell_struct_group = get_group_from_str_or_instance(supercell_struct_group)

    # Query all structures in group.
    if verbosity >= 2:
        print('Querying structures from group {}...'.format(
            supercell_struct_group.label))

    qb = QueryBuilder()
    qb.append(cls=Group, filters={'label': supercell_struct_group.label},
              tag='supercell_struct_group')
    qb.append(cls=StructureData, with_group='supercell_struct_group',
              tag='supercell_structures', project=['id'])
    sc_step_1_struct_column = len(qb.first())-1
    if verbosity >= 2:
        print('Group {} contains {} structures.'.format(
            supercell_struct_group.label, qb.count()))

    # Search Te(0)/Te(2-) supercell calculation processes and results
    if verbosity >= 2:
        print(('Querying Te(0)/Te(2-) disordering and enumeration ' +
               'processes from group {}.').format(supercell_struct_group.label))
    qb.append(cls=CalcFunctionNode, with_incoming='supercell_structures',
              filters={
                'attributes.process_label': 'introduce_Te_ox_state_disorder',
                'attributes.process_state': 'finished',
              },
              tag='Te_ox_state_disorder_processes',
              )
    qb.append(cls=CalcJobNode, with_ancestors='Te_ox_state_disorder_processes',
              filters={
                'attributes.process_label': 'SupercellCalculation',
                'attributes.process_state': 'finished'
              },
              tag='Step-2_supercell_processes',
              project=['id']
              )
    sc_step_2_process_column = len(qb.first())-1

    # Initialize a data in dict
    data = {
        'initial_struct_pks': [],
        'initial_struct_supercell_ids': [],
        'lowest_E_C': [],
        'lowest_E_C_meV_by_atom': [],
        'lowest_E_C_struct_pks': [],
        'lowest_E_C_supercell_ids': []
    }
    tic = time.perf_counter()
    MAX_NB_OF_STRUCT = 100000  # For testing purposes
    if verbosity >= 1:
        print(('Looping through {} structures to find lowest-E_C ' +
              'values and related information. Please be patient...').format(
              qb.count()))
    for row_index, row in enumerate(qb.all(flat=False)):
        initial_struct_pk = row[sc_step_1_struct_column]
        data['initial_struct_pks'].append(initial_struct_pk)
        data['initial_struct_supercell_ids'].append(
            get_supercell_id_from_structure(load_node(initial_struct_pk)))
        # For testing puposes
        if row_index >= MAX_NB_OF_STRUCT:
            break
        process_node = load_node(row[sc_step_2_process_column])
        (E_C, struct_sc_id, struct_pk) = get_lowest_Te0_Te2m_Coulomb_energy(
            process_node)
        if verbosity >= 2:
            print(('{}<{}> with initial struct {} : E_C = {:.2f} eV (structure {} '
                   + 'with supercell id \'{}\')').format(
                  process_node.process_label, process_node.pk, initial_struct_pk,
                  E_C, struct_pk, struct_sc_id))
        data['lowest_E_C'].append(E_C)
        data['lowest_E_C_meV_by_atom'].append(1000 * E_C / len(load_node(struct_pk).sites))
        data['lowest_E_C_struct_pks'].append(struct_pk)
        data['lowest_E_C_supercell_ids'].append(struct_sc_id)

    tac = time.perf_counter()
    if verbosity >= 1:
        print('Processing of {} structures took {:0.2f} s'.format(
            row_index+1, tac - tic))

    # Order data by increasing E_C
    sorted_indexes = list(np.argsort(np.asarray(data['lowest_E_C'])))
    for key in data.keys():
        data[key] = [data[key][index] for index in sorted_indexes]

    # Store data dict in database
    dataDict = Dict(dict=data)
    dataDict.label = data_label
    default_description = ('Coulomb energy data and related information' +
            ' for the lowest-energy Te(0)/Te(2-)/As(3+) configuration derived ' +
            ' from structures in group {}.').format(supercell_struct_group)
    if data_description is None:
        dataDict.description = default_description
    elif not isinstance(data_description,str):
        print(('WARNING: data_description should be a string. Using the ' +
              'default : \'{}\'').format(default_description))
        dataDict.description = default_description
    else:
        dataDict.description = data_description

    # WARNING : Once stored, modifications will not be allowed.
    # Set metadata to help later identification and usage
    dataDict.store()
    # Add data dict to group
    supercell_struct_group.add_nodes(dataDict)
    if verbosity >= 1:
        print(('Data dict node {} stored in database and added to group \'{}\' ' +
               'with :\nlabel: \'{}\',\ndescription: \'{}\'').format(
              dataDict.pk, supercell_struct_group.label, dataDict.label,
              dataDict.description))
    if verbosity >= 2:
        print('Dict {} content :'.format(dataDict.label))
        print_dict_of_lists_content(dataDict)

    return dataDict


@workfunction  # Uncoment when succesful
def assign_original_structure(final_structure: StructureData,
                              is_origin_known=lambda: Bool(False),
                              original_structure=None,
                              verbosity=lambda: Int(1)):
    """
    Find and assign an original structure recursively from an arbitrary number of processes

    Args:
        final_structure: StructureData
            The structure whose provenance should be explored
        is_origin_known: Bool
            Force assignment if the original structure is already known
        original_structure: StructureData or None
            original_structure to force the assignment process

    Returns:
        original_structure: StructureData
            The structure obtained when no high-level incoming processes can
            be found.
    """
    if is_origin_known:
        if isinstance(original_structure, StructureData):
            return {'original_structure': original_structure}
        else:
            return ExitCode(200, 'original_structure input should be a valid StructureData instance.')

    # Default parameters (defined locally to avoid the creation of unnecessary
    # nodes)
    process_classes = [WorkChainNode, WorkFunctionNode]
    process_labels = ['VaspMultiStepRelaxWorkchain', 'change_lattice_to_ref_and_assign']
    MAX_DEPTH = 6
    node_labels_to_skip = ['failed']

    # Check if structure already assigned and exit with status 400 in this case.
    assigned_origin = get_assigned_origin(final_structure, verbosity=verbosity)
    if assigned_origin is not None:
        return ExitCode(400, ('Structure {} already has an assigned structure: ' +
            '{}.').format(final_structure.pk, assigned_origin.pk))

    qb = QueryBuilder()
    qb.append(cls=StructureData, filters={'id': final_structure.pk},
              tag='final_structure')

    upgoing_depth = 0
    outgoing_structures_tag = 'final_structure'
    parent_structure = None
    while 1:
        upgoing_depth += 1
        if verbosity >= 2:
            print('upgoing_depth increased to ', upgoing_depth)
        if upgoing_depth >= MAX_DEPTH:
            print(('Maximum upgoing_depth ({}) has been reached. ' +
                   'Exiting while loop.').format(MAX_DEPTH))
            break
        outgoing_relax_tag = 'relax_nodes_upgoing_depth_' + str(upgoing_depth)
        # Find relaxation process(es) without a specified label
        # Possibility to use (e.g.) cls=[WorkChainNode, CalcJobNode]
        qb.append(cls=process_classes, with_outgoing=outgoing_structures_tag,
                  filters={
                      'attributes.process_label': {'in': process_labels},
                      'attributes.process_state': 'finished',
                      'attributes.exit_status': 0,
                      'label': {'!in': node_labels_to_skip}
                  },
                  tag=outgoing_relax_tag,
                  project=['id'])
        if qb.count() == 0:
            if verbosity >= 2:
                print('No incoming relaxation processes found. ' +
                      'Exiting while loop.')
            break
        elif qb.count() > 1:
            return ExitCode(401, ('More than one incoming processes ' +
                'found for structure(s) {} processes at upgoing depth ' +
                '{}.').format(outgoing_structures_tag, upgoing_depth))
        else:  # qb.count() == 1
            relax_node = load_node(qb.all(flat=True)[-1])
            """
            print('{}<{}> with label \'{}\''.format(
                relax_node.process_label, relax_node.pk, relax_node.label))
            """
            if relax_node.process_label == 'VaspMultiStepRelaxWorkchain':
                if relax_node.inputs.structure.pk == \
                        relax_node.outputs.relax.structure.pk:
                    print('Input and output structures are identical : {}'.format(
                          relax_node.inputs.structure.pk))
                    break
            outgoing_structures_tag = 'structures_upgoing_depth_' + \
                str(upgoing_depth)
            # process_col_index = len(qb.first()) - 1  # Not used
            # edge_filter is used to filter out the reference_structure input
            # in change_lattice_to_ref_and_assign processes

            qb.append(cls=StructureData,
                      with_outgoing=outgoing_relax_tag,
                      tag=outgoing_structures_tag,
                      edge_filters={'label': 'structure'},
                      project=['*'])

            if qb.count() == 1:
                struct_col_index = len(qb.first()) - 1
                query_results = qb.all(flat=True)
                parent_structure = query_results[struct_col_index]
                # Check whether an parent structure already has an assigned origin.
                assigned_origin = get_assigned_origin(parent_structure,
                                                      verbosity=verbosity)
                if assigned_origin is not None:
                    if verbosity >= 1:
                         print(('Original structure {} assigned to final structure ' +
                                '{} based on assignment of parent structure {}.').format(assigned_origin.pk, final_structure.pk,
                                parent_structure.pk))
                    return {'original_structure': assigned_origin}
                query_tags = qb.get_used_tags()

                if verbosity >= 2:
                    print(('Parent structure {} found as input to {}<{}> ' +
                           'process with label {} (tag: \'{}\'.').format(
                               parent_structure.pk, relax_node.process_label,
                               relax_node.pk, relax_node.label,
                               outgoing_relax_tag))
            elif not qb.count():
                return ExitCode(402, ('No input structures found for ' +
                    'relaxation processes at upgoing depth {}.').format(
                        upgoing_depth))
            else:
                return ExitCode(403, ('More than one input structures found ' +
                    'for relaxation processes at upgoing depth {}.').format(
                          upgoing_depth))

    # print('Final tag : ', query_tags[-1])
    original_structure = parent_structure
    if verbosity >= 1 and original_structure is not None:
        print('Original structure {} assigned to final structure {}.',
            original_structure.pk, final_structure.pk)
        # TODO: Write report instead (or in addition to print message).
    elif original_structure is None:
        # TODO : set a non-zero exit code.
        print('WARNING : No original structure found for structure {}'.format(
              final_structure.pk))
        if upgoing_depth >= MAX_DEPTH:
            return ExitCode(404, ('No original structure found for structure ' +
                '{}. Maximum recursion depth ({}) reached. Possibly due to an ' +
                'infinite loop. It can happen if one intermediate process has ' +
                'identical input and output structure pk.').format(
                    final_structure.pk, MAX_DEPTH))
        return ExitCode(405, ('No original structure found for structure {}. ' +
            'Reason is unknown.').format(final_structure.pk))
    # TODO ? add relaxation process list with link_names refering to the
    # relaxation setting
    return {'original_structure': original_structure}


def store_relaxed_struct_in_group(incoming_struct_group, output_group_basename,
    relax_setting, output_group_description='',
    relax_process_label='VaspMultiStepRelaxWorkchain', verbosity=1):
    """
    Find structures relaxed from structures in a group and store them in a new group

    DETAILED DESCRIPTION

    Args:
        incoming_struct_group: str or Group

        output_group_basename: str
            will be used to construct the output group name as :
            output_group_basename + '_' + relax_setting

        relax_setting: str
            label used to identify the relaxation processes in the database.
            It will be then used to construct the output group name as :
            output_group_basename + '_' + relax_setting

        process_label: str. Default = VaspMultiStepRelaxWorkchain

    Returns:
        New group containing the retrieved relaxed structures
    """

    # load group containing incoming structures
    incoming_struct_group = get_group_from_str_or_instance(
        incoming_struct_group, verbosity=verbosity)

    # Set default description
    if isinstance(output_group_description, str):
        if output_group_description == '':
            output_group_description = ('Group containing structures ' +
                'originating from group {} and relaxed with setting ' +
                '\'{}\'.').format(incoming_struct_group.label,
                                  relax_setting)
    else:
        sys.exit('output_group_description should be a string.')

    # load or create output_group
    output_group_label = output_group_basename + '_' + relax_setting
    try:
        output_group = load_group(label=output_group_label)
    except NotExistent:
        output_group = Group(label=output_group_label)
        output_group.store()
        if verbosity >= 1:
            print(('The following group was created to store structures ' +
                   'relaxed with setting {}: {}').format(relax_setting,
                                                         output_group))

    qb = QueryBuilder()
    qb.append(Group, filters={'label': incoming_struct_group.label},
              tag='incoming_struct_group')
    qb.append(cls=StructureData, with_group='incoming_struct_group',
              tag='initial_structures')

    if verbosity >= 1:
        print('Number of structures in group {}: {}.'.format(
            incoming_struct_group.label, qb.count()))

    # Find relaxation processes with specified relax_setting
    qb.append(cls=WorkChainNode, with_incoming='initial_structures',
              filters={
                  'attributes.process_label': relax_process_label,
                  'attributes.process_state': 'finished',
                  'attributes.exit_status': 0,
                  'label': relax_setting
              },
              tag='relax_nodes_' + relax_setting,
              project=['id'])
    if not qb.count():
        sys.exit(('No {} process found with label {} with structures in ' +
                  'group {} as inputs.').format(relax_process_label,
                                                relax_setting,
                                                incoming_struct_group.label))
    # final_relax_col_index = len(qb.first()) - 1  # NOT USED
    if verbosity >= 1:
        print(('Number of successful relax processes with label \'{}\' and ' +
               'structures of group {} as input: {}.').format(
                   relax_setting, incoming_struct_group.label,
                   qb.count()))

    qb.append(cls=StructureData,
              with_incoming='relax_nodes_' + relax_setting,
              tag='structures_relaxed_with_setting_' + relax_setting,
              project=['*'])
    if not qb.count():
        sys.exit(('No output structure found for {} processes with label {}' +
                  'using structures in group {} as inputs.').format(
                  relax_process_label, relax_setting,
                  incoming_struct_group.label))
    # final_struct_col_index = len(qb.first()) - 1  # NOT USED
    final_structures = [row[-1] for row in qb.all(flat=False)]

    # Update the group : only new structures will be added
    qb = QueryBuilder()
    qb.append(Group, filters={'label': output_group.label},
              tag='relaxed_struct_group')
    qb.append(cls=StructureData, with_group='relaxed_struct_group',
              tag='final_structures', project=['id'])
    stored_struct_pks = qb.all(flat=True)
    struct_to_store = [struct for struct in final_structures if struct.pk
                       not in stored_struct_pks]

    if verbosity >= 1:
        print(('Group {} contains {} previously-stored structures. {} will ' +
              'be added.').format(output_group.label, len(stored_struct_pks),
              len(struct_to_store)))

    if len(struct_to_store):
        output_group.add_nodes(struct_to_store)
        print(('{} structures relaxed with setting \'{}\' stored in ' +
              'group {} ({}).').format(len(struct_to_store), relax_setting,
              output_group.label, output_group.description))
    else:
        if verbosity >= 1:
            print('No structures to save in group {}.'.format(
                output_group.label))

    output_group.description = output_group_description

    return output_group


def store_modified_struct_in_group(incoming_struct_group, output_group_label,
        modif_process_types=[WorkFunctionNode],
        modif_process_labels=['change_lattice_to_ref_and_assign'],
        modif_labels=None, output_group_description='', verbosity=1):
    """
    Find structures modified from structures in a group and store them in a new group

    Database queries are used to :
        - find initial structures in incoming_struct_group
        - find processes using these structures as inputs among
          process types in modif_process_types list and process_label
          attributes in modif_process_labels list.
        - find output structures of the above processes
        - add structures to output_group (creating the group if necessary).

    Args:
        incoming_struct_group: str or Group

        output_group_label: str
            label of the output group wherein matching structures will be added.

        modif_process_labels: list if process types
            default is [WorkFunctionNode]
            but could include CalcJobNode, WorkChainNode, etc.

        modif_process_labels: list of str
            List of process_label attributes for the search.
            May incude for example CalcFunctionNode processl_label :
            - 'change_lattice_to_ref_and_assign'
            - 'make_supercell'

    Returns:
        New group containing the retrieved modified structures
    """

    # load group containing incoming structures
    incoming_struct_group = get_group_from_str_or_instance(
        incoming_struct_group, verbosity=verbosity)

    # Set default description
    if isinstance(output_group_description, str):
        if output_group_description == '':
            output_group_description = ('Group containing structures ' +
                'originating from group {} and modified with process(es) ' +
                '{}.').format(incoming_struct_group.label,
                              modif_process_labels)
    else:
        sys.exit('output_group_description should be a string.')

    # load or create output_group
    if not isinstance(output_group_label, str):
        sys.exit('output_group_label should be a string.')
    try:
        output_group = load_group(label=output_group_label)
    except NotExistent:
        output_group = Group(label=output_group_label)
        output_group.store()
        if verbosity >= 1:
            print(('The following group was created to store structures ' +
                   'modified with process(es) {}: {}').format(
                   modif_process_labels, output_group))
    output_group.description = output_group_description

    # Search structures in incoming_struct_group
    qb = QueryBuilder()
    qb.append(Group, filters={'label': incoming_struct_group.label},
              tag='incoming_struct_group')
    qb.append(cls=StructureData, with_group='incoming_struct_group',
              tag='initial_structures')

    if verbosity >= 1:
        print('Number of structures in group {}: {}.'.format(
            incoming_struct_group.label, qb.count()))

    # Find modification processes with specified relax_setting
    qb.append(cls=modif_process_types, with_incoming='initial_structures',
              filters={
                  'attributes.process_label': {'in': modif_process_labels},
                  'attributes.process_state': 'finished',
                  'attributes.exit_status': 0,
              },
              tag='modification_processes',
              project=['id'])
    # Add a filter on modif_labels if not None
    if isinstance(modif_labels, str):
        modif_labels = [modif_labels]
    if isinstance(modif_labels, (list, tuple)):
        qb.add_filter('modification_processes',{'label': {'in': modif_labels}})
    elif modif_labels is not None:
        print('WARNING: modif_labels should be a string, list or tuple.')

    if not qb.count():
        sys.exit(('No {} process found with structures in ' +
                  'group {}<{}> as inputs.').format(modif_process_labels,
                  incoming_struct_group.label, incoming_struct_group.pk))

    # final_relax_col_index = len(qb.first()) - 1  # NOT USED
    if verbosity >= 1:
        print(('Number of successful {} processes with ' +
               'structures in group {}<{}> as input: {}.').format(
                   modif_process_labels, incoming_struct_group.label,
                   incoming_struct_group.pk, qb.count()))

    qb.append(cls=StructureData,
              with_incoming='modification_processes',
              tag='modified_structures',
              project=['*'])
    if not qb.count():
        sys.exit(('No output structure found for {} processes ' +
                  'using structures in group {}<{}> as inputs.').format(
                  modif_process_labels, incoming_struct_group.label,
                  incoming_struct_group.pk))
    # final_struct_col_index = len(qb.first()) - 1  # NOT USED
    final_structures = [row[-1] for row in qb.all(flat=False)]

    # Update the group : only new structures will be added
    qb = QueryBuilder()
    qb.append(Group, filters={'label': output_group.label},
              tag='modified_struct_group')
    qb.append(cls=StructureData, with_group='modified_struct_group',
              tag='final_structures', project=['id'])
    stored_struct_pks = qb.all(flat=True)
    struct_to_store = [struct for struct in final_structures if struct.pk
                       not in stored_struct_pks]

    if verbosity >= 1:
        print(('Group {} contains {} previously-stored structures. {} will ' +
              'be added.').format(output_group.label, len(stored_struct_pks),
              len(struct_to_store)))

    if len(struct_to_store):
        output_group.add_nodes(struct_to_store)
        print(('{} structures modified with process(es) {} stored in ' +
              'group {}<{}> ({}).').format(len(struct_to_store), modif_process_labels,
              output_group.label, output_group.pk, output_group.description))
    else:
        if verbosity >= 1:
            print('No structures to save in group {}.'.format(
                output_group.label))

    return output_group


def assign_original_structures_in_group(struct_group, verbosity=1):
    """
    Assign an origninal structure to all structures in a group

    The function uses project_utils.def assign_original_structure()
    workfunction to facilitate database searches.

    Args:
        struct_group: Group or str
            Group instance or group label

    Returns:
        final_to_orig_struct_mapping: dict
    """
    # Initialize final_to_orig_struct_mapping
    final_to_orig_struct_mapping = {}

    # load group containing structures to assign
    struct_group = get_group_from_str_or_instance(
        struct_group, verbosity=verbosity)

    # Search all structures in final group
    qb = QueryBuilder()
    qb.append(cls=Group, filters={'label': struct_group.label},
              tag='final_structures_group')
    qb.append(cls=StructureData, with_group='final_structures_group',
              tag='final_structures', project=['id'])
    all_final_struct_pks = qb.all(flat=True)

    # Find structures that have had an assignment process
    qb.append(cls=WorkFunctionNode, with_incoming='final_structures',
              filters={
                  'attributes.process_label': 'assign_original_structure',
                  'attributes.process_state': 'finished',
                  'attributes.exit_status': 0
              },
              tag='assignment_processes', project=['id'])
    if qb.count():
        final_struct_with_assign_pks = [row[0] for row
                                        in qb.all(flat=False)]
    else:
        final_struct_with_assign_pks = []

    # Find structures that did not have a successful assignment process
    final_struct_without_assign_pks = list(set(all_final_struct_pks) -
                                           set(final_struct_with_assign_pks))

    if verbosity >= 1:
        print(('Out of {} structures in group {}, {} had already ' +
              'been assigned an original structure and {} had ' +
              'not.').format(len(all_final_struct_pks),
                struct_group.label, len(final_struct_with_assign_pks),
                len(final_struct_without_assign_pks)))

    for final_struct_pk in final_struct_without_assign_pks:
        final_struct = load_node(final_struct_pk)
        results = run_get_node(assign_original_structure , final_structure=final_struct,
                               verbosity=Int(verbosity))
        if not results.node.exit_status:
            if results.result['original_structure'] is not None:
                final_to_orig_struct_mapping[final_struct.pk] = \
                    results.result['original_structure'].pk
        else:
            print(('{}<{}> : exit_status {} - {}').format(results.node.process_label,
                results.node.pk, results.node.exit_status, results.node.exit_message))

    if verbosity >= 2:
        print('Final to oginal structure mapping :')
        print(final_to_orig_struct_mapping)

    return final_to_orig_struct_mapping
    # End of function assign_original_structures_in_group


def get_group_from_str_or_instance(group_name_or_instance, verbosity=1):
    if isinstance(group_name_or_instance, str):
        group_label = group_name_or_instance
        try:
            group = load_group(
                label=group_label)
        except NotExistent:
            sys.exit('No group with label {} found in database.'.format(
                group_label))
    elif isinstance(group_name_or_instance, int):
        try:
            group = load_group(
                pk=group_name_or_instance)
            group_label
        except NotExistent:
            sys.exit('No group with pk {} found in database.'.format(
                group_name_or_instance))
    elif isinstance(group_name_or_instance, Group):
        group = group_name_or_instance
    else:
        sys.exit('group_name_or_instance should be valid aiida.orm.Group ' +
                 'instance or Group label.')
    if verbosity >= 2:
        print(f'Group name: {group.label}')
    return group


def get_structure_from_pk_or_node(struct_or_pk):
    if isinstance(struct_or_pk, StructureData):
        return struct_or_pk
    elif isinstance(struct_or_pk, int):
        struct = load_node(struct_or_pk)
    elif isinstance(struct_or_pk, str):
        struct = load_node(uuid=struct_or_pk)
    else:
        sys.exit('Argument should be a StructureData, int or str.')
    if not isinstance(struct,StructureData):
        sys.exit('No stucture with specified PK or UUID in database.')
    return struct


def retrieve_and_store_relaxed_data(relaxed_struct_group,
                                    relax_setting,
                                    process_label='VaspMultiStepRelaxWorkchain',
                                    include_lowest_E_C=False, plot_figure=False,
                                    store_data=False, data_basename='relaxed_data',
                                    data_description=None, test_mode=False,
                                    include_cell_disto_to_ref=False,
                                    include_xray_pdf_rmsd=False,
                                    experimental_xray_pdf=None,
                                    xray_pdf_r_max=8.0,
                                    xray_pdf_sigma=0.15,
                                    ref_struct_or_id=None,
                                    verbosity=1):

    """
    Collect data associated with relaxed structures in a group

    Data is in the form of a dictionnary of lists wherein
    each key is associated with a list of corresponding values.

    Properties (and keys) include:
        'initial_struct_pks',
        'initial_struct_supercell_ids',
        'relaxed_struct_pks',
        'relax_process_pks',
        'relaxed_energies',
        'maximum_forces'
    and with include_lowest_E_C=True (IMPLEMENTATION IN PROGRESS)
        'lowest_E_C',
        'lowest_E_C_struct_pk',
        'lowest_E_C_supercell_id'

    Data can be stored in database as a Dict if store_data=True

    Args:
        relaxed_struct_group: str or Group
            the instance or label of a Group containing the relaxed
            strutcures to be examined.

        relax_setting: str
            label of the relaxation processes searched in the database
            Used also to construct the label of the Dict if
            store_data=True

        process_label: str (default='VaspMultiStepRelaxWorkchain')
            process_label attribute of the relaxation processes
            searched in the database.

        include_lowest_E_C: bool (default=False)

        include_cell_disto_to_ref: bool (default=False)
            whether distortions to a reference structure should be included

        ref_struct_or_id: StructureData, int, str or None (default=None)
            If None the reference structure will automatically obtained
            from group "AsTe3_initial_structures" using the structure
            description.

        plot_figure: bool (default=False)
            only used if include_lowest_E_C is True.
            plots relaxed E_DFT vs lowest-E_C

        store_data: bool (default=False)
            whether data should be stored as a Dict in the database
            and added to relaxed_struct_group

        data_basename: str (default='relaxed_info')
            only used if store_data is True. Used to compose the
            Dict label as : data_basename + '_' + relax_setting

        data_description: str (default=None)
            Only used if store_data is True. description of the
            Dict containing the collected data.

        test_mode: bool (default is False)
            If true store_data will automatically be set to False
            and the number of post-processed structures will be
            limited to 10.

        verbosity: int (default=1)
            verbosity level

    Returns:
        a Dict stored in the database, and added to relaxed_struct_group
        if store_data=True
        OR
        a dict if store_data=False
    """

    # load group containing incoming structures
    relaxed_struct_group = get_group_from_str_or_instance(
        relaxed_struct_group)

    if test_mode:
        MAX_NB_OF_STRUCT = 10  # for test purposes
        store_data=False

    # Monitoring processing time: start counter
    tic = time.perf_counter()

    # Search structures in group
    qb = QueryBuilder()
    qb.append(Group, filters={'label': relaxed_struct_group.label},
              tag='relaxed_struct_group')
    qb.append(cls=StructureData, with_group='relaxed_struct_group',
              tag='relaxed_structures',
              project=['*'])
    nb_of_relaxed_struct = qb.count()
    relaxed_struct_column = len(qb.first()) - 1
    if verbosity >= 2:
        print('Number of structures in group {}: {}.'.format(
            relaxed_struct_group.label, qb.count()))

    # Check/create/update oiginal-structure assignments
    assign_original_structures_in_group(
        relaxed_struct_group, verbosity=verbosity)

    # Query assignment processes and original structures. Necessary
    # since mapping is only returned for new assignments
    qb.append(cls=WorkFunctionNode, with_incoming='relaxed_structures',
              filters={
                'attributes.process_label': 'assign_original_structure',
                'attributes.process_state': 'finished',
                'attributes.exit_status': 0
              },
              tag='assignment_processes')
    qb.append(cls=StructureData, with_incoming='assignment_processes',
              tag='original_structures', project=['id'])
    initial_struct_column = len(qb.first()) - 1

    # Find relaxation processes with specified relax_setting
    qb.append(cls=WorkChainNode, with_outgoing='relaxed_structures',
              filters={
                  'attributes.process_label': process_label,
                  'attributes.process_state': 'finished',
                  'attributes.exit_status': 0,
                  'label': relax_setting
              },
              tag='relax_' + relax_setting + '_nodes' ,
              project=['*'])
    relax_process_column = len(qb.first()) - 1
    if verbosity >= 2:
        print(('Number of successful relax processes with label \'{}\' in ' +
               'group {}: {}.').format(relax_setting,
                                       relaxed_struct_group.label,
                                       qb.count()))

    if test_mode:
        qb.limit(MAX_NB_OF_STRUCT)

    data_dict = {'initial_struct_pks': [],
                 'relaxed_struct_pks': [],
                 'relax_process_pks': [],
                 'relaxed_energies': [],
                 'relaxed_energies_meV_by_atom': [],
                 'maximum_forces': []
                 }

    if include_cell_disto_to_ref:
        data_dict['cell_shape_disto_to_ref'] = []
        data_dict['cell_size_disto_to_ref'] = []
        data_dict['cell_disto_anisotropy_to_ref'] = []
        data_dict['cell_corners_msd_to_ref'] = []
        if ref_struct_or_id is None:
            # Load default reference structure
            ref_structure = get_refined_structure()
        else:
            ref_structure = get_structure_from_pk_or_node(ref_struct_or_id)

    if include_xray_pdf_rmsd:
        data_dict['xray_pdf_rmsd_opt_sigma'] = []
        data_dict['xray_pdf_sigma_opt'] = []
        data_dict['xray_pdf_rmsd_fixed_sigma'] = []  # TODO: choose sigma value
        pdf_data = get_experimental_xray_total_scattering()
        data_dict['xray_pdf_sigma_fixed'] = xray_pdf_sigma
        data_dict['xray_pdf_r_max'] = xray_pdf_r_max
        nd = nanopdfData(R_max=xray_pdf_r_max, Qmax=pdf_data['q_max'],
                         lambda1=pdf_data['wavelength'],
                         experimental_file=pdf_data['file_name'],
                         verbosity=0)

    if include_lowest_E_C:
        data_dict['lowest_E_C'] = []
        data_dict['lowest_E_C_meV_by_atom'] = []
        data_dict['lowest_E_C_struct_pk'] = []
        data_dict['lowest_E_C_supercell_id'] = []
        # Find processes associated with Te(0)/Te(2-) disorder
        qb.append(CalcFunctionNode, with_incoming='original_structures',
                  filters={
                      'attributes.process_label': 'introduce_Te_ox_state_disorder',
                      'attributes.process_state': 'finished',
                      'attributes.exit_status': 0
                  },
                  tag='Te_ox_state_disorder_introduction')
        qb.append(cls=SinglefileData,
                  with_incoming='Te_ox_state_disorder_introduction',
                  tag='supercell_Te_ox_state_initial_struct_files')
        qb.append(cls=CalcJobNode, filters={
                      'attributes.process_label': 'SupercellCalculation',
                      'attributes.process_state': 'finished',
                      'attributes.exit_status': 0
                  },
                  with_incoming='supercell_Te_ox_state_initial_struct_files',
                  tag='supercell_Te_ox_state_processes',
                  project=['*'])
        if not qb.count():
            print(('No SupercellCalculation processes associated with Te ' +
                   'ox-state disorder found for original structures of ' +
                   'group {}.').format(relaxed_struct_group))
            # TODO : deal with error in this case
        elif verbosity >= 1:
            print(('{} SupercellCalculation processes associated with Te ' +
                   'ox-state disorder found for original structures of ' +
                   'group {}.').format(qb.count(), relaxed_struct_group))
        supercell_Te_ox_state_procs_column = len(qb.first()) - 1

    for row in qb.all():
        relaxed_struct = row[relaxed_struct_column]
        data_dict['relaxed_struct_pks'].append(relaxed_struct.pk)
        initial_struct_pk = row[initial_struct_column]
        data_dict['initial_struct_pks'].append(initial_struct_pk)
        relax_node = row[relax_process_column]
        data_dict['relax_process_pks'].append(relax_node.pk)
        # Fetch maximum force from penultimate step (assumes
        # last step is a SPE
        data_dict['maximum_forces'].append(get_penultimate_step_max_force(
            relax_node.pk))

        if include_cell_disto_to_ref:
            (size_disto, shape_disto, disto_anisotropy, cell_corners_msd
            ) =  measure_cell_disortions(relaxed_struct, ref_structure)
            data_dict['cell_size_disto_to_ref'].append(size_disto)
            data_dict['cell_shape_disto_to_ref'].append(shape_disto)
            data_dict['cell_disto_anisotropy_to_ref'].append(disto_anisotropy)
            data_dict['cell_corners_msd_to_ref'].append(cell_corners_msd)

        if include_xray_pdf_rmsd:
            # TODO: calculate RMSD to experimental PDF with:
            # - optimize sigma
            # - fixed sigma
            nd.sigma = xray_pdf_sigma
            pmg_struct = relaxed_struct.get_pymatgen()
            nd.calculate_final_exact_pdf_from_structure(pmg_struct)
            nd.calculate_difference_to_experiment()
            data_dict['xray_pdf_rmsd_fixed_sigma'].append(nd.difference_rmsd)
            nd.optimize_sigma_vs_expt(pmg_struct, update_sigma=True)
            data_dict['xray_pdf_rmsd_opt_sigma'].append(nd.difference_rmsd)
            data_dict['xray_pdf_sigma_opt'].append(nd.sigma)

        if include_lowest_E_C:
            # TODO: Retrieve supercell information
            supercell_Te_ox_state_process = row[supercell_Te_ox_state_procs_column]
            (min_Ewald_energy, supercell_structure_id, structure_pk) = \
                get_lowest_Te0_Te2m_Coulomb_energy(supercell_Te_ox_state_process)
            data_dict['lowest_E_C'].append(min_Ewald_energy)
            data_dict['lowest_E_C_meV_by_atom'].append(
                1000 * min_Ewald_energy / len(relaxed_struct.sites))
            data_dict['lowest_E_C_struct_pk'].append(structure_pk)
            data_dict['lowest_E_C_supercell_id'].append(supercell_structure_id)

        # TODO : add an exception with final_energy = None
        relaxed_energy = relax_node.outputs.misc['total_energies'][
            'energy_extrapolated']
        data_dict['relaxed_energies'].append(relaxed_energy)
        data_dict['relaxed_energies_meV_by_atom'].append(
            1000 * relaxed_energy / len(relaxed_struct.sites))

        # TODO: add an exception with relax_struct_pk = None in the case where
        # relax process has exit_status 0 without relaxed structure

        if include_lowest_E_C:
            lowest_energy_info = (', supercell ids :(step 1 : {} ; ' +
                'step 2 : {})').format(
                    get_supercell_id_from_structure(initial_struct_pk),
                    supercell_structure_id, min_Ewald_energy,)
        else:
            lowest_energy_info = []
        if verbosity >= 2:
            print(('Initial struct. {}{} relaxed struct {}, ' +
                   'relaxed_energy = {:.3f} eV').format(initial_struct_pk,
                   lowest_energy_info, relaxed_struct.pk, relaxed_energy))
        if relaxed_struct.pk == initial_struct_pk:
            print(('{}<{}>: WARNING - initial and relaxed structures are ' +
                  'the same: {}.').format(relax_node.process_label,
                                          relax_node.pk, initial_struct_pk))

    if include_lowest_E_C and plot_figure:
        plt.plot(data_dict['lowest_E_C'], data_dict['relaxed_energies'], 'ro',
                 markersize=2)
        plt.title(
            'E_DFT vs Coulomic energy of best Te(0)/Te(2-) config - {}'.format(
                relaxed_struct_group.label))
        plt.xlabel('E_C before opt (eV)')
        plt.ylabel('E_DFT before opt (eV)')
        plt.show()

    tac = time.perf_counter()
    print('Processing of {} structures took {:.2f} s.'.format(
        len(data_dict['relaxed_energies']), tac-tic))

    if store_data:
        dataDict = Dict(dict=data_dict)
        dataDict.label = data_basename + '_' + relax_setting
        # Setting default data_description
        if not isinstance(data_description,str):
            print('WARNING: data_desctiption should be a string. Using default description.')
            data_description = None
        if data_description is None:
            data_description = ('Data collected for structures in group {}, ' +
                                'relaxed with {} and relax_setting : {}.').format(
                                relaxed_struct_group.label, process_label,
                                relax_setting)

        dataDict.description = data_description
        dataDict.store()
        relaxed_struct_group.add_nodes(dataDict)

        if verbosity >= 1:
            print(('Data dict node {} stored in database and added to group \'{}\' ' +
                   'with :\nlabel: \'{}\',\ndescription: \'{}\',\n' +
                   'content: ').format(dataDict.pk, relaxed_struct_group.label,
                   dataDict.label, dataDict.description))
            print_dict_of_lists_content(dataDict)
        return dataDict

    else:
        if verbosity >= 1:
            print('returning data_dict: ')
            print_dict_of_lists_content(data_dict)
        return data_dict

    # End of function retrieve_and_store_relaxed_data


def get_group_from_base_and_relax_setting(group_basename, relax_setting):
    group = load_group(label=group_basename + '_' + relax_setting)
    return group


def store_relaxed_struct_and_data(incoming_struct_group, output_group_basename,
        relax_setting, relax_process_label='VaspMultiStepRelaxWorkchain',
        data_basename='relaxed_data', data_description=None, verbosity=1,
        remove_older_data_dicts=False):
    """
    Create or update a group with new structures and associated data Dict

    Structures relaxed with a certain setting and not previously stored
    in the specified group will be added. The corresponding information
    will be stored in a new data Dict.
    """
    group = store_relaxed_struct_in_group(
        incoming_struct_group=incoming_struct_group,
        output_group_basename=output_group_basename,
        relax_setting=relax_setting, verbosity=verbosity)
    assign_original_structures_in_group(group, verbosity=verbosity)

    # cell distortions to the refined structure will be automatically
    # included if 'free_cell' in relax_setting
    data_Dict = retrieve_and_store_relaxed_data(group,
        relax_setting, include_lowest_E_C=True, plot_figure=False,
        store_data=True, data_basename=data_basename,
        data_description=data_description,
        include_cell_disto_to_ref='free_cell' in relax_setting,
        include_xray_pdf_rmsd=True,
        verbosity=verbosity)

    if remove_older_data_dicts:
        remove_older_dicts_from_group(group, data_Dict.label, verbosity=verbosity)


def remove_older_dicts_from_group(group, dict_label, verbosity=1):
    """
    Remove all but most recent Dicts with a certain label from a group
    """
    get_group_from_str_or_instance(group)
    # Check whether several data Dicts with identical labels exist in group
    data_Dicts = get_data_dicts_in_group(group, dict_labels=[dict_label],
                                         order_by_ctime='desc',
                                         verbosity=verbosity)
    if len(data_Dicts) > 1:
        if verbosity >= 1:
            print(('More than one Dict nodes with label \'{}\' found in ' +
                   'group {}<{}>.').format(dict_label, group.label,
                   group.pk))
            print('The following Dict nodes will be removed from the group:')
            for node in data_Dicts[1:]:
                print('\t- {}'.format(node))
            group.remove_nodes(data_Dicts[1:])
    elif len(data_Dicts) == 1:
        print('Group {}<{}> contains only one dict with label {}.'.format(
              group.label, group.pk, dict_label))
    else:
        print('WARNING: group {}<{}> does not contain Dict node with label {}.'.format(
              group.label, group.pk, dict_label))


def get_data_dicts_in_group(group, dict_labels=None, verbosity=1,
                            order_by_ctime='desc',
                            order_by_dict=None):
    """
    Get a list of Dict instances in a group

    Args:
        group: Group or str
            the group or group label wherein Dict instances will be searched.

        dict_labels: list
            list of Dict labels included in the search. If None all Dicts
            in group will be included.

        order_by_ctime: str or None (default: 'desc')
            Should be asc/asending or desc/desceinding. whether identified
            Dicts ehould be ordered by ascending
            creation time. If arg. order_by_dict is not None, the
            latter will be applied first, and order_by_ascending_ctime
            last.

        verbosity: int (default=1)
            verbosity level

    Returns:
        a list of Dict instances
    """
    group = get_group_from_str_or_instance(group)
    qb = QueryBuilder()
    qb.append(cls=Group, filters={'label': group.label},
              tag='group')
    qb.append(cls=Dict, with_group='group',
              tag='dicts', project=['*'])
    if dict_labels is not None:
        qb.add_filter('dicts', {'label': {'in': dict_labels}})
    if order_by_dict is not None:
        qb.order_by({Dict: order_by_dict})
    if isinstance(order_by_ctime, str):
        if order_by_ctime.lower() not in ['asc', 'ascending', 'desc', 'descending']:
            print('WARNING : order_by_ctime should be None/\'asc\'/\'desc\'. ' +
                  'Set to None.')
        else:
            if order_by_ctime.lower() in ['asc', 'ascending']:
                ctime_ordering_dir = 'asc'
            elif order_by_ctime.lower() in ['desc', 'descending']:
                ctime_ordering_dir = 'desc'
            qb.order_by({Dict: {'ctime': ctime_ordering_dir}})

    data_Dicts = [node for node in qb.all(flat=True)]
    if verbosity >= 1:
        print('{} Dicts found in group {}:'.format(
            qb.count(), group.label))
        for node in qb.all(flat=True):
            print('  Dict {} with label {}.'.format(node.pk,
                                                    node.label))
            try:
                print('  Contains data for {} structures.'.format(
                      len(node['relaxed_energies'])))
            except Exception as e:
                print(e)
            if verbosity >= 2:
                print_dict_of_lists_content(node, prefix='    ')

    return data_Dicts


def plot_relaxed_energies_from_data_Dicts(x_data_group,
                                          x_data_dict_label='relaxed_data',
                                          x_data_key = 'relaxed_energies',
                                          y_data_group=None,
                                          y_data_dict_label=None,
                                          y_data_key =None,
                                          x_label='Energy (eV)',
                                          y_label='Energy (eV)',
                                          scatter_size_data_group=None,
                                          scatter_size_data_dict_label=None,
                                          scatter_size_data_key='maximum_forces',
                                          scatter_size_label=None,
                                          scatter_size_scaling=10,
                                          scatter_color_data_group=None,
                                          scatter_color_data_dict_label=None,
                                          scatter_color_data_key='maximum_forces',
                                          scatter_color_label=None,
                                          title=None, verbosity = 1,
                                          pickradius=5,
                                          print_picking_info=False,
                                          max_data_points_to_pick=3,
                                          show_nd=True,
                                          nd_experimental_file=None,
                                          nd_experimental_title=None,
                                          nd_two_theta_range=(0,120),
                                          nd_fwhm=0.2,
                                          nd_eta=0.5,
                                          show_xrd=True,
                                          xrd_experimental_file=None,
                                          xrd_experimental_title=None,
                                          xrd_two_theta_range=(0,120),
                                          xrd_fwhm=0.2,
                                          xrd_eta=0.5,
                                          show_total_pdf=True,
                                          total_pdf_experimental_file=None,
                                          total_pdf_experimental_title=None,
                                          total_pdf_r_max=12.0,
                                          total_pdf_sigma=0.1,
                                          total_pdf_wavelength=0.709318,
                                          total_pdf_y_shift_factor=0.5,
                                          show_dist_matrix=True,
                                          dist_matrix_r_max=10,
                                          dist_matrix_sigma=0.01,
                                          **kwargs):
    """
    Plot energies of two distinct relaxation levels

    TODO: add unit to plot data (e.g. in eV/atom)
    TODO: Remove outliers from options and/or interactively
    (e.g. by selecting with Ctrl pressed)
    """
    from matplotlib.widgets import RectangleSelector

    # search data Dicts in groups for x, y, and (if relevant) scatter_size and scatter_color
    x_data_group = get_group_from_str_or_instance(x_data_group, verbosity=verbosity)
    if y_data_group is None:
        y_data_group = x_data_group
    if y_data_dict_label is None:
        y_data_dict_label = x_data_dict_label
    y_data_group = get_group_from_str_or_instance(y_data_group, verbosity=verbosity)
    searched_group_labels = [x_data_group.label, y_data_group.label]
    searched_data_dict_labels = [x_data_dict_label, y_data_dict_label]
    if scatter_size_data_group is not None:
        scatter_size_data_group = get_group_from_str_or_instance(
            scatter_size_data_group, verbosity=verbosity)
        searched_group_labels.append(scatter_size_data_group.label)
        if scatter_size_data_dict_label is None:
            scatter_size_data_dict_label = x_data_dict_label
        searched_data_dict_labels.append(scatter_size_data_dict_label)
    if scatter_color_data_group is not None:
        scatter_color_data_group = get_group_from_str_or_instance(
            scatter_color_data_group, verbosity=verbosity)
        searched_group_labels.append(scatter_color_data_group.label)
        if scatter_color_data_dict_label is None:
            scatter_color_data_dict_label = x_data_dict_label
        searched_data_dict_labels.append(scatter_color_data_dict_label)
    # Use sets to remove duplicates
    searched_group_labels = list(set(searched_group_labels))
    searched_data_dict_labels = list(set(searched_data_dict_labels))

    qb = QueryBuilder()
    qb.append(cls=Group,
              filters={
                'label': {'in': searched_group_labels}
              },
              tag='groups',project=['label'])
    group_label_column = len(qb.first()) - 1

    qb.append(cls=Dict, with_group='groups',
              filters={
                'label': {'in': searched_data_dict_labels}
              },
              tag='data_dicts', project=['label', 'id'])
    qb.order_by({Dict: {'ctime': 'desc'}})
    # TODO: make sure to only take the latest if sereral Dicts with the same
    # label exist in a group.
    data_Dict_label_column = len(qb.first()) - 2
    data_Dict_id_column = len(qb.first()) - 1

    # Load the appropriate data Dict nodes
    # Only pick the first (e.g. most recent) with searched data name
    data_dicts = 4*[None]
    data_dict_labels = [x_data_dict_label, y_data_dict_label,
                        scatter_size_data_dict_label,
                        scatter_color_data_dict_label]
    for row in qb.all(flat=False):
        for index, (data_dict, data_dict_label) in enumerate(
            zip(data_dicts, data_dict_labels)):
            if data_dict is None and row[data_Dict_label_column] == data_dict_label:
                data_dicts[index] = load_node(row[data_Dict_id_column])
        if all([data_dict is not None for data_dict in data_dicts]):
            break
    [x_data_Dict, y_data_Dict, scatter_size_data_Dict,
     scatter_color_data_Dict] = data_dicts

    # Find indexes of intersection between initial_pks in x, y data sets,
    # and in size and color if relevant.
    scatter_size_data = None
    scatter_color_data = None
    if x_data_group.pk == y_data_group.pk and \
            x_data_dict_label == y_data_dict_label and \
            all([data_Dict is None for data_Dict in
                 [scatter_size_data_Dict, scatter_color_data_Dict]]):
        x_data = x_data_Dict[x_data_key]
        y_data = x_data_Dict[y_data_key]
    else:
        # Find indexes of the interection between x and y data initial structures
        common_init_pks = list(set(x_data_Dict['initial_struct_pks']) &
                               set(y_data_Dict['initial_struct_pks']))

        x_data = []
        y_data = []
        if scatter_size_data_Dict is not None:
            scatter_size_data = []
        if scatter_color_data_Dict is not None:
            scatter_color_data = []

        for init_struct_pk in common_init_pks:
            index_pair = (x_data_Dict['initial_struct_pks'].index(init_struct_pk),
                          y_data_Dict['initial_struct_pks'].index(init_struct_pk))
            x_data.append(x_data_Dict[x_data_key][index_pair[0]])
            y_data.append(y_data_Dict[y_data_key][index_pair[1]])

            # Append scatter_size_data (using x_data_Dict or y_data_Dict if they match)
            if scatter_size_data_Dict == x_data_Dict:
                scatter_size_data.append(x_data_Dict[scatter_size_data_key][index_pair[0]])
            elif scatter_size_data_Dict == y_data_Dict:
                scatter_size_data.append(y_data_Dict[scatter_size_data_key][index_pair[1]])
            elif scatter_size_data_Dict is not None:
                index = scatter_size_data_Dict['initial_struct_pks'].index(init_struct_pk)
                scatter_size_data.append(scatter_size_data_Dict[scatter_size_data_key][index])
            # Append scatter_color_data (using x_data_Dict or y_data_Dict if they match)
            if scatter_color_data_Dict == x_data_Dict:
                scatter_color_data.append(x_data_Dict[scatter_color_data_key][index_pair[0]])
            elif scatter_color_data_Dict == y_data_Dict:
                scatter_color_data.append(y_data_Dict[scatter_color_data_key][index_pair[1]])
            elif scatter_color_data_Dict is not None:
                index = scatter_color_data_Dict['initial_struct_pks'].index(init_struct_pk)
                scatter_color_data.append(scatter_color_data_Dict[scatter_color_data_key][index])

        # Set default size or color value if one of scatter_size_data
        # and scatter_color_data is None
        if scatter_size_data is None and scatter_size_data is not None:
            scatter_size_data = 3  # TODO: Change for scatter_default_size
        elif scatter_color_data is None and scatter_size_data is not None:
            scatter_color_data = 'r'

    # Draw a plot or a scatter
    # TODO: store initial x_data, y_data (and sizes and colors for scatter plots)
    # to possibly reset the plot after outliers removal.
    # plots should be made in a function than can be called again if
    # outliers are removed.

    fig, ax = plt.subplots()
    plots = []

    def clean_plots():
        for i in reversed(range(len(plots))):
            if isinstance(plots[i], list):  # plots[i] is a list of line2D objects (plt.plot)
                for line in plots[i]:
                    line.remove()
                plots.pop(i)
            elif isinstance(plots[i], (Colorbar, PathCollection)):
                try:
                    plots[i].remove()
                    plots.pop(i)
                except ValueError:
                    plots.pop(i)

    def clean_legend():
        lgd = ax.get_legend()
        if lgd is not None:
            lgd.remove()

    def draw_plot():
        """ Draw a plot or scatter plot """
        # Remove pre-existing plots
        clean_plots()
        if scatter_size_data is None and scatter_color_data is None:
            plots.append(ax.plot(x_data, y_data, 'ro', picker=True,
                                  pickradius=pickradius, **kwargs))
            fig.set_size_inches((6.4, 4.8))
        else: # scatter_size_data or and/or scatter_color_data in ['cell_size_disto, cell_shape_disto, cell_disto_anisotropy']
            clean_legend()
            # TODO: automatically calculate appropriate size scale
            if isinstance(scatter_size_data, (int, float)):
                size_scale_factor = 1
            else:
                size_scale_factor = scatter_size_scaling / np.square(max(scatter_size_data))
            sc = ax.scatter(x_data, y_data,
                s=size_scale_factor*np.square(np.asarray(scatter_size_data)),
                c=scatter_color_data, picker=True,
                cmap='RdYlGn_r', pickradius=pickradius,
                alpha=0.5, linewidths=1)
            cb = plt.colorbar(sc, ax=ax, label=scatter_color_label)
            plots.append(sc)
            plots.append(cb)
            handles, labels = sc.legend_elements(prop='sizes', num = 5, alpha=0.2,
                                                 func=lambda x: np.sqrt(x/size_scale_factor))
            # ax.legend(handles, labels, loc='upper left', title=scatter_size_label)
            ax.legend(handles, labels, loc='upper left',
                      bbox_to_anchor=(1.3, 0.95), title=scatter_size_label)
            fig.set_size_inches((9.6, 4.8))

        if title is not None:
            ax.set(title=title)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        (x_min, x_max) = (min(x_data), max(x_data))
        (y_min, y_max) = (min(y_data), max(y_data))
        x_range = x_max-x_min
        y_range = y_max-y_min
        ax.set(xlim=[x_min - 0.05*x_range, x_max + 0.05*x_range],
               ylim=[y_min - 0.05*y_range, y_max + 0.05*y_range])


        plt.tight_layout()
        fig.canvas.draw()
        fig.canvas.flush_events()

        # End of draw_plot function

    draw_plot()

    # Define the picking function. Needs to be done within the plot function
    # to be able to use data Dicts
    def onpick(event):
        artist = event.artist
        key_pressed = event.mouseevent.key
        xmouse, ymouse = event.mouseevent.xdata, event.mouseevent.ydata
        path_indexes = event.ind

        if print_picking_info:
            print('Artist picked:', event.artist)
            print('{} vertices picked'.format(len(path_indexes)))
            print('Pick between vertices {} and {}'.format(min(path_indexes),
                                                           max(path_indexes)+1))
            print('x, y of mouse: {:.2f},{:.2f}'.format(xmouse, ymouse))
            if key_pressed is not None:
                print('{} button was pressed upon ckick event.'.format(key_pressed))

        # TODO: remove outliers and update plots key_pressed == 'control'
        # and update plots.
        print('Data points within {}-point picking tolerance: {}'.format(
            pickradius, len(path_indexes)))

        if key_pressed == 'control':
            print(50*'-' + '\nData points selected for suppression (outlyers)\n' + 50*'-')
            for index in path_indexes:
                print('  outlyer: ({:.3f}, {:.3f})'.format(x_data[index],
                                                           y_data[index]))
                # Retrieve original indexes in data Dict
                x_data.pop(index)
                y_data.pop(index)
                if scatter_size_data is not None or scatter_color_data is not None:
                    if len(scatter_size_data) > 1 and len(scatter_size_data) > index:
                        scatter_size_data.pop(index)
                    if len(scatter_color_data) > 1 and len(scatter_color_data) > index:
                        scatter_color_data.pop(index)
            print('Plotted data set is now of size {}.'.format(len(x_data)))
            #re-plot
            draw_plot()

        elif len(path_indexes) <= max_data_points_to_pick:
            for index in path_indexes:
                print('--------------------------------------')
                print('  Data point: ({:.3f}, {:.3f})'.format(x_data[index],
                                                              y_data[index]))
                print('--------------------------------------')
                # Retrieve original indexes in data Dict
                x_data_index = x_data_Dict[x_data_key].index(x_data[index])
                y_data_index = y_data_Dict[y_data_key].index(y_data[index])
                # DEBUGGING:
                print('x_data_index = ', x_data_index)
                print(('x data : {} property within Dict {}<{}> in group '
                      '{}<{}>').format(x_data_key, x_data_dict_label,
                      x_data_Dict.pk, x_data_group.label, x_data_group.pk))
                for key in x_data_Dict.keys():
                    if isinstance(x_data_Dict[key], (list, tuple, np.ndarray)):
                        print('\t{}: {}'.format(key, x_data_Dict[key][x_data_index]))
                    else:
                        print('\t{}: {}'.format(key, x_data_Dict[key]))
                print(('y data : {} property within Dict {}<{}> in group '
                      '{}<{}>').format(y_data_key, y_data_dict_label,
                      y_data_Dict.pk, y_data_group.label, y_data_group.pk))
                for key in y_data_Dict.keys():
                    if isinstance(y_data_Dict[key], (list, tuple, np.ndarray)):
                        print('\t{}: {}'.format(key, y_data_Dict[key][y_data_index]))
                    else:
                        print('\t{}: {}'.format(key, y_data_Dict[key]))

                structure = load_node(y_data_Dict['relaxed_struct_pks'][y_data_index])
                view(structure.get_ase())

                if show_nd:
                    plot_diffraction_pattern(structure, diffraction_type='neutron',
                        experimental_file=nd_experimental_file,
                        experimental_title=nd_experimental_title,
                        fwhm=nd_fwhm, eta=nd_eta,
                        two_theta_range=nd_two_theta_range)

                if show_xrd:
                    plot_diffraction_pattern(structure, diffraction_type='x-ray',
                        experimental_file=xrd_experimental_file,
                        experimental_title=xrd_experimental_title,
                        fwhm=xrd_fwhm, eta=xrd_eta,
                        two_theta_range=xrd_two_theta_range)

                if show_total_pdf:
                    title = 'Total X-ray scattering for structure {}'.format(
                        structure.pk)
                    plot_multiple_total_x_ray_scattering(structure,
                        experimental_file=total_pdf_experimental_file,
                        experimental_title=total_pdf_experimental_title,
                        sigma=total_pdf_sigma, R_max=total_pdf_r_max,
                        lambda1=total_pdf_wavelength, verbosity=0,
                        plot_title=title, y_shift_factor=total_pdf_y_shift_factor)
                plt.show()
        else:
            print(('Zoom in to ensure that {} or less data points are ' +
                   'witin picking tolerance range ({} points).').format(
                       max_data_points_to_pick, pickradius))
        return True  # end of function

    def line_select_callback(eclick, erelease):
        """
        Callback for line selection.

        *eclick* and *erelease* are the press and release events.
        """
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        if verbosity >= 2:
            print('Rectangle selector : ')
            print(f"({x1:3.2f}, {y1:3.2f}) --> ({x2:3.2f}, {y2:3.2f})")
            print(f" The buttons you used were: {eclick.button} {erelease.button}")
        # TODO: select points in the rectangle
        picked_indexes_in_plot = [index for index, (x, y) in enumerate(zip(x_data, y_data))
                            if x >= min(x1, x2) and x <= max(x1, x2) and
                            y >= min(y1, y2) and y <= max(y1, y2)]
        if verbosity >= 2:
            print('Picked data points:')
            for index in picked_indexes_in_plot:
                print('\t({}, {})'.format(x_data[index], y_data[index]))
        picked_data_indexes = []
        for index in picked_indexes_in_plot:
            # find matching x_data_Dict and y_data_Dict index(es)
            x_data_index = x_data_Dict[x_data_key].index(x_data[index])
            y_data_index = y_data_Dict[y_data_key].index(y_data[index])
            # TODO: verify concordance
            if x_data_Dict['initial_struct_pks'][x_data_index] != \
                    y_data_Dict['initial_struct_pks'][y_data_index]:
                print(('WARNING: Initial structures associated with x ({}), ' +
                       'and y ({}) data points do not match.').format(
                       x_data_Dict['initial_struct_pks'][x_data_index],
                       y_data_Dict['initial_struct_pks'][y_data_index]))
                # TODO: decide what to do in this case (several data points
                # with identical x or y values)
            else:
                picked_data_indexes.append((x_data_index, y_data_index))
        # rank index pairs by increassing relaxed energy
        picked_struct_energies = [y_data_Dict['relaxed_energies'][y_index]
                                  for (_, y_index) in picked_data_indexes]
        sorted_indexes = np.argsort(picked_struct_energies)
        picked_data_indexes = [picked_data_indexes[index] for index in sorted_indexes]

        picked_data_dict = {}
        picked_data_dict['picked_data_indexes'] = picked_data_indexes
        picked_data_dict['relaxed_energies'] = [
            y_data_Dict['relaxed_energies'][y_index] for (_, y_index) in picked_data_indexes]
        # create list of relaxed strutures based on y_data_Dict['relaxed_struct_pks']
        # TODO: use pandas to draw a table of properties.

        picked_data_dict['init_struct_pks'] = [x_data_Dict['initial_struct_pks'][x_index]
                                               for (x_index, _) in picked_data_indexes]
        if 'relaxed_struct_pks' in x_data_Dict.keys():
            picked_data_dict['x_struct_pks'] = [x_data_Dict['relaxed_struct_pks'][x_index]
                                                for (x_index, _) in picked_data_indexes]
        elif 'lowest_E_C_struct_pks' in x_data_Dict.keys():
            picked_data_dict['x_struct_pks'] = [x_data_Dict['lowest_E_C_struct_pks'][x_index]
                                                for (x_index, _) in picked_data_indexes]
        picked_data_dict['y_struct_pks'] = [y_data_Dict['relaxed_struct_pks'][y_index]
                                            for (_, y_index) in picked_data_indexes]
        picked_data_dict['relax_processes'] = [y_data_Dict['relax_process_pks'][y_index]
                                               for (_, y_index) in picked_data_indexes]
        picked_data_dict['max_forces'] = [y_data_Dict['maximum_forces'][y_index]
                                          for (_, y_index) in picked_data_indexes]
        picked_data_dict['x_values'] = [x_data_Dict[x_data_key][x_index]
                                        for (x_index, _) in picked_data_indexes]
        picked_data_dict['y_values'] = [y_data_Dict[y_data_key][y_index]
                                        for (_, y_index) in picked_data_indexes]
        # TODO: get deviation between x and y data structures
        picked_data_dict['size_disto(%)'] = [] # Initialize list of (size, shape, aniso) tuple
        picked_data_dict['shape_disto(%)'] = [] # Initialize list of (size, shape, aniso) tuple
        picked_data_dict['cell_corners_msd'] = []
        for x_struct_pk, y_struct_pk in zip(picked_data_dict['x_struct_pks'],
                                            picked_data_dict['y_struct_pks']):
            size_disto, shape_disto, _, cell_corners_msd = measure_cell_disortions(
                y_struct_pk, x_struct_pk)
            picked_data_dict['size_disto(%)'].append(100*size_disto)
            picked_data_dict['shape_disto(%)'].append(100*shape_disto)
            picked_data_dict['cell_corners_msd'].append(cell_corners_msd)

        picked_names = ['{} ({:.2f} eV)'.format(PK, E) for (PK, E) in
                        zip(picked_data_dict['y_struct_pks'],
                        picked_data_dict['relaxed_energies'])]
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)
        pd.set_option('display.width', None)
        pd.set_option('display.max_colwidth', None)
        pd.options.display.float_format = '{:.4f}'.format
        df = pd.DataFrame(picked_data_dict)

        print(df.loc[:, ['x_struct_pks', 'y_struct_pks', 'relax_processes',
                        'relaxed_energies', 'max_forces', 'size_disto(%)',
                        'shape_disto(%)']])

        if show_xrd:
            plot_diffraction_patterns(picked_data_dict['y_struct_pks'], diffraction_type='xrd',
                                      experimental_file=xrd_experimental_file,
                                      experimental_title=xrd_experimental_title,
                                      two_theta_range=xrd_two_theta_range,
                                      fwhm=xrd_fwhm, eta=xrd_eta,
                                      data_names=picked_names)
        if show_nd:
            plot_diffraction_patterns(picked_data_dict['y_struct_pks'], diffraction_type='nd',
                                      experimental_file=nd_experimental_file,
                                      experimental_title=nd_experimental_title,
                                      two_theta_range=nd_two_theta_range,
                                      fwhm=nd_fwhm, eta=nd_eta,
                                      data_names=picked_names)
        if show_total_pdf:
            plot_multiple_total_x_ray_scattering(picked_data_dict['y_struct_pks'],
                experimental_file=total_pdf_experimental_file,
                experimental_title=total_pdf_experimental_title,
                sigma=total_pdf_sigma, R_max=total_pdf_r_max,
                lambda1=total_pdf_wavelength, data_names=picked_names,
                y_shift_factor=total_pdf_y_shift_factor, verbosity=0)

        if show_dist_matrix:
            plot_distance_matrix(picked_data_dict['y_struct_pks'],
                                 r_max=dist_matrix_r_max,
                                 sigma=dist_matrix_sigma)
        plt.show()

    def toggle_selector(event):
        # Activate Rectanle selection by pressing r
        print(' Key pressed.')
        if event.key == 'r':
            if toggle_selector.RS.active:
                print(' RectangleSelector deactivated.')
                toggle_selector.RS.set_active(False)
            else:
                print(' RectangleSelector activated.')
                toggle_selector.RS.set_active(True)

    # drawtype is 'box' or 'line' or 'none'
    toggle_selector.RS = RectangleSelector(ax, line_select_callback,
        drawtype='box', useblit=True, button=[1], minspanx=5,
        minspany=5, spancoords='pixels', interactive=False)

    fig.canvas.mpl_connect('pick_event', onpick)
    fig.canvas.mpl_connect('key_press_event', toggle_selector)

    # TODO:
    # Add slider to change scatter size factor
    # Add a helper text to indicate the role of Ctrl key (remove outlyers)
    # Add a reset button to re-introduce removed outlyers
    # Plot energies in eV/atom

    return fig, ax

def get_total_x_ray_scattering(structure_or_pk, plot_pdf=True,
                               plot_title=None,
                               experimental_file=None,
                               experimental_title=None,
                               verbosity=0, **kwargs):
    """
    Args:
        kwargs may inclmude alla rguments used by nanopdfData
    """
    structure = get_structure_from_pk_or_node(structure_or_pk)
    pmg_struct = structure.get_pymatgen()
    nd = nanopdfData(verbosity=verbosity, **kwargs)
    nd.calculate_final_exact_pdf_from_structure(pmg_struct)

    # TODO : allow setting title
    if plot_pdf:
        fig, ax = nd.plot_exact_pdf(nd.R, nd.exactPDF,
                                    experimental_file=experimental_file,
                                    experimental_title=experimental_title)
        return nd.R, nd.exactPDF, fig, ax
    else:
        return nd.R, nd.exactPDF


def get_list_of_structures(structures_or_pks):
    """
    get a list of StructureData from a single or list of structure(s) or PK(s)

    Args:
        structures_or_pks: Int, str, StructureData or a list thereof
        May be a sinle PK, UIID, StructureData or a list thereof

    Returns:
        structures: list of StructureData
    """
    if isinstance(structures_or_pks, (int, str, StructureData)):
        structures_or_pks = [structures_or_pks]
    structures = [get_structure_from_pk_or_node(struct)
                  for struct in structures_or_pks]
    return structures


def plot_multiple_total_x_ray_scattering(structures_or_pks, data_names=None,
        plot_title='Exact total X-ray scattering', experimental_file=None,
        experimental_title=None, y_shift_factor=1.1, push_legend=True,
        include_experiment=True, **kwargs):
    """
    Plot total X-ray scattering for a series of structures (or PKs)
    """
    structures = get_list_of_structures(structures_or_pks)

    legend = check_data_names(data_names, structures)

    fig, ax = plt.subplots()
    y_shift = 0
    r_max = 0
    for index, structure in enumerate(structures):
        r, pdf = get_total_x_ray_scattering(structure, plot_pdf=False,
            plot_title=None, experimental_file=experimental_file,
            experimental_title=experimental_title, **kwargs)
        ax.plot(r, pdf + y_shift)
        if y_shift == 0:
            y_range = max(pdf)-min(pdf)
            y_shift = y_shift_factor * y_range
        else:
            y_shift += y_shift_factor * y_range
        if max(r) > r_max:
            r_max = max(r)

    if include_experiment or experimental_file is not None:
        if experimental_file is None:
            expt_pdf_data = get_experimental_xray_total_scattering()
            experimental_file = expt_pdf_data['file_name']
            if experimental_title is None:
                experimental_title = expt_pdf_data['file_name']
        expt_data = np.loadtxt(experimental_file, delimiter=None, skiprows=0)
        # Adding experimental ND data to plot
        ax.plot(expt_data[:,0], expt_data[:,1] + y_shift,
                color='black')
        if experimental_title is None:
            experimental_title = 'Experiment'
        legend += [experimental_title]

    ax.set(xlim=[0, max(r)])
    if isinstance(plot_title, str):
        plt.title(plot_title)
    plt.xlabel('r (\u212B)')
    plt.ylabel('Intensity (AU)')
    if push_legend:
        plt.legend(legend,loc='upper left', bbox_to_anchor= (1.0, 1.0))
        fig.set_size_inches(9.6, 4.8)
    else:
        plt.legend(legend)
        fig.set_size_inches(6.4, 4.8)
    plt.tight_layout()
    # plt.show()

    return fig, ax


def add_smoothing_functions(pattern, fwhm=0.2, eta=0.5,
                            two_theta_range=[0, 120],
                            step=None):
    """
    Add pseudo-Voigt broadening to a diffraction pattern (pymatgen)

    Args:
        pattern:
            pymatgen diffraction pattern as obtained with XRDCalculator
            or NDCalculator
        fwhm: float (default=0.2)
            Full width at half maximum in units of 2*theta (°)
        eta: float (default=0.5)
            Lorentzian to Gaussian ratio (1 for pure Lorentzian, 0 for
            pure Gaussian profile)
        two_theta_range:
        step: float or None
            two_theta step for the x axis. Default is 0.1*fwhm

    Returns:
        (x, y): tuple
            x:  two_theta axis between two_theta_range[0] and
                two_theta_range[1] with step.
            y:  simulated diffracted intensities at x.
    """
    H = fwhm  # fwhm in 2(theta) units
    if eta < 0 or eta > 1:
        print('WARNING : eta should be between 0 and 1. Using the default (0.5).')
        eta = 0.5
    if step==None:
        step = 0.1*H

    x = np.arange(two_theta_range[0], two_theta_range[1], step)
    x_hkl = pattern.x
    I_hkl = pattern.y
    # Initialize y
    y = np.zeros(len(x))
    for i in range(len(x_hkl)):
        # Gaussian broadening
        G = (2/H)*np.sqrt(np.log(2)/np.pi) * np.exp(-(4*np.log(2)/(H*H)) * np.square(x - x_hkl[i]))
        # Gaussian broadening
        L = 2/(np.pi * H) / (1 + 4/(H*H) * np.square(x - x_hkl[i]))
        y += I_hkl[i] * (eta * L + (1 - eta) * G)
    # Normalize:
    y = y/max(y)

    return x, y


def plot_diffraction_pattern(structure_or_pk,
                             diffraction_type='X-ray',
                             wavelength=None,
                             two_theta_range=(0, 80),
                             fwhm=0.2, eta=0.5,
                             include_experiment=True,
                             experimental_file=None,
                             experimental_title=None,
                             push_legend=False):
    """
    Plot neutron diffraction pattern for an input StructureData instance
    """
    structure=get_structure_from_pk_or_node(structure_or_pk)

    # Plot simulated neutron or x-raydiffraction pattern with Pymatgen
    pmg_struct = structure.get_pymatgen()
    if diffraction_type.lower() in ['xrd', 'x-ray', 'x-rays']:
        diffraction_type_name = 'X-ray diffraction'
        if wavelength is None:
            wavelength = 'CuKa'
        calculator=XRDCalculator(wavelength=wavelength, symprec=0,
                                 debye_waller_factors=None)
    elif diffraction_type.lower() in ['nd', 'neutron', 'neutrons']:
        diffraction_type_name = 'neutron diffraction'
        if wavelength is None:
            wavelength = 2.522
        calculator=NDCalculator(wavelength=wavelength, symprec=0,
                                debye_waller_factors=None)
    else:
        sys.exit('diffraction_type should be neutron/ND or X-ray/XRD.')

    pattern=calculator.get_pattern(pmg_struct, scaled=True,
                                   two_theta_range=two_theta_range)

    x, y = add_smoothing_functions(pattern, fwhm=fwhm,
        eta=eta, two_theta_range=two_theta_range)
    fig, ax = plt.subplots()
    ax.plot(x, y)
    legend = ['Structure {}'.format(structure.pk)]

    if include_experiment or experimental_file is not None:
        if experimental_file is None:
            expt_diffract_data = get_experimental_diffraction(diffraction_type=diffraction_type)
            experimental_file = expt_diffract_data['file_name']
            if experimental_title is None:
                experimental_title = expt_diffract_data['title']
        # Load experimental ND pattern
        expt_data = np.loadtxt(experimental_file, delimiter=None, skiprows=0)

        # Adding experimental ND data to plot
        ax.plot(expt_data[:,0], expt_data[:,1]/max(expt_data[:,1])+
                 1.1*max(y),color='black')
        if experimental_title is None:
            legend.append('Experiment')
        else:
            legend.append(experimental_title)

    ax.set_title(('Simulated {} pattern for structure {}').format(
        diffraction_type_name, structure.pk))
    ax.set(xlabel='2\u03B8 (°)',ylabel='Normalized Intensity (AU)',
           xlim=two_theta_range)
    if push_legend:
        ax.legend(legend,loc='upper left', bbox_to_anchor= (1.0, 1.0))
        fig.set_size_inches(12.8, 4.8)
    else:
        ax.legend(legend)
        fig.set_size_inches(9.6, 4.8)
    plt.tight_layout()
    # plt.show()

    return fig, ax


def check_data_names(data_names, structures):
    """
    Check data_names length and type and return a copy to avoide in-place modif
    """
    # Check data_names
    if isinstance(data_names, (list, tuple)):
        if len(data_names) != len(structures):
            print('WARNING: data_names and structures have different length.')
            print('Structure PKs = ',[struct.pk for struct in structures])
            print('data_names = ', data_names)
            _data_names = None
        else:
            _data_names = data_names.copy()
    elif isinstance(data_names, str):
        _data_names = [data_names]
    elif data_names is not None:
        print('WARNING: data_names should be a list or tuple. Using the default.')
        _data_names = None
    if data_names is None:
        _data_names = [f'structure {struct.pk}' for struct in structures]

    return _data_names


def plot_diffraction_patterns(structures_or_pks,
                              diffraction_type='X-ray',
                              wavelength=None,
                              two_theta_range=None,
                              data_names=None,
                              plot_title=None,
                              fwhm=0.2, eta=0.5,
                              experimental_file=None,
                              experimental_title=None):
    """
    Plot simulated neutron or x-ray  diffraction patterns for input structures
    """
    diffraction_type_name = 'X-ray diffraction'
    DEFAULT_TWO_THETA_RANGE = (0, 120)
    structures = get_list_of_structures(structures_or_pks)

    legend = check_data_names(data_names, structures)
    # Load experimental_file and set two_theta_range accordingly
    if experimental_file is not None:
        # Load experimental ND pattern
        expt_data = np.loadtxt(experimental_file,delimiter=None,skiprows=0)
        if two_theta_range is None:
            two_theta_range = (min(expt_data[:,0]), max(expt_data[:,0]))
    else:
        if two_theta_range is None:
            two_theta_range = DEFAULT_TWO_THETA_RANGE

    fig, ax = plt.subplots()
    y_shift = 0
    for structure in structures:
        # Load Pymatgen structure and calulate diffraction pattern
        pmg_struct = structure.get_pymatgen()
        if diffraction_type.lower() in ['xrd', 'x-ray', 'x-rays']:
            diffraction_type_name = 'X-ray diffraction'
            if wavelength is None:
                wavelength = 'CuKa'
            calculator=XRDCalculator(wavelength=wavelength, symprec=0,
                                     debye_waller_factors=None)
        elif diffraction_type.lower() in ['nd', 'neutron', 'neutrons']:
            diffraction_type_name = 'neutron diffraction'
            if wavelength is None:
                wavelength = 2.522
            calculator=NDCalculator(wavelength=wavelength, symprec=0,
                                    debye_waller_factors=None)
        else:
            sys.exit('diffraction_type should be neutron/ND or X-ray/XRD.')

        pattern=calculator.get_pattern(pmg_struct, scaled=True,
                                       two_theta_range=two_theta_range)
        x, y = add_smoothing_functions(pattern, fwhm=fwhm, eta=eta,
                                       two_theta_range=two_theta_range)

        y_range =max(y) - min(y)
        ax.plot(x, y + y_shift)
        y_shift += 1.05 * y_range

    if experimental_file is not None:
        # Adding experimental ND data to plot
        ax.plot(expt_data[:,0],
                expt_data[:,1] / max(expt_data[:,1]) + y_shift,
                color='black')
        if experimental_title is None:
            legend.append('Experiment')
        else:
            legend.append(experimental_title)

    if isinstance(plot_title, str):
        ax.set_title(plot_title)
    else:
        ax.set_title(('Simulated {} patterns').format(
            diffraction_type_name, structure.pk))

    ax.set(xlabel='2\u03B8 (°)',ylabel='Normalized Intensity (AU)',
           xlim=two_theta_range)
    ax.legend(legend,loc='upper left', bbox_to_anchor= (1.0, 1.0))
    plt.tight_layout()
    plt.show()

    return fig

def get_variation(a, a_0):
    return (a-a_0)/a_0

def mean_of_abs_variations(vect, vect_0):
    mean = (abs(vect[0]-vect_0[0])/vect_0[0] +
                  abs(vect[1]-vect_0[1])/vect_0[1] +
                  abs(vect[2]-vect_0[2])/vect_0[2] )/3
    return mean


def measure_cell_disortions(structure, ref_structure, verbosity=1):
    """
    Get a tuple of parameters indicating cell disortortion from a ref structure
    """

    # Convert PKs or UIIDs to StructureData objects
    structure = get_structure_from_pk_or_node(structure)
    ref_structure = get_structure_from_pk_or_node(ref_structure)
    if verbosity >= 2:
        print('Measuring cell distortions between structures {} and {}.'.format(
            structure, ref_structure))

    lgth = structure.cell_lengths
    ang = structure.cell_angles
    lgth_0 = ref_structure.cell_lengths
    ang_0 = ref_structure.cell_angles

    size_disto = mean_of_abs_variations(lgth, lgth_0)
    shape_disto = mean_of_abs_variations(ang, ang_0)
    disto_anisotropy = mean_of_abs_variations(
        [lgth[0]/lgth[1], lgth[1]/lgth[2], lgth[0]/lgth[2]],
        [lgth_0[0]/lgth_0[1], lgth_0[1]/lgth_0[2], lgth_0[0]/lgth_0[2]])

    cell = np.asarray(structure.cell)
    ref_cell = np.asarray(ref_structure.cell)
    cell_corners_msd = 1/3 * np.sum(np.square(np.linalg.norm(cell-ref_cell, axis=0)))

    return (size_disto, shape_disto, disto_anisotropy, cell_corners_msd)


def load_and_store_xrd_pdf_refined_cif_file(filename,
        node_label='AsTe3_refined_structure_cif_file',
        node_description=None,
        group_label='AsTe3_initial_structures',
        group_description=None,
        verbosity=1):

    if not isinstance(node_description, str):
        node_description = ('CIF file of disordered AsTe3 structure ' +
            'obtained from a 1x8x1 supercell of metallic Te by ' +
            'refinement against combined XRD and PDF data ' +
            '(O. Masson, May 2021)')

    if isinstance(group_label, str):
        group_name = group_label
    else:
        # Set default group name
        sys.exit('group_label should be a string.')
    try:
        group = load_group(label=group_name)
        if verbosity >= 1:
            print('Loading existing group {}<{}>'.format(group.label,
                  group.pk))
    except NotExistent:
        group = Group(label=group_name)
        if isinstance(group_description, str):
            group.description = group_description
        else:
            group.description = 'AsTe3 initial structures'
        group.store()
        if verbosity >= 1:
            print('Group {}<{}> has been created.'.format(group.label,
                  group.pk))
    try:
        single_file = SinglefileData(filename)
    except ValueError as e:
        sys.exit(e)
    single_file.label = node_label
    single_file.description = node_description
    single_file.store()
    group.add_nodes(single_file)

    if verbosity >= 1:
         print(('CIF file {} loaded in database as SinglefileData <{}> ' +
                'and stored in group {}<{}>.').format(filename, single_file.pk,
                group.label, group.pk))


def load_and_store_structure_in_group(filename,
        group_or_label,
        node_label=None,
        node_description=None,
        verbosity=1):

    group = get_group_from_str_or_instance(group_or_label)
    try:
        pmg_struct = Structure.from_file(filename)
    except FileNotFoundError as e:
        sys.exit(e)
    structure = StructureData(pymatgen_structure=pmg_struct)
    if isinstance(node_label, str):
        structure.label = node_label
    if isinstance(node_description, str):
        structure.description = node_description
    structure.store()
    group.add_nodes(structure)

    if verbosity >= 1:
         print(('Structure file {} loaded, stored in database as ' +
                'StructureData <{}> (label: {}) and added to group ' +
                '{}<{}>.').format(filename, structure.pk, structure.label,
                group.label, group.pk))
    return structure, group

def load_and_store_refined_cell_structure(filename,
        group_label='AsTe3_initial_structures',
        structure_label='Te_supercell_1x8x1_refined-cell',
        structure_description=None, verbosity=1):
    """
    Load the refined cell structure from a file and store in database
    within a group and a label appropriate for use in other functions

    WARNING: changing structure_label and group_label may have repercussions
    on other functions.

    Args:
        filename: str
            name of (e.g. cif) cile containing the refined structure
        group_label: str (default is 'AsTe3_initial_structures')
            Name of group to store the refined structure
            stick to the default unless you know what you're doing
            this will have repercussions on other functions
        structure_label: str (default is 'Te_supercell_1x8x1_refined-cell')
            Label of refined structure
            stick to the default unless you know what you're doing
            this will have repercussions on other functions
        structure_description: str (default is None)
            The default description is defined within the function

    Returns:
        StructureData, Group
    """
    structure, group = load_and_store_structure_in_group(
        '/home/cadarp02/aiida_projects/AsTe3/reference_structures/Te_sc-181_refined-cell.cif',
        group_label)
    if isinstance(structure_description, str):
        structure.description = structure_description
    else:
        structure.description = ('Model structure based on a 1x8x1 supercell '
            'of Te-metal refined against combined experimental DRX and X-ray '
            'total scattering data collected for AsTe3')
    structure.label = structure_label
    if verbosity >= 1:
        print(('Structure loaded from file :\n{}\nhas been stored as StructureData<{}> '
               'with label {} in group {}<{}>.').format(filename, structure.pk,
              structure.label, group.label, group.pk))

    return structure, group

def get_refined_structure(group_label='AsTe3_initial_structures',
                          description=None):
    if description is None:
        # Make sure it matches the description provided in
        # load_and_store_refined_cell_structure()
        description = ('Model structure based on a 1x8x1 supercell '
            'of Te-metal refined against combined experimental DRX and X-ray '
            'total scattering data collected for AsTe3')
    qb = QueryBuilder()
    qb.append(Group, filters={'label': 'AsTe3_initial_structures'}, tag='group')
    qb.append(StructureData, with_group='group',
          filters={'description': description})
    # pick the most recent in cas several structures with the same description
    # are present
    qb.order_by({StructureData: {'ctime': 'desc'}})  # mtime ???
    refined_structure = qb.first()[0]
    return refined_structure

@calcfunction
def change_lattice_to_reference(structure: StructureData,
                                reference_structure: StructureData,
                                add_random_perturbation=lambda: Bool(False),
                                random_perturbation=lambda: Float(0.05)):
    """
    Create a copy of a StructureData with the lattice of a reference structure

    Args:
        structure: StructureData
            The input structure whose lattice will be modified

        reference_structure: StructureData
            The reference structure whose lattice will be taken and applied to
            to the iunput structure. A warnning will be printed if input and
            reference structure are very different (more than 10% average
            variation on lengths or angles.

        add_random_perturbation: Bool
            Whether a random perturbation should be applied on all atoms
            the pymatgen.core.structure.Structure.perturb() method is used.

        random_perturbation: Float
            Random perturbation value in Angstroms.

    Returns:
        structure: StructureData
            The structure with modified cell parameters and atomic positions
            (if add_random_perturbation is True).
    """
    pmg_struct = structure.get_pymatgen()
    pmg_ref_struct = reference_structure.get_pymatgen()

    # Measure cell distortions and print warning if large
    (size_disto, shape_disto, disto_anisotropy, _) = measure_cell_disortions(
        structure, reference_structure)
    if size_disto > 0.1 or shape_disto > 0.1 or disto_anisotropy > 0.1:
        print(('WARNING: relatively large disortions observed between ' +
              'the cell parameters of the considered structure {} and ' +
              'the reference structure ({}).').format(structure.pk,
              reference_structure.pk))
    pmg_struct.lattice = pmg_ref_struct.lattice
    # Add random perturbation on atomic positions
    if add_random_perturbation:
        pmg_struct.perturb(random_perturbation.value)

    output_structure = StructureData(pymatgen_structure=pmg_struct)
    description = ('cell modified based on reference structure {} ' +
                  '(uuid = {})').format(reference_structure.pk,
                  reference_structure.uuid)
    if  structure.description == '':
        output_structure.description = description
    else:
        output_structure.description = (structure.description + ' - ' +
                                        description)
    return {'structure': output_structure}


def print_process_status(process_node, verbosity=1):
    if verbosity >= 2:
        exit_msg = ''
        if process_node.exit_message is not None:
            exit_msg = ' (' + process_node.exit_message + ')'
        print('{}<{}> finished with exit_status: {}{}.'.format(
              process_node.process_label, process_node.pk,
              process_node.exit_status, exit_msg))


@workfunction
def change_lattice_to_ref_and_assign(structure: StructureData,
        reference_structure: StructureData,
        random_perturbation=lambda: Float(0.05),
        verbosity=lambda: Int(1)):
    """
    Combine lattice change and assignment in a single Workfunction

    The idea is to facilitate database queries later on, to limit the depth
    of recursive searches.

    Args:
        TO BE COMPLETED

    Returns:
        structure: StructureData
            The structure modified with the different processes.
    """

    # Change lattice and assign
    (result, node) = run_get_node(change_lattice_to_reference, structure,
        reference_structure, add_random_perturbation=Bool(True),
        random_perturbation=random_perturbation)
    print_process_status(node, verbosity=verbosity.value)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))
    modified_structure = node.outputs.structure
    # Find original structure previously assigned to initial structure
    original_structure = get_assigned_origin(structure, verbosity=verbosity.value)
    # Forced assignment process
    (result, node) = run_get_node(assign_original_structure,
        modified_structure, is_origin_known=Bool(True),
        original_structure=original_structure, verbosity=verbosity)
    print_process_status(node, verbosity=verbosity.value)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))

    return {'structure': modified_structure}


def find_structures_not_yet_relaxed(initial_group_or_label, relax_node_label,
        relax_process_label='VaspMultiStepRelaxWorkchain', pks_only=True,
        verbosity=1):
    """
    Search structures that have not yet undergone a relaxation process
    identified by its label

    WARNING: first eliminate structures with seemingly succesful relaxation
    processes (exit_status=0) but ideintical input and output structures.
    """

    if verbosity >= 2:
        print(('Looking for structure not-yet relaxed with process {} and '
               'setting {} in group {}.').format(relax_process_label,
              relax_node_label, initial_group_or_label))

    initial_struct_group = get_group_from_str_or_instance(initial_group_or_label)

    qb = QueryBuilder()
    # Search all structures in group
    qb.append(cls=Group, filters={'label': initial_struct_group.label},
              tag='initial_structures_group')
    qb.append(cls=StructureData, with_group='initial_structures_group',
              tag='structures_in_group', project=['id'])
    all_structures_pks = qb.all(flat=True)

    # Search successful relaxation processes performed on structures in the group
    qb.append(cls=WorkChainNode, with_incoming='structures_in_group',
              filters={
                'attributes.process_label': relax_process_label,
                'attributes.exit_status': 0,
                'label': relax_node_label
              },
              tag='relaxation_processes',
              project=['id']
              )
    #DEBBUGGING
    if verbosity >= 2:
        print('{} successfull {} processes with label {} found.'.format(
            qb.count(), relax_process_label, relax_node_label))
    if qb.count() == 0:
        if verbosity >= 1:
            print(('No relaxation process with label {} outgoing from ' +
                   'structures in group {}<{}>.').format(
                  relax_process_label, initial_struct_group.label,
                  initial_struct_group.pk))
        struct_with_relax_proc_pks = []
    else:
        qb.append(cls=StructureData, with_incoming='relaxation_processes',
                  tag='structures_already_relaxed', project=['id'])
        struct_with_relax_proc_pks = [row[0] for row in qb.all(flat=False)]

    struct_to_relax_pks = list(set(all_structures_pks) -
                               set(struct_with_relax_proc_pks))

    if verbosity >= 1:
        print(('Among the {} structures in group {}<{}>, {} have already ' +
               'been relaxed with setting \'{}\' and {} have not.').format(
               len(all_structures_pks), initial_struct_group.label,
               initial_struct_group.pk, len(struct_with_relax_proc_pks),
               relax_node_label, len(struct_to_relax_pks)))

    if pks_only:
        return struct_to_relax_pks
    else:
        return [load_node(pk) for pk in struct_to_relax_pks]


def get_latest_data_dict_in_group(supercell_struct_group_name,
                                  lowest_E_C_data_label, verbosity=1):
    """
    Get the most recent (highest-pk) data Dict with the specified label in a group
    """
    qb = QueryBuilder()
    qb.append(cls=Group, filters={'label': supercell_struct_group_name},
              tag='supercell_struct_group')
    qb.append(cls=Dict, with_group='supercell_struct_group',
              filters={'label': lowest_E_C_data_label},
              tag='lowest_E_C_data_dict', project=['*'])
    qb.order_by({Dict: {'id': 'desc'}})
    if qb.count() == 0:
        sys.exit('No lowest_E_C data with label {} found in group {}.'.format(
            lowest_E_C_data_label, supercell_struct_group_name))
    lowest_E_C_Dict = qb.first()[0]
    if verbosity >= 2:
        print('lowest_E_C_Dict content: ')
        print_dict_of_lists_content(lowest_E_C_Dict)
    return lowest_E_C_Dict


def get_data_indexes_ordered_by_lowest_E_C(lowest_E_C_Dict,
        struct_pks_to_order, verbosity=1):
    """
    Get data indexes for structures listed in a group

    The method is based on the fact that structure pks are ordered
    by increasing lowest-E_C in Dict lowest_E_C_data_label.

    NB: keep in mind that in this particular example the Coulomb energies
    are not obtained directly on the considered structures but by
    """
    # Pick structures based on the lowest-E_C ranking in the
    # lowest_Coulomb_energies_and_info data Dict
    E_C_data_length = len(lowest_E_C_Dict['lowest_E_C'])

    # TODO: The part below is very slow and could probably be optimized.
    data_indexes = [index for index in range(E_C_data_length) if
                  lowest_E_C_Dict['initial_struct_pks'][index] in
                  struct_pks_to_order]

    if verbosity >= 3:
        print('data_indexes = ', data_indexes)

    return data_indexes


def find_structures_not_yet_modified_and_relaxed(initial_group_or_label,
        relax_node_label, modif_process_label, modif_node_label=None,
        relax_process_label='VaspMultiStepRelaxWorkchain', pks_only=True,
        verbosity=1):
    """
    Search structures that have not yet undergone a modification and then a relaxation process

    The method assumes that the structure modification process is performed by
    a single calcfunction or workfunction idendified by its process_label attribute.
    Only structures that hove not yet undergone sequentially both processes
    will be considered.

    WARNING: first eliminate structures with seemingly succesful relaxation
    processes (exit_status=0) but ideintical input and output structures.
    """

    initial_struct_group = get_group_from_str_or_instance(initial_group_or_label)

    qb = QueryBuilder()

    # Search all structures in group
    qb.append(cls=Group, filters={'label': initial_struct_group.label},
              tag='initial_structures_group')
    qb.append(cls=StructureData, with_group='initial_structures_group',
              tag='structures_in_group', project=['id'])
    all_structures_pks = qb.all(flat=True)

    # Search successful modification process
    qb.append(cls=[CalcFunctionNode, WorkFunctionNode],
              with_incoming='structures_in_group',
              filters={
                'attributes.process_label': modif_process_label,
                'attributes.exit_status': 0,
              },
              tag='modification_processes',
              project=['id']
              )
    if isinstance(modif_node_label, str):
        qb.add_filter('modification_processes', {'label': modif_node_label})

    # Search modified structures
    qb.append(cls=StructureData, with_incoming='modification_processes',
              tag='modified_structures', project=['id'])

    # Search successful relaxation processes performed on modified structures
    qb.append(cls=WorkChainNode, with_incoming='modified_structures',
              filters={
                'attributes.process_label': relax_process_label,
                'attributes.exit_status': 0,
                'label': relax_node_label
              },
              tag='relaxation_processes',
              project=['id']
              )

    # Search relaxed structures
    qb.append(cls=StructureData, with_incoming='relaxation_processes',
                  tag='structures_already_relaxed', project=['id', 'ctime'])

    if verbosity >= 2:
        df = pd.DataFrame({
            'structures_in_group': [row[0] for row in qb.all()],
            'modification_processes': [row[1] for row in qb.all()],
            'modified_structures': [row[2] for row in qb.all()],
            'relaxation_processes': [row[3] for row in qb.all()],
            'relaxed_structures': [row[4] for row in qb.all()],
            'relax_struct_ctime': [row[5] for row in qb.all()]})
        print(df)

    if qb.count() == 0:
        if verbosity >= 1:
            print(('No combination of {}({}) and {}({}) processes outgoing from ' +
                   'structures in group {}<{}>.').format(
                  modif_process_label, modif_node_label, relax_process_label,
                  relax_node_label, initial_struct_group.label,
                  initial_struct_group.pk))
        struct_with_relax_proc_pks = []
    else:
        struct_with_relax_proc_pks = [row[0] for row in qb.all(flat=False)]

    struct_to_relax_pks = list(set(all_structures_pks) -
                               set(struct_with_relax_proc_pks))

    if verbosity >= 1:
        print(('Among the {} structures in group {}<{}>, {} have already ' +
               'been modified and then relaxed with setting \'{}\' and {} have not.').format(
               len(all_structures_pks), initial_struct_group.label,
               initial_struct_group.pk, len(struct_with_relax_proc_pks),
               relax_node_label, len(struct_to_relax_pks)))

    if pks_only:
        return struct_to_relax_pks
    else:
        return [load_node(pk) for pk in struct_to_relax_pks]


def order_relaxed_structures_by_energy(relaxed_struct_pks, relax_node_label,
        relax_process_label='VaspMultiStepRelaxWorkchain', verbosity=1):

    qb = QueryBuilder()
    qb.append(cls=StructureData, filters={'id': {'in': relaxed_struct_pks}},
              tag='selected_structures', project=['id'])

    if verbosity >= 2:
        print('{} structures identified in database from their PKs.'.format(
            qb.count()))
    qb.append(cls=WorkChainNode, with_outgoing='selected_structures',
              filters={
                  'attributes.process_label': relax_process_label,
                  'attributes.exit_status': 0,
                  'label': relax_node_label,
              },
              tag='relaxation_processes',
              project=['*']
              )

    if verbosity >= 1:
        tic = time.perf_counter()
        print(('Fetching final energies for {} structures relaxed with '
               'setting {}... please wait.').format(qb.count(), relax_node_label))
    data = {
        'struct_pks': [row[0] for row in qb.all(flat=False)],
        'energy': [row[1].outputs.misc['total_energies']['energy_extrapolated']
                   for row in qb.all(flat=False)]
    }
    if verbosity >= 1:
        print('Done. Processing {} structures took {:.2f} s.'.format(
              qb.count(), time.perf_counter()-tic))
        tic = time.perf_counter()
        print(('Sorting {} structures relaxed with setting {} by energy. ' +
            'Please wait.').format(qb.count(), relax_node_label))
    sorted_indexes = list(np.argsort(np.asarray(data['energy'])))
    sorted_pks = [data['struct_pks'][index] for index in sorted_indexes]
    if verbosity >= 1:
        print('Done. Process took {:.2f} s.'.format(time.perf_counter()-tic))
    if verbosity >= 2:
        sorted_energies = [data['energy'][index] for index in sorted_indexes]
        for index, (pk, energy) in enumerate(zip(sorted_pks, sorted_energies)):
            if index < 5 or index > len(sorted_pks)-2:
                print('Rank {}: Structure {} with E = {:.2f} eV'.format(
                    index, pk, energy))
            elif index == 6:
                print('...')

    return sorted_pks


def build_and_submit_multistep_relax(structure: StructureData,
                                     multistep_input_param: Dict,
                                     relax_code: Code,
                                     relax_potential_family,
                                     relax_potential_mapping,
                                     relax_options):
    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    # Define an empty KpointsData object. Not used at this stage but needed
    # because vaspMultiStepRelaxWorkChain is based on aiida_vasp.RelaxWorkChain
    # THIS WILL HAVE TO BE CHANGED AT SOME POINT
    kpoints_data = KpointsData()
    kpoints_data.set_kpoints_mesh([0, 0, 0])

    # Set the workchain to call. No () : we use the class rather than instance
    workchain = VaspMultiStepRelaxWorkchain

    # Declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()
    # organize settings
    settings.parser_settings = {'output_params': ['total_energies',
                                                  'maximum_force'],
                                'add_maximum_stress': False}
    # set inputs for the following WorkChain execution
    inputs.code = relax_code
    # set structure
    inputs.structure = structure

    inputs.kpoints = kpoints_data  # Removed from expected inputs

    # set parameters
    # These are replaced at this stage by the multistep_input_param
    # and will be loaded later based on the latter.
    # inputs.parameters = dict_data(dict=incar)
    inputs.multistep_input_param = dict_data(dict=multistep_input_param)

    # set potentials and their mapping
    inputs.potential_family = Str(relax_potential_family)
    inputs.potential_mapping = dict_data(dict=relax_potential_mapping)
    # set options
    inputs.options = dict_data(dict=relax_options)
    # set settings
    inputs.settings = dict_data(dict=settings)
    # set workchain related inputs
    inputs.verbose = Bool(True)  # For testing purposes

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.
    # submit(workchain, **inputs)
    # Use run() command for testing purposes
    # run(workchain, **inputs)
    multistep_relax_node = submit(workchain, **inputs)
    return multistep_relax_node


def get_default_potential_mapping(structure, return_aiida_data_type=False):
    """
    Return a potential mapping of the form {'As': 'As', 'Te': 'Te', ...}
    based on structure composition

    Args:
        structure: StructureData
            The structure from which composition will be obtained
        return_aiida_data_type: bool (default: False)
            whether to return the potential mapping
    """
    potential_mapping = {}
    for kind in structure.get_kind_names():
        potential_mapping[kind] = kind
        # Add exceptions for selected elements requiring other potentials
        if return_aiida_data_type:
            potential_mapping=Dict(dict=potential_mapping)
    return potential_mapping


def build_and_submit_md(structure: StructureData,
                        parameters=None,
                        code=None, computer_label='jean-zay', code_label='vasp-5.4.4',
                        test_mode=False, options=None, potential_family=None,
                        potential_mapping=None,
                        wall_time_s=None, nb_of_nodes=1,
                        kpoints=None, kpoints_density=0.5, kpoints_offset=None,
                        max_ionic_steps=None, default_max_ionic_steps=1000, node_label='',
                        dry_run=False, verbosity=1):

    # Fetch the AiiDA datatypes which will house calculation inputs
    dict_data = DataFactory('dict')

    # set inputs for the following WorkChain execution
    # Set the workchain to call. No () : we use the class rather than instance
    workchain = MolecularDynamicsWorkChain

    inputs = AttributeDict()
    # set structure
    inputs.structure = structure

    if (isinstance(code, Code) and isinstance(potential_family, Str)
        and isinstance(potential_mapping, Dict) and
        isinstance(options, Dict)):
        inputs.code=code
        inputs.potential_family = potential_family
        inputs.potential_mapping = potential_mapping
        inputs.options = options

    else:
        if potential_mapping is None:
            potential_mapping = get_default_potential_mapping(structure,
                return_aiida_data_type=False)

        code_and_computer_data = get_vasp_code_and_computer_data(
            code_label='vasp-5.4.4', computer_label='jean-zay',
            wall_time_s=wall_time_s, nb_of_nodes=nb_of_nodes,
            test_mode=test_mode, potential_mapping=potential_mapping,
            verbosity=verbosity)
        inputs.code=code_and_computer_data['code']
        inputs.potential_family = Str(code_and_computer_data['potential_family'])
        inputs.potential_mapping = Dict(dict=code_and_computer_data['potential_mapping'])
        inputs.options = Dict(dict=code_and_computer_data['options'])

    if isinstance(parameters, Dict):
        inputs.parameters=parameters
    elif isinstance(parameters, dict):
        inputs.parameters=Dict(dict=parameters)
    else:
        # Read default parameters
        if test_mode and max_ionic_steps is None:
            max_ionic_steps=2
        elif max_ionic_steps is None or not isinstance(max_ionic_steps, int):
            if verbosity >= 2:
                print(('Using the default maximum number of ionic steps '
                       '(VASP INCAR \'nsw\'): {}').format(default_max_ionic_steps))
            max_ionic_steps=default_max_ionic_steps

        inputs.parameters=Dict(dict={
            'system': '{} - structure {} (uuid={}), NVT MD with Nose-Hoover thermostat at 300 K'.format(structure.get_formula(),
                structure.pk, structure.uuid),
            'encut': 270,
            'ediff': 1e-5,
            'prec': 'Normal',
            'ismear': 0,
            'sigma': 0.05,
            'ncore': 10,
            'lreal': 'Auto',
            'isif': 2,
            'nsw': max_ionic_steps,
            'ibrion': 0,
            'potim': 1.0,
            'tebeg':300,
            'mdalgo': 2,
            'smass': 0,
            'nblock': 1 })
        # DEBUGGING
        print('inputs.parameters = ', inputs.parameters.get_dict())
    """
    inputs.relax = Dict(dict={
        'perform': True
        'steps':
    })
    """
    # Define an empty KpointsData object. Not used at this stage but needed
    # because vaspMultiStepRelaxWorkChain is based on aiida_vasp.RelaxWorkChain
    # THIS WILL HAVE TO BE CHANGED AT SOME POINT
    if isinstance(kpoints, KpointsData):
        inputs.kpoints = kpoints
    else:
        kpoints_data = KpointsData()
        kpoints_data.set_cell_from_structure(structure)
        kpoints_data.set_kpoints_mesh_from_density(kpoints_density,
                                                   offset=kpoints_offset)
        inputs.kpoints = kpoints_data

    # Declare and organize settings: add trajectory
    settings = AttributeDict()
    settings.parser_settings = {'output_params': ['total_energies',
                                                  'maximum_force'],
                                'add_trajectory': True,
                                'add_forces': True,
                                'add_energies': True,
                                'add_poscar-structure': True}
    inputs.settings = Dict(dict=settings)

    if verbosity > 1:
        inputs.verbose = Bool(True)  # For testing purposes

    # submit the requested workchain with the supplied inputs
    # The submit() command will return an ImportError exception if the folder
    # containing the module is not automatically included in the PYTHONPATH via
    # the environment activation script.
    # Please mind that the search for modules is not recursive.
    # submit(workchain, **inputs)
    # Use run() command for testing purposes
    # run(workchain, **inputs)
    if dry_run:
        print('Ready to run molecular dynamics with the following inputs: ')
        for k, v in inputs.items():
            print(f'{k}: {v}')
        return None
    else:
        md_node = submit(workchain, **inputs)
        if isinstance(node_label, str):
            md_node.label = node_label
        if verbosity >= 2:
            print('Molecular dynamics submitted under {}<{}> with label \'{}\'.'.format(
                md_node.process_label, md_node.pk, md_node.label))
        return md_node

def get_relax_param_file(relax_setting='test', relax_param_dir=None,
                         verbosity=1):
    """
    Get csv files containing the multi-step relaxation parameters
    """
    if relax_param_dir is None:
        relax_param_dir = os.path.join(os.getcwd(), 'relax_param_files')

    relax_param_file = os.path.join(relax_param_dir,
                                    'multistep_input_param_{}.csv'.format(relax_setting))
    if not os.path.isfile(relax_param_file):
        raise FileNotFoundError('File {} not found.'.format(relax_param_file))

    if verbosity >= 2:
        print('Relaxation may be read from file: {}.'.format(
              relax_param_file))

    return relax_param_file


def workflow_queue(maxrun_workflows, running_workflows_list,
                   sleep_time_in_mn=10):
    """
    Limit the number of jobs submitted to the scheduler

    Args :
        maxrun_workflows : Int
            the maximum number of jobs sent to the computer scheduler
        running_workflows_list : list
            a list of workflows submitted to the dameon
        sleep_time_in_mn : Int, default = 10
            Sleep time in minutes when number of submitted jobs reaches the
            limit.
    Returns :
        None

    See : https://groups.google.com/g/aiidausers/c/NjNvF8noeP0/m/7ZU2oPxeBwAJ
    USAGE :
    running_workflows_list = []
    for inputs in inputs_list:
        workflow_node = submit(WorkchainClass, **inputs)
        running_workflows_list.append( workflow_node )
        workflow_queue(30, running_workflows_list)

    """
    while len(running_workflows_list) >= maxrun_workflows:
        print('{} or more jobs in queue. Waiting {} min...'.format(
            maxrun_workflows, sleep_time_in_mn))
        time.sleep(sleep_time_in_mn*60)
        for workflow in running_workflows_list.copy():
              if workflow.is_finished or workflow.is_excepted:
                running_workflows_list.remove(workflow)

    return


def prepare_and_submit_rough_fixed_cell_relax(relax_code,
        relax_potential_family, relax_potential_mapping, relax_options,
        relax_param_file, max_nb_of_relax_jobs,
        nb_of_lowest_E_C_structs, nb_of_random_structs,
        nb_of_highest_E_C_structs,
        relax_setting_label='fixed_cell_rough',
        supercell_struct_group_name= 'AsTe3_Te-based_sc181_As-Te-configurations_no-opt',
        lowest_E_C_data_label='lowest_E_C_data',
        dry_run=False, verbosity=2):
    """
    Search yet-un-relaxed structures in a group, pick them and run relaxations

    IMPLEMENTATION IN PROGRESS
    """
    # Use a database query to find structures not-yet optimized in group
    # (may be done in a separate function)
    not_yet_opt_struct_pks = find_structures_not_yet_relaxed(
        supercell_struct_group_name, relax_node_label=relax_setting_label,
        pks_only=True, verbosity=verbosity)

    lowest_E_C_Dict = get_latest_data_dict_in_group(supercell_struct_group_name,
        lowest_E_C_data_label, verbosity=verbosity)

    # Order structures by lowest-E_C in a separate function
    if verbosity >= 2:
        print(('Sorting not-yet optimized structures in group {} using '
               'the lowest-E_C ranking stored in data Dict {}<{}> therein.'
               '\n... Process can be long. Please be patient...'
               ).format(supercell_struct_group_name, lowest_E_C_Dict.label,
                        lowest_E_C_Dict.pk))
    sorted_data_indexes = get_data_indexes_ordered_by_lowest_E_C(
        lowest_E_C_Dict, not_yet_opt_struct_pks, verbosity=verbosity)

    # pick lowest-E_C, random and highest-E_C structures among not-yet-opt struct.
    picked_data = {}
    lowest_E_C_indexes = sorted_data_indexes[:nb_of_lowest_E_C_structs]

    # TODO: replace random by seed-controlled random choice with numpy random number generator
    random_indexes = random.sample(
        sorted_data_indexes[(nb_of_lowest_E_C_structs):
                            -nb_of_highest_E_C_structs],
        nb_of_random_structs)
    highest_E_C_indexes = sorted_data_indexes[-nb_of_highest_E_C_structs:]
    picked_indexes = lowest_E_C_indexes + random_indexes + highest_E_C_indexes
    for key, value in lowest_E_C_Dict.attributes_items():
        picked_data[key] = [value[index] for index in picked_indexes]
    if verbosity >= 2:
        print('picked_data = ', picked_data)

    # load multi-step relaxation parmaters from file.
    multistep_param = get_multistep_input_param_from_csv(
        relax_param_file)

    submitted_process_pks = []
    running_workflows_list = []
    for struct_pk in picked_data['initial_struct_pks']:
        if dry_run:
            print(('DRYRUN: Ready to run VaspMultiStepRelaxWorkchain '
                   'process with setting {} on structure {}.'
                   ).format(relax_setting_label, struct_pk))
        else:
            structure = load_node(struct_pk)
            # Optimize each As/Te configuration
            # WARNING : PROCEED CAREFULLY WHEN INCREASING NUMBER OF CALCULATIONS
            # - NUMBER OF DAEMON USERS
            # - INODES ON REMOTE MACHINE (USE '-write-on-scratch' COMPUTERS)
            # - AUTOMATIC RESTART
            workflow_queue(max_nb_of_relax_jobs, running_workflows_list,
                           sleep_time_in_mn=1)
            multistep_relax_node = build_and_submit_multistep_relax(
                structure, multistep_param, relax_code,
                relax_potential_family, relax_potential_mapping, relax_options)
            multistep_relax_node.label = relax_setting_label
            submitted_process_pks.append(multistep_relax_node.pk)
            print(('{}<{}> : submitting multi-step relaxation with setting ' +
                   '\'{}\' on structure {}').format(
                   multistep_relax_node.process_label, multistep_relax_node.pk,
                   multistep_relax_node.label, structure.pk))
            # Limit number of jobs submitted to scheduler
            running_workflows_list.append(multistep_relax_node)

    return submitted_process_pks


def get_vasp_code_and_computer_data(code_label='vasp-5.4.4',computer_label='jean-zay',
                                    wall_time_s=None, nb_of_nodes=1, test_mode=True,
                                    potential_mapping={'Te': 'Te', 'As': 'As'},
                                    verbosity=1):
    """
    prepare a dict default calculation parameters
    build_and_submit_multistep_relax(structure: StructureData,
                                     multistep_input_param: Dict,
                                     relax_code: Code,
                                     relax_potential_family,
                                     relax_potential_mapping,
                                     relax_options):
    Args:
        code_label: str (default is 'vasp-5.4.4')
            code prefix in 'code@computer' listed by 'verdi code list' command
        computer_label: str (default is 'vasp-5.4.4')
            computer suffix in code@computer listed by 'verdi code list'
        wall_time_s: int (default is None)
            The default will be adapted
    """
    data_dict = {
        'code': Code.get_from_string('{}@{}'.format(code_label,
                                                    computer_label)),
        'potential_family': 'PBE.54',
        'potential_mapping': potential_mapping}
    # TODO: create an automatic default potential mapping from a structure
    # something like:
    # {specie.name: specie.name for specie in structure.species}
    options = AttributeDict()
    if computer_label.lower() == 'jean-zay':
        options.account = 'def@cpu'
        options.resources = {'num_machines': nb_of_nodes,
                                   'num_mpiprocs_per_machine': 40}
        if wall_time_s is None:
            options.max_wallclock_seconds = 7200 # 02:00:00
        else:
            options.max_wallclock_seconds = wall_time_s

        if test_mode and options.max_wallclock_seconds <= 7200:
            options.qos = 'qos_cpu-dev'
        elif options.max_wallclock_seconds > 72000 and nb_of_nodes <= 4:
            options.qos = 'qos_cpu-t4'
        elif options.max_wallclock_seconds <= 72000:
            options.qos = 'qos_cpu-t3'
        else:
            options.max_wallclock_seconds = 72000
            options.qos = 'qos_cpu-t3'
            print(('WARNING: inappropriate resource request for computer {}. '
                   'qos set to {} with a wall time of {} s. See Jean-Zay '
                   'website for QoS and associated limitations.').format(
                  computer_label, options.qos, options.max_wallclock_seconds))
        options.queue_name = ''

    elif computer_label.lower() == 'curta':
        options.account = ''
        options.qos = ''
        options.resources = {'num_machines': nb_of_nodes,
                             'num_mpiprocs_per_machine': 32}
        if test_mode and nb_of_nodes == 1:
            options.queue_name = 'visu'
        else:
            options.queue_name = 'ircer'

    data_dict['options'] = options
    if verbosity >= 2:
        for k, v in data_dict.items():
            print('{}: {}'.format(k, v))

    return data_dict


def prepare_and_submit_test_relax(input_test_structures, multistep_input_param_file=None,
                                  relax_setting='test', relax_param_dir=None,
                                  code_label='vasp-5.4.4', computer_label='jean-zay',
                                  wall_time_s=None, nb_of_nodes=1,
                                  potential_mapping={'Te': 'Te', 'As': 'As'},
                                  process_node_labels='test',
                                  convert_to_single=True, test_mode=True,
                                  verbosity=1):
    """
    TO BE COMPLETED
    """
    # Convert single structure (or id) to list
    if not isinstance(input_test_structures, (list, tuple)):
        input_test_structures = [input_test_structures]
    data_dict = get_vasp_code_and_computer_data(code_label=code_label,
        computer_label=computer_label, wall_time_s=wall_time_s,
        nb_of_nodes=nb_of_nodes, test_mode=test_mode,
        potential_mapping=potential_mapping, verbosity=verbosity)
    input_param_file = get_relax_param_file(relax_setting=relax_setting,
                                            relax_param_dir=relax_param_dir)
    multistep_input_param = get_multistep_input_param_from_csv(input_param_file)
    structures = [get_structure_from_pk_or_node(s) for s in input_test_structures]
    if isinstance(process_node_labels, str):
        process_node_labels = [process_node_labels] * len(structures)

    multistep_relax_nodes = []
    for index, structure in enumerate(structures):
        process_node = build_and_submit_multistep_relax(structure,
                multistep_input_param,
                data_dict['code'], data_dict['potential_family'],
                data_dict['potential_mapping'], data_dict['options'])
        process_node.label = process_node_labels[index]
        multistep_relax_nodes.append(process_node)

    if convert_to_single and len(multistep_relax_nodes) == 1:
        # Convert list of length 1 to single element
        [multistep_relax_nodes] = multistep_relax_nodes

    return multistep_relax_nodes


def prepare_and_submit_relax(relax_code,
        relax_potential_family, relax_potential_mapping,
        relax_options, relax_param_file,
        max_nb_of_relax_jobs, initial_struct_group_name,
        nb_of_structures,
        previous_relax_setting='fixed_cell_rough',
        relax_setting='free_cell_accurate',
        dry_run=False, verbosity=2):
    """
    Search lowest-E structures relaxed with one setting to run another without structure modification
    """

    # Find structures relaxed with previous settings and yet-to-relax
    # with the current setting.
    struct_to_relax_pks = find_structures_not_yet_relaxed(
        initial_struct_group_name, relax_node_label=relax_setting,
        verbosity=verbosity)
    struct_to_relax_pks = order_relaxed_structures_by_energy(
        struct_to_relax_pks, relax_node_label=previous_relax_setting,
        verbosity=verbosity)
    # Pick the first nb_of_structures
    picked_struct_pks = struct_to_relax_pks[:nb_of_structures]

    # load multi-step relaxation parmaters from file.
    multistep_param = get_multistep_input_param_from_csv(
        relax_param_file)

    submitted_process_pks = []
    running_workflows_list = []
    for index, struct_pk in enumerate(picked_struct_pks):
        if dry_run:
            print(('Ready to run VaspMultiStepRelaxWorkchain process ' +
                   'with setting {} on structure {}.').format(
                  relax_setting, struct_pk))
        else:
            structure = load_node(struct_pk)
            # prepare and run relaxation
            workflow_queue(max_nb_of_relax_jobs, running_workflows_list,
                           sleep_time_in_mn=1)
            multistep_relax_node = build_and_submit_multistep_relax(
                structure, multistep_param, relax_code,
                relax_potential_family, relax_potential_mapping, relax_options)
            multistep_relax_node.label = relax_setting
            submitted_process_pks.append(multistep_relax_node.pk)
            print(('{}<{}> : submitting multi-step relaxation with setting ' +
                  '\'{}\' on structure {}').format(
                multistep_relax_node.process_label, multistep_relax_node.pk,
                multistep_relax_node.label, structure.pk))
            # Limit number of jobs submitted to scheduler
            running_workflows_list.append(multistep_relax_node)

    return submitted_process_pks


def prepare_and_submit_fixed_refined_cell_relax(relax_code,
        relax_potential_family, relax_potential_mapping,
        relax_options, relax_param_file,
        max_nb_of_relax_jobs, initial_struct_group_name,
        nb_of_structures,
        previous_relax_setting='free_cell_accurate',
        relax_setting='fixed_cell_accurate',
        reference_struct_group_label = 'AsTe3_initial_structures',
        reference_structure_label='Te_supercell_1x8x1_refined-cell',
        random_perturbation=0.05,
        dry_run=False, verbosity=2):
    """
    Prepare a series of calculations with cell parameters fixed
    to those of a reference 'refined' structure

    TOBECOMPLETED

    Args:

        supercell_sizes: list (or tuple) of list or (tuple)
            A calculation will be prepared for every supercell_size in supercell_sizes
            where supercell_size is an array (or list or tuple) of 3 integer values
            repersenting the supercell scaling along each respective cell vector.

    """

    # Find structures relaxed with previous settings and yet-to-relax
    # with the current setting.
    struct_to_relax_pks = find_structures_not_yet_modified_and_relaxed(
        initial_struct_group_name, relax_node_label=relax_setting,
        modif_process_label='change_lattice_to_ref_and_assign',
        verbosity=verbosity)
    struct_to_relax_pks = order_relaxed_structures_by_energy(
        struct_to_relax_pks, relax_node_label=previous_relax_setting,
        verbosity=verbosity)
    # Pick the first nb_of_structures
    picked_struct_pks = struct_to_relax_pks[:nb_of_structures]

    # load multi-step relaxation parmaters from file.
    multistep_param = get_multistep_input_param_from_csv(
        relax_param_file)

    # Load reference structure from group:
    qb = QueryBuilder()
    qb.append(Group, filters={'label': reference_struct_group_label},
              tag='group')
    qb.append(StructureData, with_group='group',
              filters={'label': reference_structure_label},
              tag='refined_Te_structure')
    reference_structure = qb.all(flat=True)[0]
    if verbosity >= 1:
        print(('Reference structure {} with label {} has been loaded ' +
               'from group {}.').format(reference_structure.pk,
               reference_structure.label, reference_struct_group_label))

    submitted_process_pks = []
    running_workflows_list = []
    for index, struct_pk in enumerate(picked_struct_pks):
        if dry_run:
            print(('Ready to change lattice of structure {}, and ' +
                   'run VaspMultiStepRelaxWorkchain process ' +
                   'with setting {}.').format(struct_pk, relax_setting))
        else:
            # TODO: Test whether modification process already exists for this structure

            # Change picked structure lattice to reference structure
            result, node = run_get_node(change_lattice_to_ref_and_assign,
                load_node(struct_pk), reference_structure,
                random_perturbation=Float(random_perturbation),
                verbosity=Int(verbosity))
            structure = result['structure']

            # prepare and run relaxation
            workflow_queue(max_nb_of_relax_jobs, running_workflows_list,
                           sleep_time_in_mn=1)
            multistep_relax_node = build_and_submit_multistep_relax(
                structure, multistep_param, relax_code,
                relax_potential_family, relax_potential_mapping, relax_options)
            multistep_relax_node.label = relax_setting
            submitted_process_pks.append(multistep_relax_node.pk)
            print(('{}<{}> : submitting multi-step relaxation with setting ' +
                  '\'{}\' on structure {}').format(
                multistep_relax_node.process_label, multistep_relax_node.pk,
                multistep_relax_node.label, structure.pk))
            # Limit number of jobs submitted to scheduler
            running_workflows_list.append(multistep_relax_node)

    return submitted_process_pks


def prepare_and_submit_fixed_supercell_relax(relax_code,
        relax_potential_family, relax_potential_mapping,
        relax_options, relax_param_file,
        max_nb_of_relax_jobs, initial_struct_group_name,
        nb_of_structures, supercell_sizes,
        previous_relax_setting='fixed_cell_accurate',
        relax_setting='fixed_supercell_accurate',
        reference_struct_group_label = 'AsTe3_initial_structures',
        reference_structure_label='Te_supercell_1x8x1_refined-cell',
        random_perturbation=0.05,
        dry_run=False, verbosity=1):
    """
    Prepare a series of calculations using supercells from best structures
    previously relaxed with another setting

    TO BE COMPLETED
    """

    # Find structures relaxed with previous settings and yet-to-relax
    # with the current setting.
    # WARNING: As is it currently written, the method below will not
    # take the considered supercell size into account.
    struct_to_relax_pks = find_structures_not_yet_modified_and_relaxed(
        initial_struct_group_name, relax_node_label=relax_setting,
        modif_process_label='change_lattice_make_supercell_and_assign',
        verbosity=verbosity)
    struct_to_relax_pks = order_relaxed_structures_by_energy(
        struct_to_relax_pks, relax_node_label=previous_relax_setting,
        verbosity=verbosity)
    # Pick the first nb_of_structures
    picked_struct_pks = struct_to_relax_pks[:nb_of_structures]

    # load multi-step relaxation parmaters from file.
    multistep_param = get_multistep_input_param_from_csv(
        relax_param_file)

    # Load reference structure from group:
    qb = QueryBuilder()
    qb.append(Group, filters={'label': reference_struct_group_label},
              tag='group')
    qb.append(StructureData, with_group='group',
              filters={'label': reference_structure_label},
              tag='refined_Te_structure')
    reference_structure = qb.all(flat=True)[0]
    if verbosity >= 1:
        print(('Reference structure {} with label {} has been loaded ' +
               'from group {}.').format(reference_structure.pk,
               reference_structure.label, reference_struct_group_label))

    submitted_process_pks = []
    running_workflows_list = []
    for index, struct_pk in enumerate(picked_struct_pks):
        if dry_run:
            print(('Ready to change lattice of structure {}, and ' +
                   'and run VaspMultiStepRelaxWorkchain process ' +
                   'with setting {}.').format(struct_pk, relax_setting))
        else:
            for supercell_size in supercell_sizes:
                if load_node(struct_pk).cell == reference_structure.cell:
                    # Make supercell without changing lattice
                    result, make_sc_node = run_get_node(make_supercell_and_assign,
                        load_node(struct_pk), a=Int(supercell_size[0]),
                        b=Int(supercell_size[1]), c=Int(supercell_size[2]),
                        random_perturbation=Float(random_perturbation),
                        verbosity=Int(verbosity))
                else:
                    # Change picked structure lattice to reference structure
                    if verbosity >= 2:
                        print('Unit cells of structure {} and reference structure {} differ.'.format(
                            struct_pk, reference_structure.pk))
                    result, make_sc_node = run_get_node(change_lattice_make_supercell_and_assign,
                        load_node(struct_pk), reference_structure,
                        a=Int(supercell_size[0]), b=Int(supercell_size[1]),
                        c=Int(supercell_size[2]),
                        random_perturbation=Float(random_perturbation),
                        verbosity=Int(verbosity))
                structure = result['structure']

                # prepare and run relaxation
                workflow_queue(max_nb_of_relax_jobs, running_workflows_list,
                               sleep_time_in_mn=1)
                multistep_relax_node = build_and_submit_multistep_relax(
                    structure, multistep_param, relax_code,
                    relax_potential_family, relax_potential_mapping, relax_options)
                multistep_relax_node.label = relax_setting
                submitted_process_pks.append(multistep_relax_node.pk)
                print(('{}<{}> : submitting multi-step relaxation with setting ' +
                       '\'{}\' on {}x{}x{} supercell structure {}').format(
                       multistep_relax_node.process_label, multistep_relax_node.pk,
                       multistep_relax_node.label, make_sc_node.inputs.a.value,
                       make_sc_node.inputs.b.value, make_sc_node.inputs.c.value,
                       structure.pk))
                # Limit number of jobs submitted to scheduler
                running_workflows_list.append(multistep_relax_node)

    return submitted_process_pks


@calcfunction
def make_supercell(structure: StructureData, a=lambda: Int(1), b=lambda: Int(1),
                   c=lambda: Int(1), verbosity=lambda: Int(1)):
    """
    Create a copy of a StructureData with the lattice of a reference structure
    """
    pmg_struct = structure.get_pymatgen()
    pmg_struct.make_supercell([a, b, c])

    output_structure = StructureData(pymatgen_structure=pmg_struct)
    description = ('{}\u00D7{}\u00D7{} supercell of structure {} ' +
                  '(uuid = {})').format(a, b, c, structure.pk,
                  structure.uuid)
    if  structure.description == '':
        output_structure.description = description
    else:
        output_structure.description = (description + ': ' +
            structure.description)

    if verbosity >= 1:
        print(('Structure {} made from a {}\u00D7{}\u00D7{} supercell ' +
               'of structure {}.').format(output_structure.uuid, a.value,
               b.value, c.value, structure.pk))

    return {'structure': output_structure}


@workfunction
def change_lattice_make_supercell_and_assign(structure: StructureData,
        reference_structure: StructureData,
        a:Int, b:Int, c:Int, random_perturbation=lambda: Float(0.05),
        verbosity=lambda: Int(1)):
    """
    Combine lattice change, supercell and assginment in a single Workfunction

    The idea is to facilitate database queries later on, to limit the depth
    of recursive searches.

    Args:

    Returns:
        structure: StructureData
            The structure modified with the different processes.
    """

    # Change lattice and assign
    (result, node) = run_get_node(change_lattice_to_reference, structure,
        reference_structure, add_random_perturbation=Bool(True),
        random_perturbation=random_perturbation)
    print_process_status(node, verbosity=verbosity.value)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))
    modified_structure = node.outputs.structure
    # Find original structure previously assigned to initial structure
    original_structure = get_assigned_origin(structure, verbosity=verbosity.value)
    # Forced assignment process
    (result, node) = run_get_node(assign_original_structure,
        modified_structure, is_origin_known=Bool(True),
        original_structure=original_structure, verbosity=verbosity)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))

    # Make supercell and assign
    (result, node) = run_get_node(make_supercell, modified_structure,
                                  a=a, b=b, c=c, verbosity=verbosity)
    print_process_status(node, verbosity=verbosity.value)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))
    modified_structure = node.outputs.structure
    # Forced assignment process
    (result, node) = (result, node) = run_get_node(assign_original_structure,
        modified_structure, is_origin_known=Bool(True),
        original_structure=original_structure, verbosity=verbosity)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))

    return {'structure': modified_structure}


@workfunction
def make_supercell_and_assign(structure: StructureData,
        a:Int, b:Int, c:Int, random_perturbation=lambda: Float(0.05),
        verbosity=lambda: Int(1)):
    """
    Combine make supercell and assginment in a single Workfunction

    The idea is to facilitate database queries later on, to limit the depth
    of recursive searches.

    Args:

    Returns:
        structure: StructureData
            The structure modified with the different processes.
    """
    # Find original structure previously assigned to initial structure
    original_structure = get_assigned_origin(structure, verbosity=verbosity.value)

    # Make supercell and assign
    (result, node) = run_get_node(make_supercell, structure,
                                  a=a, b=b, c=c, verbosity=verbosity)
    print_process_status(node, verbosity=verbosity.value)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))
    modified_structure = node.outputs.structure
    # Forced assignment process
    (result, node) = (result, node) = run_get_node(assign_original_structure,
        modified_structure, is_origin_known=Bool(True),
        original_structure=original_structure, verbosity=verbosity)
    if node.exit_status:
        return ExitCode(node.exit_status, '{} (inherited from {}<{}>)'.format(
            node.exit_message, node.process_label, node.pk))

    return {'structure': modified_structure}


def get_assigned_origin(struct_or_pk, verbosity=1):
    """
    Return original structure if assigned, None otherwise
    """
    # Get structure pk from struct_or_pk
    struct = get_structure_from_pk_or_node(struct_or_pk)

    qb = QueryBuilder()
    qb.append(StructureData, filters={'id': struct.pk}, tag='input_structure')
    qb.append(WorkFunctionNode, with_incoming='input_structure',
              filters={
                'attributes.process_label': 'assign_original_structure',
                'attributes.process_state': 'finished',
                'attributes.exit_status': 0
              },
              tag='assignment_process')
    if not qb.count():
        if verbosity >= 2:
            print('Structure {} has no assigned origin.'.format(
                struct.pk))
        return None
    else:
        qb.append(StructureData, with_incoming='assignment_process',
                  tag='original_structure', project='*')
        if qb.count() == 1:
            assigned_origin = qb.all(flat=True)[0]
            if verbosity >= 1:
                print('Assigned origin of structure {}: {}'.format(
                      struct.pk, assigned_origin.pk))

    return assigned_origin


def explore_relax_process(relax_node, verbosity=1):
    relax_node = get_processes_from_nodes_or_ids(relax_node, convert_to_single=True)
    print('{}<{}> with initial structure {}: exit status {} - {}.'.format(
        relax_node.process_label, relax_node.pk, relax_node.inputs.structure.pk,
        relax_node.exit_status, relax_node.exit_message))
    try:
        if relax_node.outputs.relax.structure.pk == relax_node.inputs.structure.pk:
            print('{}<{}>: initial and relaxed structure are the same: {}.'.format(
                relax_node.process_label, relax_node.pk, relax_node.inputs.structure.pk))
    except NotExistentAttributeError:
        print('{} process {} has no relax.structure output.'.format(
            relax_node.process_label, relax_node.pk))

    # Sort called_descendants by PK
    sorted_indexes = np.argsort(np.asarray(
        [descendant.pk for descendant in relax_node.called_descendants]))
    sorted_descendants = [relax_node.called_descendants[index] for index in sorted_indexes]
    for descendant in sorted_descendants:
        if descendant.process_label == 'RelaxWorkChain':
            print('  {}<{}>: {} - {}.'.format(descendant.process_label,
                  descendant.pk, descendant.exit_status, descendant.exit_message))
            if not descendant.exit_status:
                try:
                    if descendant.outputs.structure == descendant.inputs.structure:
                        print(('  {} process {} has identical input and output ' +
                               'structure: {}.').format(descendant.process_label,
                               descendant.pk, descendant.inputs.structure.pk))
                    else:
                        print(('  Input structure: {} ; output structure: {}').format(
                               descendant.inputs.structure.pk, descendant.outputs.structure.pk))
                except NotExistentAttributeError as e:
                    print('  {} process {} has no structure output.'.format(
                          descendant.process_label, descendant.pk))
        if descendant.process_label == 'VaspCalculation':
            print('    {}<{}>: {} - {}.'.format(descendant.process_label,
                  descendant.pk, descendant.exit_status, descendant.exit_message))
            print('    Input structure: {}'.format(descendant.inputs.structure.pk))
            try:
                print('    Output structure: {}'.format(descendant.outputs.structure.pk))
            except NotExistentAttributeError:
                print('    {} process {} does not have an output structure.'.format(
                      descendant.process_label, descendant.pk))
            if not descendant.exit_status:
                try:
                    print('      final_energy = {} eV.'.format(
                        descendant.outputs.misc['total_energies']['energy_extrapolated']))
                    print('      electronic_converged = {}.'.format(
                        descendant.outputs.misc['run_status']['electronic_converged']))
                    if descendant.outputs.misc['run_status']['ionic_converged'] is not None:
                        print('      inonic_converged: {} ; ionic steps = {}.'.format(
                            descendant.outputs.misc['run_status']['ionic_converged'],
                            descendant.outputs.misc['run_status']['last_iteration_index'][0]))
                except NotExistentAttributeError:
                    print('    {}<{}> does not have a misc output.'.format(
                        descendant.process_label, descendant.pk))
                except KeyError as e:
                    print('    {}<{}>: KeyError: {}.'.format(
                        descendant.process_label, descendant.pk, e))


def postprocess_modify_and_relax_calculations(incoming_group,
        relaxed_struct_group_basename, modified_struct_group_name,
        modif_process_types=[WorkFunctionNode],
        modif_labels=None, modif_process_labels=['change_lattice_to_ref_and_assign'],
        modified_struct_group_description='',
        relax_process_label='VaspMultiStepRelaxWorkchain',
        relax_setting='fixed_cell_accurate',
        data_basename='relaxed_data', verbosity=1,
        show_data_dicts_in_group=False):
    """
    Perform the post-processing steps on structures modified and then relaxed

    Post-processing steps include:
        - Store structures modified with the selected processes in a group
        - Identify and relabel to 'failed' relaxation processes outgoing
          from the modified structures

    Args:
        incoming_group: str or Group
            Group (or group label) containing the initial structures

        modified_struct_group_name: str
            Label that will be given to the group containing structures
            modified with the chosen modification process(es)

        relaxed_struct_group_basename: str
            will be used to construct the label that will be given to the
            group containing the final relaxed structures hacing previously
            undergone (one of) the modifiation process(es):
            relaxed_struct_group.label = relaxed_struct_group_basename + '_' +
                relax_setting

        modif_process_types: list of process classes
            Types of mofidication processes that should be incuded in the search
            may include WorkFunctionNode, CalcFunctionNode, etc.

        modif_process_labels: list of str
            List of process_label attributes of the structure modification
            processes that should be included in the search (e.g.
            'make_supercell', 'change_lattice_to_ref_and_assign',
            'change_lattice_make_supercell_and_assign', etc.)
            Note that the search is currently not recursive. All modification
            processes should be combined in a single workfuntion.
            The corresponding process_types should be inbluded in
            modif_process_types.

        modif_labels: list of str (default: None)
            List of labels of the searched modification processes.
            Such labels may used to distinguish between similar processes
            conducted with different settings, for example.

        relax_process_label: str (default: 'VaspMultiStepRelaxWorkchain')
            process_label attribute of the relaxation process

        relax_setting: str
            label of the relaxation process, used here to designate
            the relaxation settings (e.g. 'fixed_cell_rough', 'free_cell_accurate')

    Returns: Tuple of:
        modified_struct_group: Group
            Group containing all structures modified with (one of) the
            specified modification process(es).

        relaxed_struct_group: Group
            Group containing all structures relaxed with the specified
            relaxation process and relax_setting (setting) after having
            undergone (one of) the specified modification process(es).

        data_Dict: Dict
            Dict containing information retrieved on all relaxed structures
            relaxed_struct_group.
    """

    # Store structures modified with the selected processes in a group
    modified_struct_group = store_modified_struct_in_group(
        incoming_struct_group=incoming_group,
        output_group_label=modified_struct_group_name,
        modif_process_types=modif_process_types,
        modif_process_labels=modif_process_labels,
        modif_labels=None,
        output_group_description=modified_struct_group_description,
        verbosity=verbosity)

    # Identify and relabel to 'Failed' relaxation processes incoming
    # to the initial structures in incoming_group
    relabel_processes_with_identical_in_and_out_struct(
        group_or_label=modified_struct_group,
        process_label=relax_process_label,
        node_labels=relax_setting, test_mode=False,
        verbosity=verbosity)

    # assign an original structure to each structure in modified_struct_group
    assign_original_structures_in_group(modified_struct_group,
                                        verbosity=verbosity)

    # Store relaxed structures in a group
    relaxed_struct_group = store_relaxed_struct_in_group(
        incoming_struct_group=modified_struct_group,
        output_group_basename = relaxed_struct_group_basename,
        relax_setting=relax_setting, verbosity=verbosity)

    # Assign an original structure to each structure in relaxed_struct_group
    assign_original_structures_in_group(relaxed_struct_group,
                                        verbosity=verbosity)

    # Create a dict containing relevant information on the relaxed structures
    # and store it in relaxed_struct_group
    data_Dict = retrieve_and_store_relaxed_data(relaxed_struct_group,
        relax_setting, include_lowest_E_C=False, plot_figure=False,
        store_data=True, data_basename=data_basename,
        data_description=None,
        include_cell_disto_to_ref='free_cell' in relax_setting,
        include_xray_pdf_rmsd=True)
    if show_data_dicts_in_group:
        # Print information on data Dicts with the same label
        # in the group.
        get_data_dicts_in_group(relaxed_struct_group,
                                dict_labels=[data_basename + '_' + relax_setting],
                                verbosity=verbosity)

    result = (modified_struct_group, relaxed_struct_group, data_Dict)
    if verbosity >= 1:
        print('The following Group and Dict instances have been created or updated :')
        for node in result:
            if isinstance(node, Group):
                node_type='Group'
            elif isinstance(node, Dict):
                node_type='DIct'
            print(f' * {node_type}<{node.pk}>: with label: {node.label}')

    return result
    # End of function postprocess_modify_and_relax_calculations


def show_relax_statistics(relax_processes_or_ids, properties=None):
    """
    Print a table with relaxation statistics

    Uses the multistep_summary output of the VaspMultiStepRelaxWorkchain

    Args:
        relax_processes_or_ids: list of VaspMultiStepRelaxWorkchain PKs, UUIDs or nodes

        properties: None, str, or list or str (default is None)
            if None, all properties present in multistep_summary output
            of the first process will be showed
            Otherwise only properties listed

    """
    if isinstance(relax_processes_or_ids, (int, str)):
        relax_processes_or_ids = [load_node(relax_processes_or_ids)]
    elif isinstance(relax_processes_or_ids, WorkChainNode):
        relax_processes_or_ids = [relax_processes_or_ids]
    elif not isinstance(relax_processes_or_ids, (list, tuple)):
        sys.exit('relax_processes_or_ids should be a (list of) WorkChainNode(s), ' +
                 'UUID(s) or PK(s).')

    data = []

    # Edit properties from first process:
    if isinstance(relax_processes_or_ids[0], (int, str)):
        relax_process = load_node(relax_processes_or_ids[0])
    elif not isinstance(relax_processes_or_ids[0], WorkChainNode):
            sys.exit(('relax_process is of type {} whereas it should be a ' +
                      'WorkChainNode.').format(type(relax_processes_or_ids[0])))
    else:
        relax_process = relax_processes_or_ids[0]
    summary_dict = relax_process.outputs.multistep_summary.get_dict()
    avail_properties = list(summary_dict.keys())
    nb_of_steps = len(summary_dict[avail_properties[0]])

    if properties is None:
        properties = avail_properties
    elif isinstance(properties, str):
        properties = [properties]

    for prop in properties:
        if prop in avail_properties:
            print('\n---------------------------------------------------')
            print('{:12}{}'.format('', prop))
            print('---------------------------------------------------')
            header = '{:12}'.format('Process')
            for step_index in range(nb_of_steps):
                header += '{:12}'.format(f'Step {step_index}')
            print(header)
            print('---------------------------------------------------')

            for relax_index, relax_process in enumerate(relax_processes_or_ids):
                if isinstance(relax_process, (int, str)):
                    relax_process = load_node(relax_process)
                if not isinstance(relax_process, WorkChainNode):
                    sys.exit(('relax_process is of type {} whereas it should be a ' +
                              'WorkChainNode.').format(type(relax_process)))
                # Read multistep_summary output
                summary_dict = relax_process.outputs.multistep_summary.get_dict()
                row = '{:<12}'.format(relax_process.pk)
                for step_index in range(len(summary_dict[prop])):
                    # DEBUGGING
                    # print(type(summary_dict[prop][step_index]))
                    if isinstance(summary_dict[prop][step_index], float):
                        row += '{:<12.3f}'.format(summary_dict[prop][step_index])
                    else:
                        row += '{:<12}'.format(str(summary_dict[prop][step_index]))

                print(row)
            print('---------------------------------------------------')

    # df = pandas.DataFrame([data])


def get_processes_from_nodes_or_ids(processes_or_ids, convert_to_single=False):
    if isinstance(processes_or_ids, (int, str)):
        processes = [load_node(processes_or_ids)]
    elif isinstance(processes_or_ids, WorkChainNode):
       processes = [processes_or_ids]
    elif not isinstance(processes_or_ids, (list, tuple)):
        sys.exit('processes_or_ids should be a (list of) WorkChainNode(s), ' +
                 'UUID(s) or PK(s).')
    else:  # relax_processes_or_ids is a list or tuple
        processes = []
        for process_or_id in processes_or_ids:
            if isinstance(process_or_id, (int, str)):
                process_or_id = load_node(process_or_id)
            if isinstance(process_or_id, WorkChainNode):
                processes.append(process_or_id)
            else:
                sys.exit('Elements in processes_or_ids should be valid WorkChainNode(s), ' +
                 'UUID(s) or PK(s).')
    if convert_to_single and len(processes) == 1:
        [processes] = processes
    return processes

def get_process_from_node_or_id(process_or_id):
    return get_processes_from_nodes_or_ids(process_or_id, convert_to_single=True)


def get_relax_statistics(relax_processes_or_ids, properties=None, verbosity=1):
    """
    Create a table with relaxation statistics using Pandas

    Uses the multistep_summary output of the VaspMultiStepRelaxWorkchain

    Args:
        relax_processes_or_ids: list of VaspMultiStepRelaxWorkchain PKs, UUIDs or nodes

        properties: None, str, or list or str (default is None)
            if None, all properties present in multistep_summary output
            of the first process will be showed
            Otherwise only properties listed

    """
    # Check relax_processes_or_ids and transform to list of processes if necessary
    relax_processes = get_processes_from_nodes_or_ids(relax_processes_or_ids)

    # Get relaxation process properties
    data_dict = {
        'relax_id': [node.pk for node in relax_processes],
        'relax_exit_status': [node.exit_status for node in relax_processes],
        'relax_input_struct_id': [node.inputs.structure.pk for node in relax_processes],
        'relax_output_struct_id':[],
        'relax_final_energy': [],
        'relax_max_force': [],
        'relax_ionic_steps': [],
        }
    calc_properties = ['calc_id', 'process_exit_status', 'input_struct_id',
                       'output_struct_id', 'energy', 'ionic_steps', 'max_force',
                       'ionic_converged']

    for relax_index, relax_node in enumerate(relax_processes):
        if verbosity >= 3:
            print('{}<{}>: exit status {} ({}).'.format(
                relax_node.process_label, relax_node.pk,
                relax_node.exit_status, relax_node.exit_message))
        try:
            data_dict['relax_output_struct_id'].append(relax_node.outputs.relax.structure.pk)
        except Exception as e:
            data_dict['relax_output_struct_id'].append(None)
            if verbosity >= 3:
                print(e)
        try:
            data_dict['relax_final_energy'].append(relax_node.outputs.misc[
                'total_energies']['energy_extrapolated'])
        except Exception as e:
            data_dict['relax_final_energy'].append(None)
            if verbosity >= 3:
                print(e)
        try:
            data_dict['relax_max_force'].append(relax_node.outputs.misc[
                'maximum_force'])
        except Exception as e:
            data_dict['relax_max_force'].append(None)
            if verbosity >= 3:
                print(e)
        try:
            data_dict['relax_ionic_steps'].append(sum(relax_node.outputs.multistep_summary[
                'ionic_steps']))
        except Exception as e:
            data_dict['relax_ionic_steps'].append(None)
            if verbosity >= 3:
                print(e)

        calc_index = -1
        for descendant in relax_node.called_descendants:
            if isinstance(descendant, CalcJobNode):
                calc_index += 1
                if verbosity >= 3:
                    print('  Calc {}: {}<{}> with exit status {} ({}).'.format(
                        calc_index, descendant.process_label, descendant.pk,
                        descendant.exit_status, descendant.exit_message))
                for prop in calc_properties:
                    prop_name = 'calc_{:0d}_{}'.format(calc_index, prop)
                    prop_value = 'N.C.'
                    # Initialize property if not existent
                    if prop_name not in list(data_dict.keys()):
                        data_dict[prop_name] = ['N.C.']*len(relax_processes)
                    if prop == 'calc_id':
                        prop_value = descendant.pk
                    if prop == 'input_struct_id':
                        prop_value = descendant.inputs.structure.pk
                    if prop == 'process_exit_status':
                        try:
                            prop_value = int(descendant.exit_status)
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)
                    if prop == 'energy':
                        try:
                            prop_value = descendant.outputs.misc['total_energies']['energy_extrapolated']
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)
                    if prop == 'ionic_steps':
                        try:
                            prop_value = int(descendant.outputs.misc['run_status']['last_iteration_index'][0])
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)
                    if prop == 'max_force':
                        try:
                            prop_value = descendant.outputs.misc['maximum_force']
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)
                    if prop == 'ionic_converged':
                        try:
                            prop_value = descendant.outputs.misc['run_status']['ionic_converged']
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)
                    if prop == 'output_struct_id':
                        try:
                            prop_value = descendant.outputs.structure.pk
                        except Exception as e:
                            if verbosity >= 3:
                                print(e)

                    data_dict[prop_name][relax_index] = prop_value
    if verbosity >= 3:
        for k, v in data_dict.items():
            print('{} of length {}'.format(k, len(v)))
    df = pd.DataFrame(data_dict)
    if verbosity >= 2:
        print(df)
    return df

def get_cell_param_and_diff_to_ref(struct_or_ids,
                                        ref_struct_or_id=None,
                                        verbosity=1):
    """
    Get a table of unit cell parameters for a (list of) structure(s) or id(s)

    A comparison with the cell parameters of one or a set of reference
    structure(s) will be made if ref_struct_or_id is not None

    Args:
        struct_or_ids: int, str, StructureData or a list or tuple thereof
            the (list of) structure(s) to get the table of cell parameters from.
        ref_struct_or_id: None, int, str, StructureData or a list or tuple thereof
            if None no comparisons to other structures will be made.
            if a single structure or id os provided, all structures will be compared
            to the same reference structure
            if a list or tuple of the same length as struct_or_ids is provided,
            each structure will be compared to the reference structure with the
            same index.
        verbosity: int (default: 1)
            The table will be printed if verbosity >= 2

    Returns:
        data: pandas.DataFrame
    """

    structures = get_list_of_structures(struct_or_ids)

    a = np.asarray([struct.cell_lengths[0] for struct in structures])
    b = np.asarray([struct.cell_lengths[1] for struct in structures])
    c = np.asarray([struct.cell_lengths[2] for struct in structures])
    alpha = np.asarray([struct.cell_angles[0] for struct in structures])
    beta = np.asarray([struct.cell_angles[1] for struct in structures])
    gamma = np.asarray([struct.cell_angles[2] for struct in structures])

    data_dict = {'struct_id': [struct.pk for struct in structures],
                 'a': a, 'b': b, 'c': c,
                 'alpha': alpha, 'beta': beta, 'gamma': gamma }

    make_comparisons = False
    ref_structures = None
    if ref_struct_or_id is not None:
        if isinstance(ref_struct_or_id, (list, tuple)):
            if len(ref_struct_or_id) != len(structures):
                print('WARNING: lengths of reference and focus structures should ' +
                      'be the same. No comparison will be made.')
            else:
                make_comparisons = True
                ref_structures = get_list_of_structures(ref_struct_or_id)

                a_0 = np.asarray([struct.cell_lengths[0] for struct in ref_structures])
                b_0 = np.asarray([struct.cell_lengths[1] for struct in ref_structures])
                c_0 = np.asarray([struct.cell_lengths[2] for struct in ref_structures])
                alpha_0 = np.asarray([struct.cell_angles[0] for struct in ref_structures])
                beta_0 = np.asarray([struct.cell_angles[1] for struct in ref_structures])
                gamma_0 = np.asarray([struct.cell_angles[2] for struct in ref_structures])

                distos = [measure_cell_disortions(struct, ref_struct)
                          for (struct, ref_struct) in zip(structures, ref_structures)]

        elif isinstance(ref_struct_or_id, (StructureData, int, str)):
            ref_struct = get_structure_from_pk_or_node(ref_struct_or_id)
            make_comparisons = True
            # A common reference structure is used for all structures
            a_0 = ref_struct.cell_lengths[0]
            b_0 = ref_struct.cell_lengths[1]
            c_0 = ref_struct.cell_lengths[2]
            alpha_0 = ref_struct.cell_angles[0]
            beta_0 = ref_struct.cell_angles[1]
            gamma_0 = ref_struct.cell_angles[2]

            distos = [measure_cell_disortions(struct, ref_struct)
                      for struct in structures]
        else:

            print('WARNING: ref_struct_or_id type should be: str, int, ' +
                  'StructureData, list or tuple. No comparison will be made.')

    if make_comparisons:
        if ref_structures is not None:
            data_dict['ref_struct_id'] = [struct.pk for struct in ref_structures]
        else:
            data_dict['ref_struct_id'] = [ref_struct.pk]*len(structures)
        data_dict['Delta_a'] = (a - a_0) / a_0
        data_dict['Delta_b'] = (b - b_0) / b_0
        data_dict['Delta_c'] = (c - c_0) / c_0
        data_dict['Delta_alpha'] = (alpha - alpha_0) / alpha_0
        data_dict['Delta_beta'] = (beta - beta_0) / beta_0
        data_dict['Delta_gamma'] = (gamma - gamma_0) / gamma_0
        data_dict['size_disto'] = np.asarray([disto[0] for disto in distos])
        data_dict['shape_disto'] = np.asarray([disto[1] for disto in distos])
        data_dict['disto_aniso'] = np.asarray([disto[2] for disto in distos])
        data_dict['cell_corners_msd'] = np.asarray([disto[3] for disto in distos])

    data = pd.DataFrame(data_dict)
    if verbosity >= 1:
        print(data)

    return data


def get_lowest_E_C_struct_from_initial(initial_struct_or_id, verbosity=1):
    """
    ADD DESCRIPTION
    """
    initial_struct = get_structure_from_pk_or_node(initial_struct_or_id)

    qb = QueryBuilder()
    qb.append(StructureData,
              filters={'id': initial_struct.pk},
              tag='AsTe_config')
    qb.append(cls=CalcFunctionNode, with_incoming='AsTe_config',
              filters={
                'attributes.process_label': 'introduce_Te_ox_state_disorder',
                'attributes.process_state': 'finished',
                'attributes.exit_status': 0,
              },
              tag='Te_ox_state_disorder_process',
              project=['*']
              )
    qb.append(cls=CalcJobNode, with_ancestors='Te_ox_state_disorder_process',
              filters={
                'attributes.process_label': 'SupercellCalculation',
                'attributes.process_state': 'finished',
                'attributes.exit_status': 0,
              },
              tag='supercell_process',
              project=['*'])
    qb.append(cls=StructureData, with_incoming='supercell_process',
              tag='lowest_E_C_structure',
              project=['*'])
    if verbosity >= 2:
        print(qb.all())

    return qb.first()[-1]


def export_structure(structure_or_id, file_name, fmt=None):

    structure = get_structure_from_pk_or_node(structure_or_id)

    try:
        if fmt is not None:
            structure.get_pymatgen().to(filename=file_name, fmt=fmt)
        else:
            structure.get_pymatgen().to(filename=file_name)

    except Exception as e:
        print(e)
        from ase.io import write
        try:
            ase_struct = structure.get_ase()
            result = write(file_name, structure.get_ase())
        except Exception as e:
            print(e)

# End of module local_utils


def get_distance_between_structures(struct_or_id_1, struct_or_id_2,
                                    r_max=10, sigma=0.01,
                                    show_plot=False, verbosity=1):

    struct_1 = get_structure_from_pk_or_node(struct_or_id_1)
    struct_2 = get_structure_from_pk_or_node(struct_or_id_2)
    pmg_struct_1 = struct_1.get_pymatgen()
    pmg_struct_2 = struct_2.get_pymatgen()
    dmd = dt.distanceMatrixData(Rmax=r_max, sigma=sigma)
    dist = dmd.calculate_cosine_distance(pmg_struct_1, pmg_struct_2,
                                          showPlot=show_plot)
    if verbosity >= 2:
        print('Cosine distance between structures {} and {}: {:.3f}.'.format(
              struct_1.pk, struct_2.pk, dist))
    return dist


def plot_distance_matrix(struct_or_ids, r_max=10, sigma=0.02,
                         system_name='AsTe3', cmap=None, tickLabels=[]):
    """
    Plot a matrix of distances between structures with pyama tools
    """
    structures = get_list_of_structures(struct_or_ids)
    dmd = dt.distanceMatrixData(Rmax=r_max, sigma=sigma)
    structureIDs = [struct.pk if struct.pk is not None else 'no-PK'
                    for struct in structures]
    dmd.calculate_distance_matrix([struct.get_pymatgen() for struct in structures],
                                  structureIDs=structureIDs,
                                  systemName=system_name)
    fig, ax = dmd.plot_distance_matrix(cmap=cmap, tickLabels=tickLabels)
    return fig, ax


def plot_distance_matrix_from_group(group_or_label, data_dict_label='',
        sort_by_energy=True, max_nb_of_struct=20,
        r_max=10, sigma=0.02, verbosity=1):
    """
    Plot distance matrix for structures in a group, possibly sorted by energy.
    """
    group = get_group_from_str_or_instance(group_or_label)
    data_Dict = get_data_dicts_in_group(group, dict_labels=[data_dict_label],
                                        order_by_ctime='desc',
                                        verbosity=verbosity)[0]
    if sort_by_energy:
        sorted_indexes = np.argsort(data_Dict['relaxed_energies'])
        struct_pks = [data_Dict['relaxed_struct_pks'][index] for index in
                      sorted_indexes]
    else:
        struct_pks = data_Dict['relaxed_struct_pks']

    fig = plot_distance_matrix(struct_pks[:max_nb_of_struct], r_max=r_max,
                               sigma=sigma)
    return fig


def get_relax_processes_incoming_to_group(group_or_label, relax_setting,
                                      relax_process_label='VaspMultiStepRelaxWorkchain'):
    """
    Get the relaxation processes that lead to structures in a group
    """
    group = get_group_from_str_or_instance(group_or_label)
    qb = QueryBuilder()
    qb.append(Group, filters={'label': group.label}, tag='group')
    qb.append(StructureData, with_group='group', tag='relaxed_structures', project=['id'])
    qb.append(WorkChainNode, with_outgoing='relaxed_structures',
              filters={
                'attributes.process_label': relax_process_label,
                'attributes.exit_status': 0
              }, project=['*'])
    return [row[1] for row in qb.all()]


def get_initial_structures_incoming_to_group(group_or_label, relax_setting,
        relax_process_label='VaspMultiStepRelaxWorkchain',
        modif_process_types=[WorkFunctionNode],
        modif_process_labels=['change_lattice_to_ref_and_assign',
                              'change_lattice_make_supercell_and_assign',
                              'make_supercell_and_assign'],
        return_full_search_result=False):
    """
    Get the initial structures that lead to structures in a group after modification and relaxation processes

    In the case where no modification process is used the function get_relax_processes_incoming_to_group
    may be used instead as follows:
        process_nodes = get_relax_processes_incoming_to_group(...)
        initial_strucytures = [node.inputs.structure for node in process_nodes]

    """
    group = get_group_from_str_or_instance(group_or_label)
    qb = QueryBuilder()
    qb.append(Group, filters={'label': group.label}, tag='group')
    qb.append(StructureData, with_group='group', tag='relaxed_structures', project=['id'])
    qb.append(WorkChainNode, with_outgoing='relaxed_structures',
              filters={
                'attributes.process_label': relax_process_label,
                'attributes.exit_status': 0
              },
              tag='relax_processes', project=['id'])
    qb.append(StructureData, with_outgoing='relax_processes',
              tag='modified_structures', project=['id'])
    qb.append(modif_process_types, with_outgoing='modified_structures',
              filters={
                'attributes.process_label': {'in': modif_process_labels},
                'attributes.exit_status': 0
              },
              tag='modif_processes', project=['id'])
    qb.append(StructureData, with_outgoing='modif_processes',
              edge_filters={'label': 'structure'},  # Exlude reference_structure input from node
              tag='initial_structures',
              project=['*'])
    initial_structures = [row[-1] for row in qb.all()]
    if return_full_search_result:
        search_results = {
            'init_struct_pks': [struct.pk for struct in initial_structures],
            'modif_proc_pks': [row[-2] for row in qb.all()],
            'modified_struct_pks': [row[-3] for row in qb.all()],
            'relax_struct_pks': [row[-4] for row in qb.all()],
            'relaxed_struct_pks': [row[-5] for row in qb.all()]
        }
        return initial_structures, search_results
    else:
        return initial_structures

def get_sorted_energy_data_from_group(group_or_label, nb_of_structures=None):
    """
    Get a dict of energy data sorted by increasing energy
    """
    _data_dict = get_data_dicts_in_group(group_or_label, order_by_ctime='desc')[0]
    sorted_indexes = np.argsort(_data_dict['relaxed_energies'])
    if isinstance(nb_of_structures, int):
        if nb_of_structures < len(sorted_indexes):
            sorted_indexes = sorted_indexes[:nb_of_structures]
    sorted_data_dict = {}
    for key in _data_dict.keys():
        sorted_data_dict[key] = [_data_dict[key][i] for i in sorted_indexes]
    return sorted_data_dict

def get_experimental_xray_total_scattering():
    """
    get a dict of experimental x-ray scattering file and parameters from the appropriate location on disk
    """
    pdf_data = {}
    pdf_data['file_name'] = ('/home/cadarp02/sujets-recherche/AsTe3/totalScattering/'
                        'X-ray-totalScattering_Oliver_2021-01/AsTe3-cristal_2021-10-21.dat')
    pdf_data['wavelength'] = 0.559422
    pdf_data['q_max'] = 21.795247  # Ag K_alpha_1
    pdf_data['title'] = 'AsTe3 - experimental X-ray total scattering'

    return pdf_data

def get_experimental_diffraction(diffraction_type='xrd'):
    """
    get a dict of experimental x-ray scattering file and parameters from the appropriate location on disk
    """
    diffraction_data = {}
    if diffraction_type.lower() in ['xrd', 'x-rays', 'x-ray']:
        diffraction_data['file_name'] = ('/home/cadarp02/sujets-recherche/AsTe3/'
            'Diffraction/AsTe3_sintered_SPS_XRD_2021-01-04.dat')
        # diffraction_data['wavelength'] = 1.54184 # 'CuKa' in pymatgen.analysis.diffraction.xrd  (Cu K_alpha_1)
        diffraction_data['title'] = 'AsTe3 (SPS-sintered) experimental X-ray diffraction'
    elif diffraction_type.lower() in ['neutron', 'neutrons', 'nd']:
        diffraction_data['file_name'] = ('/home/cadarp02/sujets-recherche/AsTe3/'
            'Diffraction/Neutron_AndreaPiarristeguy/261764_LAMPascii_Te65_429K')
        diffraction_data['title'] = 'AsTe3 (SPS-sintered) neutron diffraction at 429 K'
    return diffraction_data

def get_group_resource_consumption(group_or_label, verbosity=1):
    """
    Get a dict containing the resource consumption of all calculations in a group
    """
    group = get_group_from_str_or_instance(group_or_label)
    data_dict = get_data_dicts_in_group(group)[-1]
    group_consumption_dict = {
        'nb_of_relax_proc': 0,
        'nb_of_calc_proc': 0,
        'nb_of_calc_per_relax': [],
        'total_user_time_s': 0,
        'total_cpu_time_cpuh': 0,
        'user_time_per_relax_s': [],
        'cpu_time_per_relax_cpuh': [],
    }
    if verbosity >= 2:
        print('{} processes listed in data_dict {} with label {} in group {}<{}>'.format(
            len(data_dict['relax_process_pks']), data_dict.pk, data_dict.label,
            group.label, group.pk))
    for relax_proc_pk in data_dict['relax_process_pks']:
        relax_proc = load_node(relax_proc_pk)
        group_consumption_dict['nb_of_relax_proc'] += 1
        group_consumption_dict['nb_of_calc_per_relax'].append(0)
        group_consumption_dict['user_time_per_relax_s'].append(0.0)
        group_consumption_dict['cpu_time_per_relax_cpuh'].append(0.0)
        for descendant in relax_proc.called_descendants:
            if descendant.process_label == 'VaspCalculation':
                group_consumption_dict['nb_of_calc_proc'] += 1
                group_consumption_dict['nb_of_calc_per_relax'][-1] += 1
                try:
                    user_time_s = descendant.outputs.misc['run_stats']['total_cpu_time_used']
                    cpu_time_cpuh = user_time_s / 3600 * relax_proc.inputs.options['resources'][
                        'num_machines'] * relax_proc.inputs.options['resources']['num_mpiprocs_per_machine']
                    group_consumption_dict['total_user_time_s'] += user_time_s
                    group_consumption_dict['user_time_per_relax_s'][-1] += user_time_s
                    group_consumption_dict['total_cpu_time_cpuh'] += cpu_time_cpuh
                    group_consumption_dict['cpu_time_per_relax_cpuh'][-1] += cpu_time_cpuh
                except Exception as e:
                    if verbosity >= 2:
                        print('{}<{}>, {}<{}>: could not retrieve calc time or resources.\n{}'.format(
                            relax_proc.process_label, relax_proc.pk, descendant.process_label,
                            descendant.pk, e))
    if verbosity >= 2:
        print('group_consumption_dict = ', group_consumption_dict)
    return group_consumption_dict


def load_reference_systems(labels=None, pattern=None,
                           ref_struct_group_label='reference_structures',
                           convert_to_single=True, verbosity=1):
    if isinstance(labels, str):
        label_filter = labels
    elif isinstance(labels, (list, tuple)):
        label_filter = {'in': labels}
    if isinstance(pattern, str):
        label_filter = {'like': '%{}%'.format(pattern)}

    qb = QueryBuilder()
    qb.append(Group, filters={'label': ref_struct_group_label}, tag='ref_struct_group')
    qb.append(StructureData, with_group='ref_struct_group',
              filters={'label': label_filter}, tag='structures')
    ref_systems = [row[0] for row in qb.all()]
    if verbosity >= 1:
        for ref_system in ref_systems:
            print('Reference system {} with label {} loaded from group {}'.format(
                ref_system.pk, ref_system.label, ref_struct_group_label))
    if convert_to_single and len(ref_systems) == 1:
        [ref_systems] = ref_systems
    return ref_systems


def find_structures_without_dos_and_bands(group_or_name,
        dos_and_bands_process_label='DosAndBandsWorkChain',
        dos_and_bands_workchain_label='dos_and_bands_pbe',
        searched_exit_status=0, verbosity=1):
    """
    Find structures that do not yet have DOS or bands in a group

    IN PROGRESS
    """
    # Query for structures in group
    group = get_group_from_str_or_instance(group_or_name)
    qb = QueryBuilder()
    qb.append(Group, filters={'label': group.label}, tag='group')
    qb.append(StructureData, with_group='group', tag='structures',
              project='id')
    struct_pks_in_group = set([row[0] for row in qb.all()])
    if verbosity >= 2:
        print('{} structures found in group {}<{}>'.format(len(struct_pks_in_group),
            group.label, group.pk))
    qb.append(WorkChainNode, with_incoming='structures',
              filters={
                    'attributes.process_label': dos_and_bands_process_label,
                    'attributes.process_state': 'finished',
                    'attributes.exit_status': searched_exit_status,
                    'label': dos_and_bands_workchain_label,
                  }
              )
    struct_pks_with_dab = set([row[0] for row in qb.all()])
    struct_pks_without_dab = list(struct_pks_in_group - struct_pks_with_dab)

    if verbosity >= 1:
        print(('Among {} structures in group {}<{}>, {} have been used for DOS and '
               'BANDS calculation with {} and label \'{}\' and {} have not.').format(
            len(struct_pks_in_group), group.label, group.pk, len(struct_pks_with_dab),
            dos_and_bands_process_label, dos_and_bands_workchain_label,
            len(struct_pks_without_dab)))

    return struct_pks_without_dab

def sort_structures_in_group_by_energy(group_or_name, structure_pks=None,
                                       relax_process_label='VaspMultiStepRelaxWorkchain',
                                       verbosity=0):
    """
    Sort structures in a group by the energy of the relaxation process that was used
    to create them
    """
    group = get_group_from_str_or_instance(group_or_name)
    qb = QueryBuilder()
    qb.append(Group, filters={'label': group.label}, tag='group')
    filters = {}
    if isinstance(structure_pks, (list, tuple)):
        filters['id'] = {'in': structure_pks}
    qb.append(StructureData, with_group='group', tag='structures',
              filters=filters, project='id')
    qb.append(WorkChainNode, with_outgoing='structures',
              filters={
                    'attributes.process_label': relax_process_label,
                    'attributes.process_state': 'finished',
                    'attributes.exit_status': 0,
              }, project='*', tag='relax_processes')
    query_results = qb.all()
    relax_labels = set([row[1].label for row in query_results])
    if len(relax_labels) > 1:
        print('WARNING: structures in group {}<{}> have been relaxed with '
              ' processes with {} different labels: {}'.format(group.label, group.pk,
              len(relax_labels), relax_labels))
    relaxed_struct_pks = [row[0] for row in query_results]
    relaxed_energies = [row[1].outputs.misc['total_energies']['energy_extrapolated']
                        for row in query_results]
    sorted_indexes = np.argsort(relaxed_energies)
    sorted_struct_pks = [relaxed_struct_pks[i] for i in sorted_indexes]
    if verbosity >= 1:
        print('{} structures have been sorted by energy'.format(len(sorted_struct_pks)))

    return sorted_struct_pks


def build_and_submit_dos_and_bands(code, options, structure, parameters,
                                   potential_family, potential_mapping=None,
                                   parser_settings=None, perform_relax=Bool(False),
                                   relax_kpoints_distance=Float(0.25), relax_kpoints=None,
                                   dos_kpoints_distance=Float(0.125), dos_kpoints=None,
                                   bands_kpoints_distance=Float(0.125), bands_sigma=Float(0.01),
                                   process_description=None, verbosity=1):
    """
    Build and submit a preprun, DOS and bands calculation with the DosAndBandsWorkChain

    Args:
        parser_settings: dict, AttributeDict or None (default is None)
            content of inputs.settings['parser_settings']

    """

    # First, we need to fetch the AiiDA datatypes which will
    # house the inputs to our calculation
    dict_data = DataFactory('dict')
    kpoints_data = DataFactory('array.kpoints')

    # Then, we set the workchain you would like to call
    workchain = DosAndBandsWorkChain

    # And finally, we declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()

    # Organize settings
    if not parser_settings:
        settings.parser_settings = AttributeDict({})  # parser settings will be set
                                       # automatically by the DosAndBandsWorkChain
    if isinstance(parser_settings, AttributeDict):
        settings.parser_settings = parser_settings
    else:
        settings.parser_settings = AttributeDict(parser_settings)
    inputs.settings = dict_data(dict=settings)

    # Set inputs for the following WorkChain execution
    # Set code
    if isinstance(code, str):
        inputs.code = Code.get_from_string(code)
    else:
        inputs.code = code

    # Set structure
    inputs.structure = structure

    # Set parameters
    if not parameters:
        # Define a minimal set of input (INCAR) parameters
        inputs.parameters = AttributeDict({
            'incar': {
                'gga': 'PE',
                'ediff': 1e-5,
                'lreal': 'Auto',
                'ncore': 10,
            }
        })  # to be completed ?
    elif isinstance(parameters, Dict):
        inputs.parameters = parameters
    else:
        inputs.parameters = dict_data(dict=parameters)

    # Set potentials and their mapping
    # TODO: load default potential_family of None
    if isinstance(potential_family, str):
        inputs.potential_family = Str(potential_family)
    else:
        inputs.potential_family = Str(potential_family)

    if not potential_mapping:
        inputs.potential_mapping = get_default_potential_mapping(inputs.structure)
    elif isinstance(potential_mapping, Dict):
        inputs.potential_mapping = potential_mapping
    else:
        inputs.potential_mapping = dict_data(dict=potential_mapping)
    # Set options

    if isinstance(options, Dict):
        inputs.options = options
    else:
        inputs.options = dict_data(dict=options)

    # Set workchain related inputs, in this case, give more explicit output to repor
    inputs.verbose = Bool(True)

    # Setting the relax parameters for the (prerun) relaxation or SPE
    relax = AttributeDict()
    if isinstance(perform_relax, Bool):
        relax.perform = perform_relax
    else:
        relax.perform = Bool(perform_relax)

    if isinstance(relax_kpoints, KpointsData):
        relax.kpoints = relax_kpoints
    else:
        relax.kpoints_distance = Float(relax_kpoints_distance)

    inputs.relax = relax

    # Bands inputs:
    bands = AttributeDict()
    if isinstance(bands_kpoints_distance, Float):
        bands.kpoints_distance = bands_kpoints_distance
    else:
        bands.kpoints_distance = Float(bands_kpoints_distance)
    bands.smearing = AttributeDict()
    # TODO: allow setting a bands_gaussian_smearing
    bands.smearing.gaussian = Bool(True)
    if isinstance(bands_sigma, Float):
        bands.smearing.sigma = bands_sigma
    else:
        bands.smearing.sigma = Float(bands_sigma)
    inputs.bands = bands

    # Dos inputs:
    dos = AttributeDict()
    if isinstance(dos_kpoints_distance, Float):
        dos.kpoints_distance = dos_kpoints_distance
    else:
        dos.kpoints_distance = Float(dos_kpoints_distance)
    inputs.dos = dos

    # Extract electronic band structure
    # inputs.extract_bands = Bool(True)
    # Submit the requested workchain with the supplied inputs

    if verbosity >= 1:
        print('Ready to submit {} workchain with the following inputs:\n{}'.format(
            workchain.__name__, inputs))

    if verbosity >= 2:# DEBUGGING:
        print('inputs.options:\n', inputs.options.get_dict())

    process_node = submit(workchain, **inputs)

    if process_description:
        process_node.description = process_description

    return process_node


def prepare_and_submit_dos_and_bands_calculations(code,
        potential_family, potential_mapping, options,
        nb_of_structures, max_nb_of_jobs=10, parameters=None, parser_settings=None,
        initial_struct_group_name='AsTe3_Te-based_sc181_relaxed_free_cell_accurate_dftd3bj',
        previous_relax_setting='free_cell_accurate',
        dos_and_bands_workchain_label='dos_and_bands_pbe',
        relax_kpoints_distance=0.25, dos_kpoints_distance=0.12,
        bands_kpoints_distance=0.12, bands_sigma=0.01,
        dry_run=False, verbosity=1):
    """
    Prepare a series of DOS and BANDS calculations for best-energy structures
    in a group

    TO BE COMPLETED
    """

    # Find structures witout dos or bands among those in the considered group
    struct_pks = find_structures_without_dos_and_bands(initial_struct_group_name,
        dos_and_bands_workchain_label=dos_and_bands_workchain_label, verbosity=verbosity)
    struct_pks = sort_structures_in_group_by_energy(initial_struct_group_name,
        structure_pks=struct_pks, verbosity=verbosity)

    # Pick the first nb_of_structures
    picked_struct_pks = struct_pks[:nb_of_structures]

    submitted_process_pks = []
    running_workflows_list = []
    for index, struct_pk in enumerate(picked_struct_pks):
        if dry_run:
            print(('Ready to run DOS and bands calculations with setting ('
                   'label) {} on structure {}.').format(
                       dos_and_bands_workchain_label, struct_pk))
        else:
            structure = load_node(struct_pk)

            # prepare and run relaxation
            workflow_queue(max_nb_of_jobs, running_workflows_list,
                           sleep_time_in_mn=1)

            # TODO: DEFINE  build_and_submit_dos_and_bands
            # DEFINE A GENERIC FUNCTION FOR THIS IN AIIDA_UTILS
            process_node = build_and_submit_dos_and_bands(code, options,
                structure, parameters, potential_family, potential_mapping,
                parser_settings=parser_settings,
                relax_kpoints_distance=relax_kpoints_distance,
                dos_kpoints_distance=dos_kpoints_distance,
                bands_kpoints_distance=bands_kpoints_distance,
                bands_sigma=bands_sigma, process_description=None, verbosity=verbosity)

            process_node.label = dos_and_bands_workchain_label

            print(('{}<{}> : submitting prerun, DOS, and BANDS with label ' +
                   '\'{}\' on structure {}').format(
                   process_node.process_label, process_node.pk,
                   process_node.label, structure.pk))
            # Limit number of jobs submitted to scheduler
            running_workflows_list.append(process_node)

    return submitted_process_pks


def plot_dos_by_site(dos_and_bands_process_or_id):
    """
    TO BE COMPLETED
    """

    """
    dos_and_bands_process = get_process_from_node_or_id(dos_and_bands_process_or_id)
    color_maps = {
        'Te': mpl.colormaps['Reds'],
        'As': mpl.colormaps['Greens']}
    """


def prepare_vasp_hybrid_bands_calculations(structure_or_id,
        export_dir, kpoints_distance=0.12):
    """
    TO BE COMPLETED

    This function shoud prepare all necessay files to do the following:
    Prepare band structure calculations at hybrid level for best-global-score structures:
    (1) run get_explicit_kpoints_path (from aiida.tools) to get:
            - a primitive structure
            - a list of explicit kpoints (with names) for the band structure calculation
    (2) export primitive structure as poscar
    (3) run SCF wih VASP ON PRIMITIVE STRUCTURE to:
            - get the WAVECAR for subsequent non-self-consistent bands calculation
            - get explicit kpoints from IBZKPT
    (3) concatenate IBZKPT and explicit k-points path for bands with 0 weight in KPOINTS file (do not forget to change actual number of k-points accordingly.
    (4) run bands calculation with VASP using the primitive structure

    """
    structure = get_structure_from_pk_or_node(structure_or_id)
    parameters = {
        "reference_distance": kpoints_distance
    }
    kpts_path = get_explicit_kpoints_path(structure, method='seekpath',
                                          **parameters)
    #TODO: Create export_dir if not there
    os.makedirs(export_dir, exist_ok=True)

    # export primitive structure as POSCAR
    kpts_path['primitive_structure'].get_pymatgen().to(
        filename=os.path.join(export_dir, 'POSCAR'),
        fmt='poscar')

    path_labels = {'indexes': [], 'normalized_x_ticks': [], 'labels': []}
    normalized_linearcoords = kpts_path['parameters']['explicit_kpoints_linearcoord'
        ] / np.max(kpts_path['parameters']['explicit_kpoints_linearcoord'])

    for index, label in kpts_path['explicit_kpoints'].labels:
        path_labels['indexes'].append(index)
        path_labels['normalized_x_ticks'].append(normalized_linearcoords[index])
        path_labels['labels'].append(label)

    with open(os.path.join(export_dir, 'path_labels.json'), 'w') as f:
        json.dump(path_labels, f)

    with open(os.path.join(export_dir, 'explicit_kpoints.txt'), 'w') as f:
        for kpt_coord in kpts_path['explicit_kpoints'].get_array('kpoints'):
            # write explicit k-point coordinates with 0 weight
            f.write('{:>20.8e}{:>20.8e}{:>20.8e}{:>10d}\n'.format(
                kpt_coord[0], kpt_coord[1], kpt_coord[2], 0))





